# Kubernetes glossary
- **Cluster** A set of worker machines, called [nodes](https://kubernetes.io/docs/concepts/architecture/nodes/), that run containerized applications. Every cluster has at least one worker node.
* **Node** A node may be a virtual or physical machine, depending on the cluster. Each node is managed by the [control plane](https://kubernetes.io/docs/reference/glossary/?all=true#term-control-plane) and contains the services necessary to run [Pods](https://kubernetes.io/docs/concepts/workloads/pods/). Typically you have several nodes in a cluster; in a learning or resource-limited environment, you might have only one node.
  The [components](https://kubernetes.io/docs/concepts/overview/components/#node-components) on a node include the [kubelet](https://kubernetes.io/docs/reference/generated/kubelet), a [container runtime](https://kubernetes.io/docs/setup/production-environment/container-runtimes), and the [kube-proxy](https://kubernetes.io/docs/reference/command-line-tools-reference/kube-proxy/)
* **POD** A _Pod_ (as in a pod of whales or pea pod) is a group of one or more [containers](https://kubernetes.io/docs/concepts/containers/), with shared storage and network resources, and a specification for how to run the containers. A Pod's contents are always co-located and co-scheduled, and run in a shared context. A Pod models an application-specific "logical host": it contains one or more application containers which are relatively tightly coupled. In non-cloud contexts, applications executed on the same physical or virtual machine are analogous to cloud applications executed on the same logical host.

# Kubernetes infrastructure at your home
Following chapters describe how to setup full k8s environment at your local machine. It might be very useful while testing kubernetes specific features of the `POD` or plugins, or for bug reproduction. Having everything set up locally significantly shortens the feedback loop.

To setup Kubernetes we're gonna use [minikube](https://minikube.sigs.k8s.io/docs/start/).

To view the Cluster state use [lens](https://k8slens.dev/)

## Start minikube
`minikube start `

## Enable DNS
That will allow to access k8s services using local URLs. Follow steps described [here](
https://minikube.sigs.k8s.io/docs/handbook/addons/ingress-dns/)

## Deploying a Docker image to the minikube
Minikube manages it's own instance of docker, with separate image registry. To build images for the minikube we need to use that registry, that can be done by calling:
* `eval $(minikube docker-env)`

Now this shell can be used to build the Docker images via `docker build ...args` command.

## Build plugin-tool-deployment
The `plugin-tool-deployment` container is monitoring for new plugins to show up in the cluster, and apply ingress configuration on them upon startup. That allows to reach them using local urls.

Build the container image in `minikube` [registry](#deploying-a-docker-image-to-the-minikube)
* `cd tools/local_k8s/plugin-tool-deployment/`
* `docker build -t plugin-tool-create-deployment -f Dockerfile_create .`

The image will be later used in `pod_deploy.yaml` configuration.
## Build local POD to the container
This will build the container of your local version of the POD, allows fix-and-check procedure to be quick.

NOTE: all commands has to be done from the same shell

* Build image in `minikube` image [registry](#deploying-a-docker-image-to-the-minikube)


  Note `use_kubernetes=true`, that adds `kubectl` executable into container
  `docker build -t pod-for-k8s --build-arg use_kubernetes=true .`
* Deploy POD to the minikube

  `kubectl apply -f tools/local_k8s/pod_deploy.yml`

* Use dns name: `http://pod.test` to access POD

## Pull image to k8s
Seems like fetching large containers by minikube is a problem, [returns an error after some time](https://github.com/kubernetes/minikube/issues/14806)

Solution is to manually pull the image first, example

 `minikube ssh docker pull gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest`

## Deploying plugins in k8s
* Note HTTPS ingress is not yet implemented for local k8s. POD needs to be patched accordingly:
```rust
/// Figure out correct URL depending on the environnement the Plugin will be run on.
pub fn resolve_webserver_url(item: &CreateItem, cli: &CliOptions) -> Result<String> {
    ...

    let webserver_url = if cli.use_kubernetes {
        // Kubernetes
        if let Some(base) = &cli.plugins_public_domain {
	        // format!("https://{}.{}", container_id, base)   CHANGE HTTPS TO HTTP
            format!("http://{}.{}", container_id, base)
        } ...
```
* You can deploy and start plugin using [POD HTTP API](./docs/HTTP_API.md)
* When plugin starts, the `plugin-tool-deployment` container will patch its ingress configuration using `attach-services.sh` that will expose the plugin webserver via local url.
* Having all above steps done correctly, POD and plugins are reachable via their URL addresses.
  POD is able to communicate with the plugin and vice versa. For example item triggering works.

## Side notes
## Kubectl command to run plugin, used by pod
```
kubectl run --restart=Never c0000000000-gitlabmemriio5050szi-4fa6094c51d64874a135-4b6ac9a0 '--labels=app=c0000000000-gitlabmemriio5050szi-4fa6094c51d64874a135-4b6ac9a0,type=plugin,owner=10eab6008d5642cf42abd2aa41f847cb' --port=8080 --image-pull-policy=Always '--image=gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest' '--overrides={"apiVersion":"v1","kind":"Pod","spec":{"containers":[{"env":[{"name":"POD_FULL_ADDRESS","value":"http://172.17.0.5:3030"},{"name":"POD_TARGET_ITEM","value":"{\"containerId\":\"c0000000000-gitlabmemriio5050szi-4fa6094c51d64874a135-4b6ac9a0\",\"containerImage\":\"gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest\",\"dateCreated\":1665349205938,\"dateModified\":1665349205938,\"dateServerModified\":1665349205938,\"deleted\":false,\"id\":\"4fa6094c-51d6-4874-a135-9017fcdedf18\",\"pluginModule\":\"plugin_to_trigger.plugin\",\"pluginName\":\"ClassifierPlugin\",\"rowid\":870,\"status\":\"idle\",\"targetItemId\":\"4fa6094c-51d6-4874-a135-9017fcdedf18\",\"type\":\"PluginRun\",\"webserverPort\":8008,\"webserverUrl\":\"http://localhost\"}"},{"name":"POD_PLUGINRUN_ID","value":"4fa6094c-51d6-4874-a135-9017fcdedf18"},{"name":"POD_OWNER","value":"0000000000000000000000000000000000000000000000000000000000000000"},{"name":"POD_AUTH_JSON","value":"{\"data\":{\"nonce\":\"72c936beb1a14bc7be48d8a4bf3c9416922fbc8f55c9abdb\",\"encryptedPermissions\":\"b33fe2b4514380eb2ed78bd070d6ddbc5027484879a4d08f5e04bf5a82cfae6bfe228a3b977a26c43f0fa4b191b3bd17\"}}"},{"name":"PLUGIN_DNS","value":"http://localhost"},{"name":"PYTHONUNBUFFERED","value":"1"}],"image":"gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest","imagePullPolicy":"Always","name":"c0000000000-gitlabmemriio5050szi-4fa6094c51d64874a135-4b6ac9a0","ports":[{"containerPort":8080,"protocol":"TCP"}],"resources":{"limits":{"cpu":"2","memory":"3.5Gi"},"requests":{"cpu":"1","memory":"2Gi"}}}]}}'
```