#!/usr/bin/env python3
import json
import sys
import os
import re
import subprocess
from shutil import which
import argparse
from time import sleep

RED = "\033[0;31m"
NC = "\033[0m"
GREEN = "\033[0;32m"


def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--namespace", help="kubernetes namespace", default="dev")
    parser.add_argument("--plugin-name", help="name of the plugin on k8s", required=True)
    parser.add_argument(
        "--plugin-alias", help="name you want to use to identify the attached plugin", default="attached-plugin")
    parser.add_argument(
        "--plugin-forward-port", help="port used to forward POD requests to K8S", default="3000")
    parser.add_argument(
        "--pod-listening-port", help="port that POD is listening for incoming requests", default="3030")

    args = parser.parse_args()

    plugin_name = args.plugin_name
    plugin_forward_port = args.plugin_forward_port
    pod_listening_port = args.pod_listening_port
    namespace = args.namespace
    plugin_alias = args.plugin_alias

    kubectl = which("kubectl")
    if not kubectl:
        print(f"kubectl command not found", file=sys.stderr)
        exit(1)

    ngrok = which("ngrok")
    if not ngrok:
        print(f"ngrok command not found", file=sys.stderr)
        exit(1)

    ngrok_pid = None
    kube_pid = None

    try:
        cmd = [kubectl, "port-forward", f"svc/{plugin_name}",
               f"{plugin_forward_port}:80", "-n", namespace]
        print(f"running {cmd} ...")

        kube_pid = subprocess.Popen(
            cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        handle_kubectl_output(kube_pid)

        cmd = [ngrok, "http", pod_listening_port, "--log=stdout"]
        print(f"running {cmd} ...")
        ngrok_pid = subprocess.Popen(
            cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)

        url = handle_ngrok_output(ngrok_pid)

        print(f"POD available on {RED}{url}{NC}, use that as callback URL for the plugin")

        attach_to_the_plugin(plugin_forward_port, plugin_alias)

        print()
        print(f"{GREEN}Port forwarding and tunneling are running. Ctrl + C to terminate them.{NC}")
        while True:
            input("")

    finally:
        if kube_pid:
            kube_pid.terminate()

        if ngrok_pid:
            ngrok_pid.terminate()


def attach_to_the_plugin(plugin_forward_port, plugin_alias):
    OWNER_KEY = "OWNER_KEY"
    DB_KEY = "DB_KEY"
    POD_URL = "POD_URL"

    pod_owner = os.getenv(OWNER_KEY)
    pod_url = os.getenv(POD_URL)
    db_key = os.getenv(DB_KEY)

    if all((pod_owner, pod_url, db_key)):
        print(f"Attaching to the plugin, from {pod_url} for owner {pod_owner}")
        request = json.dumps({
            "auth": {
                "type": "ClientAuth",
                "databaseKey": db_key
            },
            "payload": {
                "id": plugin_alias,
                "webserverPort": int(plugin_forward_port),
                "webserverUrl": "http://localhost"
            }
        })

        try:
            res = subprocess.run(check=True, timeout=5, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                 args=["curl", "--fail-with-body", "-X", "POST", "-H", "content-type: application/json", f"{pod_url}/v4/{pod_owner}/plugin/attach", "-d", request])
            print(f"Attached PluginRun ID {res.stdout.decode()}")
            print(f"{GREEN}Plugin attached successfully{NC}")
        except subprocess.CalledProcessError as err:
            print(f"{RED}{err.stderr}{NC}")
            raise

    else:

        print("To attach local pod to the plugin, run following command.")

        print(
            f"Please set the {RED}POD_URL, OWNER_KEY, DATABASE_KEY{NC} values in the request, you can also change the {GREEN}id{NC} of the plugin")

        print('''
curl -H "content-type: application/json" \\
    $POD_URL/v4/$OWNER_KEY/plugin/attach \\
    -d @- << EOF
    {j}
EOF
    '''.format(j=json.dumps(
            {
                "auth": {
                    "type": "ClientAuth",
                    "databaseKey": "DATABASE_KEY"
                },
                "payload": {
                    "id": plugin_alias,
                    "webserverPort": int(plugin_forward_port),
                    "webserverUrl": "http://localhost"
                }
            }, indent=4)))


def handle_kubectl_output(kube_pid):
    sleep(1)
    kube_out = kube_pid.stdout.readline()

    if len(kube_out) == 0 and kube_pid.poll() is not None:
        kube_err = kube_pid.stderr.read()
        print("Error while setting up kubectl forwarding:", file=sys.stderr)
        print(kube_err.decode("utf-8"), file=sys.stderr)
        raise RuntimeError("kubectl returned error")

    print(kube_out)


def handle_ngrok_output(ngrok_pid):
    sleep(1)
    if ngrok_pid.poll() is not None:
        ngrok_err = ngrok_pid.stderr.read()
        print("Error while setting up tunneling:", file=sys.stderr)
        print(ngrok_err.decode("utf-8"), file=sys.stderr)
        raise RuntimeError("ngrok returned error")

    line = ngrok_pid.stdout.readline()

    while b"started tunnel" not in line:
        line = ngrok_pid.stdout.readline()

    line = line.decode()
    url = re.search(r"url=([a-z0-9\:/\.\-]+)\s?", line).groups()[0]

    return url


if __name__ == "__main__":
    main()
