use criterion::{criterion_group, criterion_main, Criterion};
use libpod::{
    async_db_connection::{AsyncConnection, AsyncTx},
    database_api,
    internal_api::{graphql, search},
    test_helpers::{default_cli, get_memory_connection},
};

mod test_dataset;
use serde_json::json;

#[cfg(test)]
pub fn deep_join(c: &mut Criterion) {
    let runtime = tokio::runtime::Runtime::new().unwrap();

    let mut conn = get_memory_connection().unwrap();

    runtime.block_on(prepare_database(&mut conn));

    let schema = runtime
        .block_on(
            conn.in_read_transaction(|tx: AsyncTx| async move { database_api::get_schema(&tx) }),
        )
        .unwrap();

    // search
    let search_json = json!({
        "type": "Person1",
        "[[edges]]": {}
    });

    // graphql search
    let query = "
     query {
         Person1 {
            id
            friend1 {
                id
                 friend2 {
                     id
                 }
            }
         }
     }"
    .to_owned();

    c.bench_function("search api", |b| {
        b.to_async(tokio::runtime::Runtime::new().unwrap())
            .iter(|| {
                let schema_2 = schema.clone();
                let search_json_2 = search_json.clone();
                let mut conn_2 = AsyncConnection::from_other(&conn);

                async move {
                    conn_2
                        .in_read_transaction(|tx: AsyncTx| async move {
                            search(
                                &tx,
                                &schema_2,
                                serde_json::from_value(search_json_2).unwrap(),
                            )
                        })
                        .await
                }
            })
    });

    c.bench_function("graphql deep join", |b| {
        b.to_async(tokio::runtime::Runtime::new().unwrap())
            .iter(|| {
                let schema_2 = schema.clone();
                let mut conn_2 = AsyncConnection::from_other(&conn);
                let query_2 = query.clone();

                async move {
                    conn_2
                        .in_read_transaction(|tx: AsyncTx| async move {
                            graphql(&tx, &schema_2, query_2)
                        })
                        .await
                }
            })
    });
}

async fn prepare_database(conn: &mut AsyncConnection) {
    conn.set_prepared_statement_cache_capacity(0);

    conn.in_write_transaction(|tx: AsyncTx| async move {
        tx.execute("PRAGMA cache_size = 0;", [])?;
        Ok(())
    })
    .await
    .unwrap();

    conn.in_write_transaction(|tx: AsyncTx| async move {
        let cli = default_cli();
        let pod_owner = "";

        test_dataset::prepare_graphql_dataset(&tx, pod_owner, cli).await;

        let schema = database_api::get_schema(&tx).unwrap();

        // search
        let search_json = json!({
            "type": "Person1",
            "[[edges]]": {}
        });

        let res = search(&tx, &schema, serde_json::from_value(search_json).unwrap()).unwrap();
        assert_eq!(res.len(), 1000);

        // graphql search
        let query = "
        query {
            Person1 {
               id
               friend1 {
                   id
                    friend2 {
                        id
                    }
               }
            }
        }"
        .to_owned();

        let _res = graphql(&tx, &schema, query);

        // print!("{:}", serde_json::to_string_pretty(&json!(_res.unwrap())).unwrap());

        Ok(())
    })
    .await
    .unwrap();
}

criterion_group!(benches, deep_join);
criterion_main!(benches);
