use reqwest::Method;
use serde::de::{self, Visitor};
use serde::{Deserialize, Serialize};
use serde_json::Value;
use std::str::FromStr;
use std::{collections::HashMap, fmt::Debug, fmt::Display};

//
// Wrapper structs:
//
#[derive(Clone, Serialize, Deserialize, Debug)]
pub struct PodOwner(String);

impl Display for PodOwner {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let s = self.0.as_str();
        if s.len() < 12 {
            write!(f, "{s}")
        } else {
            write!(f, "{}..{}", &s[..5], &s[s.len() - 5..])
        }
    }
}

use std::ops::Deref;
impl Deref for PodOwner {
    type Target = String;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<String> for PodOwner {
    fn from(s: String) -> Self {
        Self(s)
    }
}

impl FromStr for PodOwner {
    type Err = core::convert::Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(PodOwner(s.into()))
    }
}

/// Plugin authentication data.
#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginAuthData {
    /// Nonce - the number used only once.
    pub nonce: String,
    /// Encrypted database permissions.
    pub encrypted_permissions: String,
}

/// Plugin authentication structure.
/// Please use create_plugin_auth to generate this.
#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginAuth {
    /// Plugin authentication data.
    pub data: PluginAuthData,
}

/// Client authentication structure.
#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct ClientAuth {
    /// Database authentication key.
    pub database_key: String,
}

/// Authentication key enum.
#[derive(Serialize, Deserialize, Debug)]
#[serde(tag = "type")]
pub enum AuthKey {
    /// Plugin authentication structure.
    PluginAuth(PluginAuth),
    /// Client authentication structure.
    ClientAuth(ClientAuth),
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PayloadWrapper<T> {
    pub auth: AuthKey,
    pub payload: T,
}

//
// Item API:
//

#[derive(Serialize, Deserialize, Debug, Clone, Default)]
#[serde(rename_all = "camelCase")]
pub struct CreateItem {
    pub rowid: Option<i64>,
    pub id: Option<String>,
    /// Mandatory type of the item, serialized as just "type" (without underscore)
    pub _type: String,
    pub date_created: Option<i64>,
    pub date_modified: Option<i64>,
    pub date_server_modified: Option<i64>,
    #[serde(default)]
    pub deleted: bool,
    #[serde(flatten)]
    pub fields: HashMap<String, Value>,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
#[serde(rename_all = "camelCase")]
pub struct OauthRequestTokenPayload {
    pub callback_url: String,
    pub service: String,
}

#[derive(Serialize, Deserialize, Debug, Clone)]
#[serde(rename_all = "camelCase")]
pub struct OauthAccessTokenPayload {
    pub oauth_token: String,
    pub oauth_verifier: String,
    pub oauth_token_secret: String,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub enum Platforms {
    Gitlab,
    Twitter,
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Oauth2AuthUrlRequest {
    /// list of platform specific scopes
    pub scopes: Vec<String>,
    pub redirect_uri: String,
    pub platform: Platforms,
}

#[derive(Serialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Oauth2AuthUrlResponse {
    /// list of platform specific scopes
    pub authorize_url: String,
    pub pkce_verifier: String,
    pub csfr_token: String,
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Oauth2AccessTokenRequest {
    pub platform: Platforms,
}

#[derive(Serialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Oauth2AccessTokenResponse {
    pub access_token: String,
    pub token_type: String,
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Oauth2AuthorizeTokenRequest {
    /// Code returned by Oauth server, used in exchange to access_code
    pub auth_code: String,
    /// pkce_challenge = SHA256(pkce_verifier) was used in auth_url,
    /// pkce_verifier is an additional piece required to pass while exchanging
    /// to access_code
    pub pkce_verifier: String,
    // Must be the same as in authorization url
    pub redirect_uri: String,
    pub platform: Platforms,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct UpdateItem {
    pub id: String,
    #[serde(flatten)]
    pub fields: HashMap<String, Value>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Bulk {
    #[serde(default)]
    pub create_items: Vec<CreateItem>,
    #[serde(default)]
    pub update_items: Vec<UpdateItem>,
    #[serde(default)]
    pub delete_items: Vec<String>,
    #[serde(default)]
    pub create_edges: Vec<CreateEdge>,
    #[serde(default)]
    pub search: Vec<Search>,
}

#[derive(Serialize, Deserialize, Debug, Default)]
#[serde(rename_all = "camelCase")]
pub struct BulkResponse {
    pub create_items: Vec<String>,
    pub create_edges: Vec<String>,
    pub search: Vec<Vec<Value>>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct CreateEdge {
    #[serde(rename = "_source")]
    pub source: String,
    #[serde(rename = "_target")]
    pub target: String,
    #[serde(rename = "_name")]
    pub name: String,
    #[serde(default, rename = "_self")]
    pub self_id: Option<String>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct DeleteEdgeBySourceTarget {
    #[serde(rename = "_source")]
    pub source: String,
    #[serde(rename = "_target")]
    pub target: String,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct GetEdges {
    pub item: String,
    pub direction: EdgeDirection,
    pub expand_items: bool,
}
#[derive(Serialize, Deserialize, Debug, PartialEq, Eq)]
pub enum EdgeDirection {
    Outgoing,
    Incoming,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct EdgeSearch {}

#[derive(Serialize, Deserialize, Default, Debug)]
#[serde(rename_all = "camelCase")]
pub struct Search {
    pub id: Option<String>,
    pub _type: Option<String>,
    #[serde(rename = "dateServerModified>=")]
    pub date_server_modified_gte: Option<i64>,
    #[serde(rename = "dateServerModified<")]
    pub date_server_modified_lt: Option<i64>,
    pub deleted: Option<bool>,
    #[serde(rename = "_sortOrder", default = "default_api_sort_order")]
    pub sort_order: SortOrder,
    #[serde(rename = "_limit", default = "default_api_limit")]
    pub limit: u64,
    #[serde(rename = "_returnProperties", default)]
    pub return_properties: Vec<String>,
    #[serde(rename = "[[edges]]")]
    pub forward_edges: Option<EdgeSearch>,
    #[serde(rename = "~[[edges]]")]
    pub backward_edges: Option<EdgeSearch>,
    #[serde(flatten)]
    pub other_properties: HashMap<String, Value>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct SearchV2 {
    pub id: Option<String>,
    pub _type: Option<String>,
    #[serde(rename = "dateServerModified>=")]
    pub date_server_modified_gte: Option<i64>,
    #[serde(rename = "dateServerModified<")]
    pub date_server_modified_lt: Option<i64>,
    pub deleted: Option<bool>,
    #[serde(rename = "_sortOrder", default = "default_api_sort_order")]
    pub sort_order: SortOrder,
    #[serde(rename = "_limit", default = "default_api_limit")]
    pub limit: u64,
    #[serde(rename = "_returnProperties", default)]
    pub return_properties: Vec<String>,
    #[serde(rename = "_returnEdges", default)]
    pub return_edges: Vec<String>,
    #[serde(flatten)]
    pub other_properties: HashMap<String, Value>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct SendEmail {
    pub to: String,
    pub subject: String,
    pub body: String,
}

#[derive(Deserialize, Debug)]
pub struct TraceRequest {
    pub secret: String,
    pub filter: String,
}

#[derive(Debug, Clone)]
pub enum PluginIdentifier {
    Id(String),
    Alias(String),
}

impl AsRef<PluginIdentifier> for PluginIdentifier {
    fn as_ref(&self) -> &PluginIdentifier {
        self
    }
}

/// Deserialize the string that can represent a hexstring - return PluginIdentifier::Id,
/// or a human friendly name - return PluginIdentifier::Alias
impl<'de> Deserialize<'de> for PluginIdentifier {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        struct PluginIdVisitor;
        impl Visitor<'_> for PluginIdVisitor {
            type Value = PluginIdentifier;
            fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                formatter.write_str("UUID or plugin alias name")
            }

            fn visit_str<E>(self, value: &str) -> std::result::Result<PluginIdentifier, E>
            where
                E: de::Error,
            {
                let hexstring = value.replace('-', "");
                if hex::decode(hexstring).is_ok() {
                    Ok(PluginIdentifier::Id(value.to_string()))
                } else {
                    Ok(PluginIdentifier::Alias(value.to_string()))
                }
            }
        }

        deserializer.deserialize_str(PluginIdVisitor)
    }
}
#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginStatusReq {
    pub plugins: Vec<PluginIdentifier>,
}

#[derive(Serialize, Deserialize, Debug, PartialEq, Eq)]
#[serde(rename_all = "camelCase")]
pub enum PluginStatus {
    Running(Value),
    Unreachable,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct PluginStatusMetadata {
    pub status: PluginStatus,

    #[serde(skip_serializing_if = "Option::is_none")]
    pub alias: Option<String>,
}
#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginStatusRes {
    pub plugins: HashMap<String, PluginStatusMetadata>,
    pub shared_plugins: HashMap<String, PluginStatusMetadata>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct CreateUserReq {
    pub owner_key: PodOwner,
    pub database_key: String,
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginGetApiReq {
    pub id: PluginIdentifier,
}

type Endpoint = String;
pub type PluginGetApiRes = HashMap<Endpoint, PluginEndpointMetadata>;

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginEndpointMetadata {
    method: String,
    args: HashMap<String, String>,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "UPPERCASE")]
pub enum PluginCallMethod {
    Get,
    Post,
    Put,
    Delete,
    Patch,
}

impl From<&PluginCallMethod> for Method {
    fn from(plugin_call_method: &PluginCallMethod) -> Self {
        match plugin_call_method {
            PluginCallMethod::Get => Method::GET,
            PluginCallMethod::Post => Method::POST,
            PluginCallMethod::Put => Method::PUT,
            PluginCallMethod::Delete => Method::DELETE,
            PluginCallMethod::Patch => Method::PATCH,
        }
    }
}

#[derive(Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct PluginCallApiReq {
    pub id: PluginIdentifier,
    pub endpoint: String,
    pub method: PluginCallMethod,
    pub query: Option<Value>,
    pub json_body: Option<Value>,
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct PluginCallApiRes {
    pub status_code: u16,
    pub body: Value,
}

/// Pretty print body, using Display, not Debug to have the output clean
impl Debug for PluginCallApiRes {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("PluginCallApiRes")
            .field("status_code", &self.status_code)
            .field("body", &format_args!("{:#}", self.body))
            .finish()
    }
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct PluginRunReq {
    pub id: String,
    pub container_image: String,
    pub plugin_module: String,
    pub plugin_name: String,
    pub status: Option<String>,
    pub target_item_id: String,
}

#[derive(Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct PluginAttachReq {
    pub id: PluginIdentifier,
    pub webserver_port: u32,
    pub webserver_url: String,
}

#[derive(Serialize, Deserialize, Debug, Clone, Default)]
pub enum SortOrder {
    /// Ascending
    #[default]
    Asc,
    /// Descending
    Desc,
}

impl std::fmt::Display for SortOrder {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        std::fmt::Debug::fmt(self, f)
    }
}

fn default_api_sort_order() -> SortOrder {
    SortOrder::Asc
}
fn default_api_limit() -> u64 {
    u64::MAX
}

//
// Files API:
//

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all = "camelCase")]
pub struct GetFile {
    pub sha256: String,
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_serialization_examples() {
        let client_auth = ClientAuth {
            database_key: "0000".to_string(),
        };
        let client_auth_ser = serde_json::to_string(&client_auth).unwrap();
        assert_eq!(client_auth_ser, r#"{"databaseKey":"0000"}"#);
        let auth_key = AuthKey::ClientAuth(client_auth);
        let auth_key_ser = serde_json::to_string(&auth_key).unwrap();
        assert_eq!(
            auth_key_ser,
            r#"{"type":"ClientAuth","databaseKey":"0000"}"#
        );
    }
}
