use rusqlite::TransactionBehavior;
use std::ops::Deref;
use std::rc::Rc;
use tracing::error;

use crate::database_pool::PooledConnection;
use crate::error::{ErrorContext, Result};
use crate::internal_error;

pub trait AsyncTxHandler {
    type Output;
    type Future: core::future::Future<Output = Self::Output>;

    fn call(self, tx: AsyncTx) -> Self::Future;
}

impl<Func, Fut> AsyncTxHandler for Func
where
    Func: FnOnce(AsyncTx) -> Fut,
    Fut: core::future::Future,
{
    type Output = Fut::Output;
    type Future = Fut;

    #[inline]
    fn call(self, tx: AsyncTx) -> Self::Future {
        (self)(tx)
    }
}

pub struct AsyncConnection {
    pub inner: Rc<PooledConnection>,
}

impl AsyncConnection {
    pub fn new(conn: PooledConnection) -> Self {
        AsyncConnection {
            inner: Rc::new(conn),
        }
    }

    pub fn from_other(conn: &AsyncConnection) -> Self {
        AsyncConnection {
            inner: conn.inner.clone(),
        }
    }

    /// Starts a transaction and gathers write lock immediately,
    /// might wait `busy_timeout` milliseconds if other write
    /// transaction is on the way.
    /// In SQLite there can be exactly one writer at a time, so write TXes
    /// has to be serialized in order to avoid deadlocks
    pub async fn in_write_transaction<F, T>(&mut self, f: F) -> Result<T>
    where
        F: AsyncTxHandler<Output = Result<T>>,
    {
        // EXCLUSIVE is similar to IMMEDIATE in that a write transaction is started immediately.
        // EXCLUSIVE and IMMEDIATE are the same in WAL mode, but in other journaling modes,
        // EXCLUSIVE prevents other database connections from reading the database while
        // the transaction is underway.
        // that prevents writer starvation in other journaling modes
        let res = self.in_transaction(f, TransactionBehavior::Exclusive).await;

        if let Err(e) = &res {
            error!("Error {e} in async write transaction");
        }

        res
    }

    /// Executes F inside transaction with default "deferred" behavior, on current thread.
    /// Commits the transaction in case of success, rollbacks otherwise.
    /// Returns the execution result.
    /// There can be many read transactions on the fly, they do not block on each other.
    /// Make sure all operations inside TX are read only, if they are not, that might
    /// cause a deadlock and fail the TX immediately with SQLITE_BUSY error
    pub async fn in_read_transaction<F, T>(&mut self, f: F) -> Result<T>
    where
        F: AsyncTxHandler<Output = Result<T>>,
    {
        let res = self.in_transaction(f, TransactionBehavior::Deferred).await;

        if let Err(e) = &res {
            error!("Error {e} in async read transaction");
        }

        res
    }

    // &mut self for exclusive access to inner connection
    pub async fn in_transaction<F, T>(&mut self, f: F, behavior: TransactionBehavior) -> Result<T>
    where
        F: AsyncTxHandler<Output = Result<T>>,
    {
        // Hide usage of tx in here, to disallow misuse by the caller
        let tx = AsyncTx::new(self.inner.clone(), behavior)?;
        let res = f.call(tx.clone()).await;

        if res.is_ok() {
            tx.commit()
                .context_str("while committing the transaction")?;
        } else {
            let _ = tx.rollback();
        }

        res
    }
}

impl Deref for AsyncConnection {
    type Target = PooledConnection;

    fn deref(&self) -> &Self::Target {
        self.inner.as_ref()
    }
}

#[derive(Clone)]
pub struct AsyncTx {
    pub inner: Rc<PooledConnection>,
}

impl AsyncTx {
    fn new(conn: Rc<PooledConnection>, behavior: TransactionBehavior) -> Result<Self> {
        let query = match behavior {
            TransactionBehavior::Deferred => "BEGIN DEFERRED",
            TransactionBehavior::Immediate => "BEGIN IMMEDIATE",
            TransactionBehavior::Exclusive => "BEGIN EXCLUSIVE",
            _ => return Err(internal_error!("Unsupported TX behavior to set")),
        };
        conn.execute_batch(query)?;

        Ok(Self { inner: conn })
    }

    fn commit(&self) -> Result<()> {
        self.inner.execute_batch("COMMIT")?;
        Ok(())
    }

    fn rollback(&self) -> Result<()> {
        self.inner.execute_batch("ROLLBACK")?;
        Ok(())
    }
}

impl Deref for AsyncTx {
    type Target = PooledConnection;

    // Yes it's famous Deref antipattern,
    // as an excuse, rusqlite does this too
    fn deref(&self) -> &Self::Target {
        &self.inner
    }
}

// ----------------------------------------------------
// Version 1, works somewhat with regular functions, fails with closures - there is no way
// to tell lifetimes in closure
// trait MyHandler<'conn> {
//     type Output;
//     type Future: core::future::Future<Output = Self::Output>;

//     fn call(self, tx: Rc<Tx<'conn>>) -> Self::Future;
// }

// impl<'conn, Func, Fut> MyHandler<'conn> for Func
// where
//     Func: FnOnce(Rc<Tx<'conn>>) -> Fut,
//     Fut: core::future::Future,
// {
//     type Output = Fut::Output;
//     type Future = Fut;

//     #[inline]
//     fn call(self, tx: Rc<Tx<'conn>>) -> Self::Future {
//         (self)(tx)
//     }
// }

// async fn in_tx<'conn, F, T>(conn: &'conn mut PooledConnection, f: F) -> Result<T>
// where
//     F: MyHandler<'conn, Output = Result<T>>,
// {
//     let tx = Rc::new(conn.transaction()?);

//     let a = f.call(tx.clone()).await?;

//     let tx = Rc::try_unwrap(tx).expect("Failed to unwrap transaction");
//     let _ = tx.commit();

//     Ok(a)
// }

// async fn some_async_fn(tx: Rc<Tx<'_>>) -> Result<()> {
//     tx.prepare_cached("asdf")?;
//     Ok(())
// }
