use clap::Parser;
use lazy_static::lazy_static;
use std::net::IpAddr;

#[derive(Parser, Debug, Clone)]
#[command(
    name = "Pod, the open-source backend for Memri project.",
    // setting = AppSettings::DeriveDisplayOrder,
    // setting = AppSettings::UnifiedHelpMessage,
    version=std::env!("GIT_DESCRIBE")
)]
pub struct CliOptions {
    /// Port to listen to.
    #[arg(short, long, default_value = "3030")]
    pub port: u16,

    /// Comma-separated list of Pod owners (hex-encoded hashes of public keys).
    /// See `docs/HTTP_API.md#api-authentication-credentials` on the format of the owner keys.
    ///
    /// Only those owners are allowed to call Pod endpoints.
    ///
    /// Each Pod owner has its own database and files directory,
    /// the owners do not intersect data-wise.
    /// Pod does not store any data on owners in any external databases.
    ///
    /// A magic value of "ANY" will allow any owner to connect to the Pod.
    #[arg(
        short = 'o',
        long,
        name = "OWNERS",
        required = true,
        env = "POD_OWNER_HASHES"
    )]
    pub owners: String,

    /// If specified, all Plugin containers will be started using kubernetes (`kubectl`).
    /// Otherwise and by default, docker containers are used.
    #[arg(long)]
    pub use_kubernetes: bool,

    /// Sets maximum number of plugins that can be run at once per one POD owner.
    #[arg(
        long,
        name = "K8S_LIMIT",
        env = "POD_KUBERNETES_PLUGIN_LIMIT_PER_OWNER",
        default_value = "5"
    )]
    pub kubernetes_plugin_limit_per_owner: usize,

    /// Ignores limits on kubernetes containers, it takes precedence to 'kubernetes-plugin-limit-per-owner' flag
    #[arg(long, env = "POD_KUBERNETES_IGNORE_LIMITS")]
    pub kubernetes_ignore_limits: bool,

    /// Set the callback address for plugins launched from within Pod.
    /// This should be the Pod-s address as seen by external plugins.
    /// It defaults to "pod:3030" if Pod is inside docker,
    /// or "localhost:3030" on Linux,
    /// or "host.docker.internal:3030" on other operating systems.
    #[arg(
        short = 's',
        long,
        name = "ADDRESS",
        env = "POD_PLUGINS_CALLBACK_ADDRESS"
    )]
    pub plugins_callback_address: Option<String>,

    /// Set the origin range for PodClients created in the Plugin.
    /// Request will fail if pod callback url is outside the allowed origins.
    #[arg(
        long,
        name = "ALLOWED_ORIGINS",
        env = "POD_ALLOWED_ORIGINS",
        value_delimiter(','),
        default_value = "*"
    )]
    pub pod_allowed_origins: Vec<String>,

    /// Plugin's public address, as seen by Memri clients and the external network.
    /// This is passed to the Plugins so that they can know (and advertise) where they are.
    ///
    /// If not specified, Pod will assume the plugin runs on `localhost`.
    #[arg(long, env)]
    pub plugins_public_domain: Option<String>,

    /// Docker network to use when running plugins, e.g. `docker run --network=XXX ...`
    /// If not set, "host" will be used, which means that started plugins
    /// will share the network with the host system.
    /// If Pod itself is running inside docker, please run both Pod and plugins
    /// in identical network that will then not be shared with the host system
    /// (this is covered in docker-compose.yml by default).
    #[arg(
        long,
        name = "PLUGINS_DOCKER_NETWORK",
        env = "POD_PLUGINS_DOCKER_NETWORK"
    )]
    pub plugins_docker_network: Option<String>,

    /// File to read https public certificate from.
    #[arg(
        short = 'c',
        long,
        default_value = "./data/certs/pod.crt",
        name = "CERTIFICATE_FILE"
    )]
    pub tls_pub_crt: String,

    /// File to read https private key from.
    #[arg(
        short = 'k',
        long,
        default_value = "./data/certs/pod.key",
        name = "KEY_FILE"
    )]
    pub tls_priv_key: String,

    /// Override a certain Plugin container to run a script instead.
    /// The value specified should be of the form `image_name=/path/to/your/local/script`.
    /// Whenever Pod is supposed to run a container with this name, it will run
    /// the specified script instead. This argument can be useful during plugin development,
    ///
    /// for example: `my_plugin_container_name=/home/john/memri/my_plugin/script.py`
    /// or `my_plugin_container_name=C:\\memri\\my_plugin\\plugin.exe`
    /// or `my_plugin_container_name=/Users/john/memri/my_plugin/script.py`
    ///
    /// Note that running scripts instead of containers is not secure. A container is limited
    /// in its access to the filesystem, but running a script is only secure if you know and trust
    /// the script.
    #[arg(long, value_parser(parse_key_val), action(clap::ArgAction::Append))]
    pub insecure_plugin_script: Vec<(String, String)>,

    /// Do not use https when starting the server, instead run on http://127.0.0.1.
    /// Running on loopback interface (127.0.0.1) means that only apps
    /// from within the same computer will be able to access Pod.
    /// This option might be used during development as an alternative to self-signed certificates.
    #[arg(short = 't', long, env = "POD_NON_TLS")]
    pub non_tls: bool,

    /// Unsafe version of --non-tls that runs on a public network, e.g. "http://0.0.0.0".
    /// This option will force Pod to not use https,
    /// and instead run http on the provided network interface.
    /// WARNING: This is heavily discouraged as an intermediary
    /// (even your router on a local network)
    /// could spoof the traffic sent to the server and do a MiTM attack.
    /// Please consider running Pod on a non-public network (--non-tls),
    /// or use Pod with https encryption.
    #[arg(long, name = "NETWORK_INTERFACE", env = "POD_INSECURE_NON_TLS")]
    pub insecure_non_tls: Option<IpAddr>,

    /// Run server as a "SharedServer". See `/docs/SharedServer.md` documentation
    /// for details on what it is, and how it works.
    #[arg(long, env)]
    pub shared_server: bool,

    /// SMTP relay server to use (advanced option).
    #[arg(long, env)]
    pub email_smtp_relay: Option<String>,

    /// SMTP relay server port to use (advanced option).
    #[arg(long, default_value = "465", env)]
    pub email_smtp_port: u16,

    /// SMTP relay server user (advanced option).
    #[arg(long, env)]
    pub email_smtp_user: Option<String>,

    /// SMTP relay server password (advanced option).
    #[arg(long, env)]
    pub email_smtp_password: Option<String>,

    /// DB credentials used by the POD to store non volatile information
    #[arg(long, env = "POD_OWNER_KEY_FOR_POD", requires = "db_key_for_pod")]
    pub owner_key_for_pod: Option<String>,

    /// DB credentials used by the POD to store non volatile information
    #[arg(long, env = "POD_DB_KEY_FOR_POD", requires = "owner_key_for_pod")]
    pub db_key_for_pod: Option<String>,

    #[arg(long, env = "POD_SHARED_PLUGINS", action(clap::ArgAction::Append))]
    pub shared_plugins: Vec<String>,

    /// Fixed size of opened connections per database,
    /// if not set, calculation is done using # of cores.
    #[arg(long, env = "POD_OPENED_CONNECTIONS")]
    pub opened_connections: Option<u32>,
}

fn parse_key_val(s: &str) -> Result<(String, String), String> {
    let pos = s
        .find('=')
        .ok_or_else(|| format!("invalid KEY=value: no `=` found in `{s}`"))?;
    Ok((
        s[..pos]
            .parse()
            .map_err(|e| format!("failed to parse KEY {e}"))?,
        s[pos + 1..]
            .parse()
            .map_err(|e| format!("failed to parse value {e}"))?,
    ))
}

lazy_static! {
    pub static ref PARSED: CliOptions = get_cli_options();
}

#[cfg(not(test))]
fn get_cli_options() -> CliOptions {
    CliOptions::parse()
}

#[cfg(test)]
fn get_cli_options() -> CliOptions {
    tests::test_cli()
}

#[cfg(test)]
pub mod tests {
    use super::CliOptions;
    use std::net::{IpAddr, Ipv4Addr};

    /// Example test CLI. Purely for convenience,
    /// you can instantiate your own / unrelated ones as well.
    pub fn test_cli() -> CliOptions {
        CliOptions {
            port: 3030,
            owners: "ANY".to_string(),
            use_kubernetes: false,
            kubernetes_plugin_limit_per_owner: 5,
            kubernetes_ignore_limits: false,
            plugins_callback_address: None,
            pod_allowed_origins: vec!["*".to_string()],
            plugins_public_domain: None,
            plugins_docker_network: None,
            insecure_plugin_script: Vec::new(),
            tls_pub_crt: "".to_string(),
            tls_priv_key: "".to_string(),
            non_tls: true,
            insecure_non_tls: Some(IpAddr::V4(Ipv4Addr::new(0, 0, 0, 0))),
            shared_server: false,
            email_smtp_relay: None,
            email_smtp_port: 465,
            email_smtp_user: None,
            email_smtp_password: None,
            owner_key_for_pod: None,
            db_key_for_pod: None,
            shared_plugins: Vec::new(),
            opened_connections: None,
        }
    }
}
