use crate::{
    api_model::SortOrder,
    async_db_connection::AsyncTx as Tx,
    bad_request, database_utils,
    database_utils::check_item_has_all_properties,
    db_model::ItemBase,
    error::{ErrorContext, Result},
    graphql_utils::Filter,
    internal_error,
    schema::{Schema, SchemaPropertyType},
};

use rusqlite::{params, types::ToSqlOutput, Row, Rows};
use serde_json::Value;

use std::collections::HashMap;
use std::fmt::Write;
use tracing::{debug, warn};

pub type Rowid = i64;
pub type DbTime = i64;

#[allow(clippy::too_many_arguments)]
pub fn insert_item_base(
    tx: &Tx,
    id: &str,
    _type: &str,
    date_created_millis: DbTime,
    date_modified_millis: DbTime,
    date_server_modified_millis: DbTime,
    deleted: bool,
) -> Result<Rowid> {
    let mut stmt = tx
        .prepare_cached(
            "INSERT INTO items (\
            id, \
            type, \
            dateCreated, \
            dateModified, \
            dateServerModified, \
            deleted \
        ) VALUES (?, ?, ?, ?, ?, ?);",
        )
        .context_str("Failed to prepare/compile INSERT statement")?;
    stmt.insert(params![
        id,
        _type,
        date_created_millis,
        date_modified_millis,
        date_server_modified_millis,
        deleted,
    ])
    .context_str("Failed to execute insert_item with parameters")
}

pub fn get_item_base(tx: &Tx, rowid: Rowid) -> Result<Option<ItemBase>> {
    let database_search = DatabaseSearch {
        rowid: Some(rowid),
        id: None,
        _type: None,
        date_server_modified_gte: None,
        date_server_modified_lt: None,
        deleted: None,
        sort_order: SortOrder::Asc,
        _limit: 1,
    };
    let item = search_items(tx, &database_search)?.into_iter().next();
    Ok(item)
}

pub fn get_item_rowid(tx: &Tx, id: &str) -> Result<Option<Rowid>> {
    let mut stmt = tx.prepare_cached("SELECT rowid FROM items WHERE id = ?;")?;
    let mut rows = stmt.query_map(params![id], |row| row.get(0))?;
    if let Some(row) = rows.next() {
        let rowid: i64 = row?;
        Ok(Some(rowid))
    } else {
        Ok(None)
    }
}

#[derive(Debug)]
pub struct DatabaseSearch<'a> {
    pub rowid: Option<Rowid>,
    pub id: Option<&'a str>,
    pub _type: Option<&'a str>,
    pub date_server_modified_gte: Option<DbTime>,
    pub date_server_modified_lt: Option<DbTime>,
    pub deleted: Option<bool>,
    pub sort_order: SortOrder,
    pub _limit: u64,
}

impl Default for DatabaseSearch<'_> {
    fn default() -> Self {
        Self {
            _limit: u64::MAX,

            rowid: Default::default(),
            id: Default::default(),
            _type: Default::default(),
            date_server_modified_gte: Default::default(),
            date_server_modified_lt: Default::default(),
            deleted: Default::default(),
            sort_order: Default::default(),
        }
    }
}

fn parse_item_base(row: &Row) -> Result<ItemBase> {
    Ok(ItemBase {
        rowid: row.get(0)?,
        id: row.get(1)?,
        _type: row.get(2)?,
        date_created: row.get(3)?,
        date_modified: row.get(4)?,
        date_server_modified: row.get(5)?,
        deleted: row.get(6)?,
    })
}

#[allow(clippy::comparison_chain)]
pub fn search_items(tx: &Tx, query: &DatabaseSearch) -> Result<Vec<ItemBase>> {
    search_items_with_properties(tx, query, None, None)
}

#[allow(clippy::comparison_chain)]
pub fn search_items_with_properties(
    tx: &Tx,
    query: &DatabaseSearch,
    schema: Option<&Schema>,
    other_properties: Option<&HashMap<String, Value>>,
) -> Result<Vec<ItemBase>> {
    let mut sql_query = "\
        SELECT \
            rowid, \
            id, \
            type, \
            dateCreated, \
            dateModified, \
            dateServerModified, \
            deleted \
        FROM \
            items \
        WHERE "
        .to_string();
    let mut params_vec: Vec<ToSqlOutput> = Vec::new();
    if let Some(r) = query.rowid {
        add_sql_param(&mut sql_query, "rowid", &Comparison::Equals);
        params_vec.push(r.into());
    }
    if let Some(id) = &query.id {
        add_sql_param(&mut sql_query, "id", &Comparison::Equals);
        params_vec.push((*id).into());
    }
    if let Some(typ) = &query._type {
        add_sql_param(&mut sql_query, "type", &Comparison::Equals);
        params_vec.push((*typ).into());
    }
    if let Some(dt) = query.date_server_modified_gte {
        add_sql_param(
            &mut sql_query,
            "dateServerModified",
            &Comparison::GreaterOrEquals,
        );
        params_vec.push(dt.into());
    }
    if let Some(dt) = query.date_server_modified_lt {
        add_sql_param(&mut sql_query, "dateServerModified", &Comparison::LessThan);
        params_vec.push(dt.into());
    }
    if let Some(deleted) = query.deleted {
        add_sql_param(&mut sql_query, "deleted", &Comparison::Equals);
        params_vec.push(deleted.into());
    }
    sql_query.push_str("1 "); // older sqlite versions do not support `true`

    let _ = write!(
        &mut sql_query,
        "ORDER BY dateServerModified {}",
        query.sort_order
    );
    sql_query.push(';');
    debug!("Executing search SQL: {}", sql_query);

    let mut stmt = tx
        .prepare_cached(&sql_query)
        .context(|| format!("SQL query: {sql_query}"))?;

    for (index, param) in params_vec.into_iter().enumerate() {
        // SQLite parameters are 1-based, not 0-based, so we need to add 1 to the index.
        stmt.raw_bind_parameter(index + 1, param)?;
    }
    let mut rows: Rows = stmt.raw_query();

    let mut result = Vec::new();
    let mut num_left = query._limit;
    let mut last_date: Option<DbTime> = None;

    while let Some(row) = rows.next()? {
        let item = parse_item_base(row)?;
        let rowid = item.rowid;

        if let (Some(_schema), Some(_other_properties)) = (schema, other_properties) {
            if !check_item_has_all_properties(tx, _schema, rowid, _other_properties)? {
                continue;
            }
        }
        if num_left > 1 {
            num_left -= 1;
        } else if num_left == 1 {
            num_left = 0;
            last_date = Some(item.date_server_modified);
        } else if let Some(last_date) = last_date {
            if last_date != item.date_server_modified {
                break;
            }
        } else {
            break;
        }

        result.push(item);
    }

    Ok(result)
}

/// Search for items that have a certain property equal to certain value
pub fn search_strings(tx: &Tx, property_name: &str, value: &str) -> Result<Vec<Rowid>> {
    let mut stmt = tx.prepare_cached("SELECT item FROM strings WHERE name = ? AND value = ?;")?;
    let mut rows = stmt.query(params![property_name, value])?;
    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        let item: Rowid = row.get(0)?;
        result.push(item);
    }
    Ok(result)
}

pub fn get_strings_for_item(tx: &Tx, item_rowid: Rowid) -> Result<HashMap<String, String>> {
    let mut stmt = tx.prepare_cached("SELECT name, value FROM strings WHERE item = ?;")?;
    let mut rows = stmt.query(params![item_rowid])?;
    let mut result = HashMap::new();
    while let Some(row) = rows.next()? {
        result.insert(row.get(0)?, row.get(1)?);
    }
    Ok(result)
}
pub struct StringsNameValue {
    pub name: String,
    pub value: String,
}
pub fn get_strings_records_for_item(tx: &Tx, item_rowid: Rowid) -> Result<Vec<StringsNameValue>> {
    let mut stmt = tx.prepare_cached("SELECT name, value FROM strings WHERE item = ?;")?;
    let mut rows = stmt.query(params![item_rowid])?;
    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        result.push(StringsNameValue {
            name: row.get(0)?,
            value: row.get(1)?,
        });
    }
    Ok(result)
}

pub fn get_integers_for_item(tx: &Tx, item_rowid: Rowid) -> Result<HashMap<String, i64>> {
    let mut stmt = tx.prepare_cached("SELECT name, value FROM integers WHERE item = ?;")?;
    let mut rows = stmt.query(params![item_rowid])?;
    let mut result = HashMap::new();
    while let Some(row) = rows.next()? {
        result.insert(row.get(0)?, row.get(1)?);
    }
    Ok(result)
}
pub struct IntegersNameValue {
    pub name: String,
    pub value: i64,
}
pub fn get_integers_records_for_item(tx: &Tx, item_rowid: Rowid) -> Result<Vec<IntegersNameValue>> {
    let mut stmt = tx.prepare_cached("SELECT name, value FROM integers WHERE item = ?;")?;
    let mut rows = stmt.query(params![item_rowid])?;
    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        result.push(IntegersNameValue {
            name: row.get(0)?,
            value: row.get(1)?,
        });
    }
    Ok(result)
}

pub fn get_reals_for_item(tx: &Tx, item_rowid: Rowid) -> Result<HashMap<String, f64>> {
    let mut stmt = tx.prepare_cached("SELECT name, value FROM reals WHERE item = ?;")?;
    let mut rows = stmt.query(params![item_rowid])?;
    let mut result = HashMap::new();
    while let Some(row) = rows.next()? {
        result.insert(row.get(0)?, row.get(1)?);
    }
    Ok(result)
}
pub struct RealsNameValue {
    pub name: String,
    pub value: f64,
}
pub fn get_reals_records_for_item(tx: &Tx, item_rowid: Rowid) -> Result<Vec<RealsNameValue>> {
    let mut stmt = tx.prepare_cached("SELECT name, value FROM reals WHERE item = ?;")?;
    let mut rows = stmt.query(params![item_rowid])?;
    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        result.push(RealsNameValue {
            name: row.get(0)?,
            value: row.get(1)?,
        });
    }
    Ok(result)
}

pub struct GQLSearchArgs {
    pub _type: Option<String>,
    pub properties: Vec<String>,
    pub rowids: Option<Vec<Rowid>>, // TODO rowids should be in filters
    pub sort_property: String,
    pub sort_order: SortOrder,
    pub limit: Option<u32>,
    pub offset: Option<u32>,
    pub filters: Option<Filter>,
}

pub struct GQLPropertiesResult {
    pub rowid: Rowid,
    pub properties: serde_json::Map<String, Value>,
}

pub struct EdgeResult {
    pub self_id: Rowid,
    pub source: Rowid,
    pub name: String,
    pub target: Rowid,
}

pub fn get_edges(
    tx: &Tx,
    item_id: Rowid,
    edge_name: Option<&String>,
    reverse: bool,
) -> Result<Vec<EdgeResult>> {
    let mut query = if reverse {
        "SELECT self, target, name, source FROM edges WHERE target = ?".to_owned()
    } else {
        "SELECT self, source, name, target FROM edges WHERE source = ?".to_owned()
    };
    if let Some(e) = edge_name {
        query.push_str(format!(" AND name = '{e}'").as_str());
    }
    let mut stmt = tx.prepare_cached(&query)?;
    let mut rows = stmt.query(params![item_id])?;

    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        let self_id = row.get(0)?;
        let source = row.get(1)?;
        let mut name = row.get(2)?;
        if reverse {
            name = format!("{}{:?}", "~", name);
        }
        let target = row.get(3)?;
        result.push(EdgeResult {
            self_id,
            source,
            name,
            target,
        })
    }
    Ok(result)
}

pub fn search_gql_properties(
    tx: &Tx,
    query: &GQLSearchArgs,
    schema: &Schema,
) -> Result<Vec<GQLPropertiesResult>> {
    // Returns results from a GQLSearchArgs, filtered, sorted, paginated, no edges.
    let sql_query = query_from_gqlsearch(query, schema)?;
    let mut stmt = tx
        .prepare_cached(&sql_query)
        .context(|| format!("SQL query: {sql_query}"))?;

    let column_names = stmt
        .column_names()
        .into_iter()
        .map(|s| s.to_owned())
        .collect::<Vec<_>>();
    let mut rows: Rows = stmt.query(params![])?;

    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        let res = row_to_gql_properties_result(row, &column_names, schema)?;
        result.push(res);
    }
    Ok(result)
}

fn row_to_gql_properties_result(
    row: &Row,
    column_names: &[String],
    schema: &Schema,
) -> Result<GQLPropertiesResult> {
    let mut json = serde_json::Map::new();
    for (i, name) in column_names.iter().enumerate() {
        if name != "rowid" {
            let valueref = row.get_ref(i)?;
            match schema.property(name) {
                Some(property_type) => {
                    let value = database_utils::valueref_to_json_value(valueref, property_type)?;
                    json.insert(name.to_string(), value);
                }
                None => warn!("Ignoring item property {}: does not exist in schema", name,),
            }
        }
    }
    let result = GQLPropertiesResult {
        rowid: row.get("rowid")?,
        properties: json,
    };
    Ok(result)
}

fn query_from_gqlsearch(gql_search: &GQLSearchArgs, schema: &Schema) -> Result<String> {
    let properties = &gql_search.properties;
    let property_queries = build_sql_property_queries(properties, schema)?;
    let filters = build_filters(gql_search)?;

    let sort_property = get_sort_property(gql_search, schema)?;

    let mut result = format!(
        "SELECT\n\
        rowid,\n\
        {property_queries}\n\
        FROM items\n\
        WHERE {filters}\n\
        ORDER BY {sort_property} {sort_order}, rowid",
        sort_order = &gql_search.sort_order,
    );

    if let Some(limit) = gql_search.limit {
        result.push_str(format!("\nLIMIT {limit}").as_str());
    }
    if let Some(offset) = gql_search.offset {
        result.push_str(format!("\nOFFSET {offset}").as_str());
    }
    Ok(result)
}

fn get_sort_property(gql_search: &GQLSearchArgs, schema: &Schema) -> Result<String> {
    let sort_property = &gql_search.sort_property;
    if gql_search.properties.contains(sort_property) | schema.is_base_property(sort_property) {
        Ok(sort_property.to_owned())
    } else {
        Err(bad_request! {"Could not sort on `{sort_property}`, property not in query"})
    }
}

fn build_filters(gql_search: &GQLSearchArgs) -> Result<String> {
    let mut result = Vec::new();
    if let Some(t) = &gql_search._type {
        result.push(format!("type = '{t}'"));
    }
    if let Some(r) = &gql_search.rowids {
        let ids_vec: Vec<String> = r.iter().map(|&id| id.to_string()).collect();
        let ids: String = ids_vec.join(", ");
        result.push(format!("rowid IN ({ids})"))
    }

    let res = result.join("\nAND ");
    let filters = if let Some(f) = gql_search.filters.to_owned() {
        let filter_str = f.to_sql_string();
        format!("{res} AND {filter_str}")
    } else {
        res
    };

    Ok(filters)
}

fn build_sql_property_queries(properties: &[String], schema: &Schema) -> Result<String> {
    let mut result: Vec<String> = properties
        .iter()
        .filter_map(|p| build_sql_property_query(p, schema).unwrap_or(None))
        .collect();
    result.push("items.*".to_string());
    Ok(result.join(",\n"))
}

fn build_sql_property_query(property_name: &String, schema: &Schema) -> Result<Option<String>> {
    let property_query: Option<String> = if schema.is_base_property(property_name) {
        None
    } else {
        match schema.property(property_name) {
            Some(property_type) => {
                let table_name = database_utils::get_table_from_type(property_type);
                let q = format!(
                    "(SELECT {table_name}.value from {table_name} \
                        WHERE {table_name}.item = items.rowid \
                        AND {table_name}.name = '{property_name}' LIMIT 1) \
                        AS {property_name}"
                );
                Some(q)
            }
            None => None,
        }
    };
    Ok(property_query)
}

pub fn check_integer_exists(tx: &Tx, item_rowid: Rowid, name: &str, value: i64) -> Result<bool> {
    let mut stmt =
        tx.prepare_cached("SELECT 1 FROM integers WHERE item = ? AND name = ? AND value = ?;")?;
    Ok(stmt.exists(params![item_rowid, name, value])?)
}

pub fn check_string_exists(tx: &Tx, item_rowid: Rowid, name: &str, value: &str) -> Result<bool> {
    let mut stmt =
        tx.prepare_cached("SELECT 1 FROM strings WHERE item = ? AND name = ? AND value = ?;")?;
    Ok(stmt.exists(params![item_rowid, name, value])?)
}

pub fn check_real_exists(tx: &Tx, item_rowid: Rowid, name: &str, value: f64) -> Result<bool> {
    let mut stmt =
        tx.prepare_cached("SELECT 1 FROM reals WHERE item = ? AND name = ? AND value = ?;")?;
    Ok(stmt.exists(params![item_rowid, name, value])?)
}

pub fn check_integer_exists_by_name(tx: &Tx, item_rowid: Rowid, name: &str) -> Result<bool> {
    let mut stmt = tx.prepare_cached("SELECT 1 FROM integers WHERE item = ? AND name = ?;")?;
    Ok(stmt.exists(params![item_rowid, name])?)
}
pub fn check_string_exists_by_name(tx: &Tx, item_rowid: Rowid, name: &str) -> Result<bool> {
    let mut stmt = tx.prepare_cached("SELECT 1 FROM strings WHERE item = ? AND name = ?;")?;
    Ok(stmt.exists(params![item_rowid, name])?)
}
pub fn check_real_exists_by_name(tx: &Tx, item_rowid: Rowid, name: &str) -> Result<bool> {
    let mut stmt = tx.prepare_cached("SELECT 1 FROM reals WHERE item = ? AND name = ?;")?;
    Ok(stmt.exists(params![item_rowid, name])?)
}

pub fn update_item_date_server_modified(tx: &Tx, rowid: Rowid, date: DbTime) -> Result<()> {
    let sql = "UPDATE items SET dateServerModified = ? WHERE rowid = ?;";
    let mut stmt = tx.prepare_cached(sql)?;
    stmt.execute(params![date, rowid])?;
    Ok(())
}

pub fn update_item_base(
    tx: &Tx,
    rowid: Rowid,
    date_modified: DbTime,
    date_server_modified: DbTime,
    deleted: Option<bool>,
) -> Result<()> {
    let mut sql = "UPDATE items SET dateModified = ?, dateServerModified = ? ".to_string();
    match deleted {
        Some(true) => sql.push_str(", deleted = 1 WHERE rowid = ?;"),
        Some(false) => sql.push_str(", deleted = 0 WHERE rowid = ? AND NOT type = 'Edge';"),
        None => sql.push_str("WHERE rowid = ?;"),
    }
    let mut stmt = tx.prepare_cached(&sql)?;
    let result = stmt.execute(params![date_modified, date_server_modified, rowid])?;
    if result == 1 {
        Ok(())
    } else {
        Err(bad_request! {
            "Could not update item base {rowid}, item not found or incorrect un-deletion",
        })
    }
}

pub fn dangerous_permament_remove_item(tx: &Tx, rowid: Rowid) -> Result<()> {
    let mut stmt = tx.prepare_cached("DELETE FROM integers WHERE item = ?;")?;
    stmt.execute(params![rowid])?;
    let mut stmt = tx.prepare_cached("DELETE FROM reals WHERE item = ?;")?;
    stmt.execute(params![rowid])?;
    let mut stmt = tx.prepare_cached("DELETE FROM strings WHERE item = ?;")?;
    stmt.execute(params![rowid])?;
    let mut stmt = tx.prepare_cached("DELETE FROM items WHERE rowid = ?;")?;
    stmt.execute(params![rowid])?;
    Ok(())
}

pub fn insert_integer(tx: &Tx, item: Rowid, name: &str, value: i64) -> Result<()> {
    let mut stmt = tx.prepare_cached("INSERT INTO integers VALUES(?, ?, ?);")?;
    stmt.execute(params![item, name, value])?;
    Ok(())
}

pub fn insert_real(tx: &Tx, item: Rowid, name: &str, value: f64) -> Result<()> {
    let mut stmt = tx.prepare_cached("INSERT INTO reals VALUES(?, ?, ?);")?;
    stmt.execute(params![item, name, value])?;
    Ok(())
}

pub fn insert_string(tx: &Tx, item: Rowid, name: &str, value: &str) -> Result<()> {
    let mut stmt = tx.prepare_cached("INSERT INTO strings VALUES(?, ?, ?);")?;
    stmt.execute(params![item, name, value])?;
    Ok(())
}

pub fn delete_property(tx: &Tx, item: Rowid, name: &str) -> Result<()> {
    let mut stmt = tx.prepare_cached("DELETE FROM integers WHERE item = ? AND name = ?;")?;
    stmt.execute(params![item, name])?;
    let mut stmt = tx.prepare_cached("DELETE FROM strings WHERE item = ? AND name = ?;")?;
    stmt.execute(params![item, name])?;
    let mut stmt = tx.prepare_cached("DELETE FROM reals WHERE item = ? AND name = ?;")?;
    stmt.execute(params![item, name])?;
    Ok(())
}

pub fn insert_edge(
    tx: &Tx,
    self_rowid: Rowid,
    source: Rowid,
    name: &str,
    target: Rowid,
) -> Result<Rowid> {
    let mut stmt =
        tx.prepare_cached("INSERT INTO edges(self, source, name, target) VALUES(?, ?, ?, ?);")?;
    stmt.execute(params![self_rowid, source, name, target])?;
    Ok(tx.last_insert_rowid())
}

pub fn get_edge_self_by_source_target(
    tx: &Tx,
    source: Rowid,
    target: Rowid,
) -> Result<Option<Rowid>> {
    let mut stmt = tx.prepare_cached("SELECT self FROM edges WHERE source = ? AND target = ?;")?;
    let mut rows = stmt.query_map(params![source, target], |row| row.get(0))?;
    if let Some(row) = rows.next() {
        let rowid: i64 = row?;
        Ok(Some(rowid))
    } else {
        Ok(None)
    }
}

pub fn delete_edge(tx: &Tx, self_rowid: Rowid) -> Result<()> {
    let mut stmt = tx.prepare_cached("DELETE FROM edges WHERE self = ?;")?;
    stmt.execute(params![self_rowid])?;
    Ok(())
}

pub fn delete_edge_by_source_target(tx: &Tx, source: Rowid, target: Rowid) -> Result<()> {
    let mut stmt = tx.prepare_cached("DELETE FROM edges WHERE source = ? AND target = ?;")?;
    stmt.execute(params![source, target])?;
    Ok(())
}

pub struct EdgeBase {
    pub rowid: Rowid,
    pub name: String,
    pub source: Rowid,
    pub target: Rowid,
}

/// Edge pointer, either for outgoing edge or for incoming edge.
/// `item` is essentially either `source` or `target`, depending on the direction.
pub struct EdgePointer {
    pub rowid: Rowid,
    pub name: String,
    pub item: Rowid,
}

pub fn get_self_edge(tx: &Tx, self_rowid: Rowid) -> Result<Option<EdgeBase>> {
    let mut stmt = tx.prepare_cached("SELECT source, name, target FROM edges WHERE self = ?;")?;
    let mut rows = stmt.query(params![self_rowid])?;
    if let Some(row) = rows.next()? {
        Ok(Some(EdgeBase {
            rowid: self_rowid,
            source: row.get(0)?,
            name: row.get(1)?,
            target: row.get(2)?,
        }))
    } else {
        Ok(None)
    }
}

pub fn get_outgoing_edges(tx: &Tx, source: Rowid) -> Result<Vec<EdgePointer>> {
    let mut stmt = tx.prepare_cached("SELECT self, target, name FROM edges WHERE source = ?;")?;
    let mut rows = stmt.query(params![source])?;
    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        let rowid = row.get(0)?;
        let item = row.get(1)?;
        let name = row.get(2)?;
        result.push(EdgePointer { rowid, name, item })
    }
    Ok(result)
}

pub fn get_incoming_edges(tx: &Tx, target: Rowid) -> Result<Vec<EdgePointer>> {
    let mut stmt = tx.prepare_cached("SELECT self, source, name FROM edges WHERE target = ?;")?;
    let mut rows = stmt.query(params![target])?;
    let mut result = Vec::new();
    while let Some(row) = rows.next()? {
        let rowid = row.get(0)?;
        let item = row.get(1)?;
        let name = row.get(2)?;
        result.push(EdgePointer { rowid, name, item })
    }
    Ok(result)
}

pub fn get_schema(tx: &Tx) -> Result<Schema> {
    let mut stmt = tx
        .prepare_cached(
            "SELECT itemType.value, propertyName.value, valueType.value \
        FROM \
            items as item, \
            strings as itemType, \
            strings as propertyName, \
            strings as valueType \
        WHERE \
            item.type = 'ItemPropertySchema' \
            AND item.deleted = 0 \
            AND itemType.item = item.rowid \
            AND propertyName.item = item.rowid \
            AND valueType.item = item.rowid \
            AND itemType.name = 'itemType' \
            AND propertyName.name = 'propertyName' \
            AND valueType.name = 'valueType';",
        )
        .context_str("Failed to prepare SQL get_schema query")?;
    let mut rows = stmt.query([])?;
    let mut schema = Schema::empty();
    while let Some(row) = rows.next()? {
        let item_type: String = row.get(0)?;
        let property_name: String = row.get(1)?;
        let value_type: String = row.get(2)?;
        let value_type = SchemaPropertyType::from_string(&value_type).map_err(|e| {
            internal_error! {"{e}"}
        })?;
        schema.set_type_unchecked(item_type, property_name, value_type);
    }

    let mut edge_stmt = tx
        .prepare_cached(
            "SELECT edgeName.value, sourceType.value, targetType.value \
        FROM \
            items as item, \
            strings as edgeName, \
            strings as sourceType, \
            strings as targetType \
        WHERE \
            item.type = 'ItemEdgeSchema' \
            AND item.deleted = 0 \
            AND edgeName.item = item.rowid \
            AND sourceType.item = item.rowid \
            AND targetType.item = item.rowid \
            AND edgeName.name = 'edgeName' \
            AND sourceType.name = 'sourceType' \
            AND targetType.name = 'targetType';",
        )
        .context_str("Failed to prepare SQL get_schema edge query")?;

    let mut edge_rows = edge_stmt.query([])?;
    while let Some(row) = edge_rows.next()? {
        let edge_name: String = row.get(0)?;
        let source_type: String = row.get(1)?;
        let target_type: String = row.get(2)?;
        schema.set_edgetype_unchecked(source_type, edge_name, target_type);
    }

    Ok(schema)
}

pub fn get_schema_item_id_for_property(
    tx: &Tx,
    item_type: &str,
    property_name: &str,
) -> Result<String> {
    let sql = "\
        SELECT \
            item.id \
        FROM \
            items as item, \
            strings as nameEntry, \
            strings as typeEntry \
        WHERE \
            item.type = 'ItemPropertySchema' \
            AND item.deleted = 0 \
            AND typeEntry.item = item.rowid \
            AND nameEntry.item = item.rowid \
            AND typeEntry.name = 'itemType' \
            AND typeEntry.value = ? \
            AND nameEntry.name = 'propertyName' \
            AND nameEntry.value = ? \
    ;";
    let mut stmt = tx
        .prepare_cached(sql)
        .context_str("Failed to prepare SQL query")
        .context_str(sql)?;
    let mut rows: Rows = stmt.query(params![item_type, property_name])?;
    if let Some(row) = rows.next()? {
        Ok(row.get(0)?)
    } else {
        Err(internal_error! { "Failed to find property {item_type}.{property_name}"})
    }
}

pub fn get_schema_edge_item_id(
    tx: &Tx,
    edge_name: &str,
    source_type: &str,
    target_type: &str,
) -> Result<String> {
    let sql = "\
        SELECT \
            item.id \
        FROM \
            items as item, \
            strings as edgeName, \
            strings as sourceType, \
            strings as targetType \
        WHERE \
            item.type = 'ItemEdgeSchema' \
            AND item.deleted = 0 \
            AND sourceType.item = item.rowid \
            AND targetType.item = item.rowid \
            AND edgeName.name = 'edgeName' \
            AND edgeName.value = ? \
            AND sourceType.name = 'sourceType' \
            AND sourceType.value = ? \
            AND targetType.name = 'targetType' \
            AND targetType.value = ? \
    ;";
    let mut stmt = tx
        .prepare_cached(sql)
        .context_str("Failed to prepare SQL query")
        .context_str(sql)?;
    let mut rows: Rows = stmt.query(params![edge_name, source_type, target_type])?;
    if let Some(row) = rows.next()? {
        Ok(row.get(0)?)
    } else {
        Err(
            internal_error! { "Failed to find the edge {source_type} -- {edge_name} -> {target_type}" },
        )
    }
}

pub fn delete_schema_items_by_item_type_and_prop(
    tx: &Tx,
    item_type: &str,
    property_name: &str,
) -> Result<()> {
    let sql = "\
        SELECT \
            item.rowid \
        FROM \
            items as item, strings as itemTypeStr, strings as propNameStr \
        WHERE \
            item.type = 'ItemPropertySchema' \
            AND item.rowid = itemTypeStr.item \
            AND item.rowid = propNameStr.item \
            AND itemTypeStr.name = 'itemType' \
            AND itemTypeStr.value = ? \
            AND propNameStr.name = 'propertyName' \
            AND propNameStr.value = ? \
        ;";
    let mut stmt = tx.prepare_cached(sql)?;
    let mut rows = stmt.query(params![item_type, property_name])?;
    while let Some(row) = rows.next()? {
        let rowid: Rowid = row.get(0)?;
        dangerous_permament_remove_item(tx, rowid)?;
    }
    Ok(())
}

#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub enum Comparison {
    Equals,
    GreaterThan,
    GreaterOrEquals,
    LessThan,
    LessOrEquals,
}

fn add_sql_param(query: &mut String, column: &str, operation: &Comparison) {
    query.push_str(column);
    match operation {
        Comparison::Equals => query.push_str(" = "),
        Comparison::GreaterThan => query.push_str(" > "),
        Comparison::GreaterOrEquals => query.push_str(" >= "),
        Comparison::LessThan => query.push_str(" < "),
        Comparison::LessOrEquals => query.push_str(" <= "),
    };
    query.push_str("? AND ");
}

#[cfg(test)]
pub mod tests {
    use crate::{async_db_connection::AsyncTx, test_helpers::get_memory_connection};

    use super::{
        super::{error::Result, schema},
        *,
    };

    use serde_json::json;
    use std::ops::Not;

    pub fn random_id() -> String {
        rand::random::<i64>().to_string()
    }

    #[tokio::test]
    async fn test_insert_item() -> Result<()> {
        let mut conn = get_memory_connection()?;

        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let item1 = insert_item_base(&tx, &random_id(), "Person", date, date, date, false)?;
            let item2 = insert_item_base(&tx, &random_id(), "Book", date, date, date, false)?;
            assert_eq!(item2 - item1, 1);
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_insert_properties() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let item = insert_item_base(&tx, &random_id(), "Person", date, date, date, false)?;
            insert_integer(&tx, item, "age", 20)?;
            insert_real(&tx, item, "attack", 13.5)?;
            insert_string(&tx, item, "trait", "resilient")?;
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_insert_edge() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let source = insert_item_base(&tx, &random_id(), "Person", date, date, date, false)?;
            let target = insert_item_base(&tx, &random_id(), "Person", date, date, date, false)?;
            assert_eq!(target - source, 1);
            let item = insert_item_base(&tx, &random_id(), "Edge", date, date, date, false)?;
            let edge = insert_edge(&tx, item, source, "friend", target)?;
            assert_eq!(edge - target, 1);
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_default_schema() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let schema = get_schema(&tx)?;
            assert!(schema.all_properties_len() >= 3);
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_full_schema() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let item = insert_item_base(
                &tx,
                &random_id(),
                "ItemPropertySchema",
                date,
                date,
                date,
                false,
            )?;
            insert_string(&tx, item, "itemType", "Person")?;
            insert_string(&tx, item, "propertyName", "age")?;
            insert_string(&tx, item, "valueType", "Integer")?;

            let item = insert_item_base(
                &tx,
                &random_id(),
                "ItemPropertySchema",
                date,
                date,
                date,
                false,
            )?;
            insert_string(&tx, item, "itemType", "Person")?;
            insert_string(&tx, item, "propertyName", "name")?;
            insert_string(&tx, item, "valueType", "Text")?;

            let schema = get_schema(&tx)?;
            assert_eq!(schema.property("age"), Some(&SchemaPropertyType::Integer));
            assert_eq!(schema.property("name"), Some(&SchemaPropertyType::Text));
            assert!(schema.all_properties_len() >= 3);
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_delete_schema_items() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let item: Rowid = insert_item_base(
                &tx,
                &random_id(),
                "ItemPropertySchema",
                date,
                date,
                date,
                false,
            )?;
            insert_string(&tx, item, "itemType", "Person")?;
            insert_string(&tx, item, "propertyName", "age")?;
            insert_string(&tx, item, "valueType", "Integer")?;

            assert_eq!(
                search_items_params(&tx, Some(item), None, None, None, None, None)?.len(),
                1
            );
            dangerous_permament_remove_item(&tx, item)?;
            assert_eq!(
                search_items_params(&tx, Some(item), None, None, None, None, None)?.len(),
                0
            );

            Ok(())
        })
        .await
    }

    fn search_items_params(
        tx: &Tx,
        rowid: Option<Rowid>,
        id: Option<&str>,
        _type: Option<&str>,
        date_server_modified_gte: Option<DbTime>,
        date_server_modified_lt: Option<DbTime>,
        deleted: Option<bool>,
    ) -> Result<Vec<ItemBase>> {
        let database_search = DatabaseSearch {
            rowid,
            id,
            _type,
            date_server_modified_gte,
            date_server_modified_lt,
            deleted,
            sort_order: SortOrder::Asc,
            _limit: u64::MAX,
        };
        search_items(tx, &database_search)
    }

    #[tokio::test]
    async fn test_search() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let item1 = insert_item_base(&tx, "one", "Person", date, date, date, false)?;
            let _item2 = insert_item_base(&tx, "two", "Book", date, date, date + 1, false)?;
            let _item3 = insert_item_base(&tx, "three", "Street", date, date, date + 2, false)?;
            assert_eq!(
                search_items_params(&tx, None, None, Some("Person"), None, None, None)?.len(),
                1,
            );
            assert_eq!(
                search_items_params(&tx, None, None, Some("Void"), None, None, None)?.len(),
                0,
            );
            assert_eq!(
                search_items_params(&tx, Some(item1), None, None, None, None, None)?.len(),
                1,
            );
            assert_eq!(
                search_items_params(&tx, None, Some("one"), None, None, None, None)?.len(),
                1,
            );
            assert_eq!(
                search_items_params(&tx, None, Some("nothing"), None, None, None, None)?.len(),
                0,
            );
            assert_eq!(
                search_items_params(&tx, None, None, None, Some(date), None, None)?.len(),
                3,
            );
            assert_eq!(
                search_items_params(&tx, None, None, None, Some(date), Some(date + 3), None)?.len(),
                3,
            );
            assert_eq!(
                search_items_params(&tx, None, None, None, Some(date - 1), Some(date), None)?.len(),
                0,
            );
            assert_eq!(
                search_items_params(&tx, None, None, None, None, None, Some(true))?.len(),
                0,
            );
            assert_eq!(
                search_items_params(&tx, None, None, None, Some(date - 1), None, Some(false))?
                    .len(),
                3,
            );
            assert!(search_items_params(&tx, None, None, None, None, None, None)?.len() >= 3);
            {
                let mut query = DatabaseSearch {
                    rowid: None,
                    id: None,
                    _type: None,
                    date_server_modified_gte: Some(date),
                    date_server_modified_lt: None,
                    deleted: None,
                    sort_order: SortOrder::Asc,
                    _limit: 1,
                };
                assert_eq!(search_items(&tx, &query)?.len(), 1);

                query._limit = 3;
                let three_results = search_items(&tx, &query)?;
                assert_eq!(three_results.len(), 3);

                let first = &three_results[0];
                let third = &three_results[2];
                assert_ne!(first.rowid, third.rowid);
                assert_ne!(first.date_server_modified, third.date_server_modified);

                query.sort_order = SortOrder::Asc;
                let search_asc = &search_items(&tx, &query)?;
                query.sort_order = SortOrder::Desc;
                let search_desc = &search_items(&tx, &query)?;
                assert_eq!(search_asc[0].rowid, search_desc[2].rowid);
                assert_eq!(search_asc[2].rowid, search_desc[0].rowid);

                query._limit = 100;
                assert!(search_items(&tx, &query)?.len() < 100); // there aren't 100 items there
            }
            assert_eq!(
                search_items_params(
                    &tx,
                    Some(item1),
                    Some("one"),
                    Some("Person"),
                    Some(date),
                    Some(date + 1),
                    Some(false)
                )?
                .len(),
                1,
            );
            Ok(())
        })
        .await
    }

    // #[tokio::test]
    // async fn test_get_edges() -> Result<()> {
    //     let mut conn = get_memory_connection()?;
    //     conn.in_write_transaction(|tx: AsyncTx| async move {

    //     // Create edges
    //     let edges = vec![
    //         EdgeResult {
    //             rowid: 0,
    //             source: 0,
    //             name: "edge_1".to_owned(),
    //             target: 10,
    //         },
    //         EdgeResult {
    //             rowid: 1,
    //             source: 1,
    //             name: "edge_1".to_owned(),
    //             target: 10,
    //         },
    //         EdgeResult {
    //             rowid: 2,
    //             source: 0,
    //             name: "edge_2".to_owned(),
    //             target: 11,
    //         },
    //         EdgeResult {
    //             rowid: 3,
    //             source: 1,
    //             name: "edge_2".to_owned(),
    //             target: 11,
    //         },
    //     ];

    //     for edge in edges.iter() {
    //         insert_edge(&tx, edge.rowid, edge.source, &edge.name, edge.target)?;
    //     }

    //     // Test retreive all edges
    //     let ids = vec![0, 1];
    //     let result_1 = get_edges(&tx, ids, None)?;
    //     assert_eq!(result_1.len(), 4);
    //     for (i, r) in result_1.iter().enumerate() {
    //         let other = &edges[i];
    //         assert_eq!(r.rowid, other.rowid);
    //         assert_eq!(r.source, other.source);
    //         assert_eq!(r.name, other.name);
    //         assert_eq!(r.target, other.target);
    //     }

    //     // Test retreive with name=edge_1
    //     let ids = vec![0, 1];
    //     let result_2 = get_edges(&tx, ids, Some(&"edge_1".to_owned()))?;
    //     assert_eq!(result_2.len(), 2);
    //     for (i, r) in result_2.iter().enumerate() {
    //         let other = &edges[i];
    //         assert_eq!(r.rowid, other.rowid);
    //         assert_eq!(r.source, other.source);
    //         assert_eq!(r.name, other.name);
    //         assert_eq!(r.target, other.target);
    //     }

    //     Ok(())
    // }

    #[tokio::test]
    async fn test_gql_search() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let item: Rowid = insert_item_base(
                &tx,
                &random_id(),
                "ItemPropertySchema",
                date,
                date,
                date,
                false,
            )?;
            insert_string(&tx, item, "itemType", "Message")?;
            insert_string(&tx, item, "propertyName", "content")?;
            insert_string(&tx, item, "valueType", "Text")?;
            let item2: Rowid = insert_item_base(
                &tx,
                &random_id(),
                "ItemPropertySchema",
                date,
                date,
                date,
                false,
            )?;
            insert_string(&tx, item2, "itemType", "Message")?;
            insert_string(&tx, item2, "propertyName", "some_int")?;
            insert_string(&tx, item2, "valueType", "Integer")?;

            let _item1 = insert_item_base(&tx, "0x1", "Message", date, date, date, false)?;
            insert_string(&tx, _item1, "content", "test")?;

            // Build query
            let rowids = vec![0, 1, 2, _item1];
            let properties = vec![
                String::from("id"),
                String::from("content"),
                String::from("some_int"),
            ];
            let item_type = String::from("Message");
            let search = GQLSearchArgs {
                _type: Some(item_type),
                properties,
                rowids: Some(rowids),
                limit: Some(100),
                offset: Some(0),
                sort_property: "content".to_owned(),
                sort_order: SortOrder::default(),
                filters: None,
            };

            let schema = get_schema(&tx)?;
            let gql_result = search_gql_properties(&tx, &search, &schema)?;
            assert_eq!(gql_result.len(), 1);
            let properties = &gql_result.first().unwrap().properties;
            assert_eq!(properties.len(), 8); // 3 requested + 5 base
            assert_eq!(properties.get("id"), Some(&json!("0x1")));
            assert_eq!(properties.get("content"), Some(&json!("test")));
            assert_eq!(properties.get("some_int"), Some(&Value::Null));
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_limiting() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();
            let _item1 = insert_item_base(&tx, "one", "Person", date, date, date, false)?;
            let _item2 = insert_item_base(&tx, "two", "Book", date, date, date, false)?;
            let _item3 = insert_item_base(&tx, "three", "Street", date, date, date + 1, false)?;
            let query = DatabaseSearch {
                rowid: None,
                id: None,
                _type: None,
                date_server_modified_gte: Some(date),
                date_server_modified_lt: None,
                deleted: None,
                sort_order: SortOrder::Asc,
                _limit: 1,
            };
            // 1 main item + 1 other item with identical `dateServerModified`
            assert_eq!(search_items(&tx, &query)?.len(), 2);
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_property_checks() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let date = schema::utc_millis();

            let item: Rowid = insert_item_base(
                &tx,
                &random_id(),
                "ItemPropertySchema",
                date,
                date,
                date,
                false,
            )?;
            insert_string(&tx, item, "itemType", "Person")?;
            insert_string(&tx, item, "propertyName", "age")?;
            insert_string(&tx, item, "valueType", "Integer")?;

            assert!(check_string_exists(&tx, item, "itemType", "Person")?);
            assert!(check_string_exists(&tx, item, "itemType", "Person2")?.not());

            // The property should have a String value,
            // so normally this would be a schema check error.
            // However, database_api is the lowest layer and it's unaware of schemas.
            // The result is a successful check with the result "no, such integer value is not found")
            assert!(check_integer_exists(&tx, item, "itemType", 1)?.not());
            assert!(check_real_exists(&tx, item, "itemType", 1.)?.not());

            Ok(())
        })
        .await
    }
}
