use crate::{
    any_error,
    async_db_connection::AsyncConnection,
    bad_request, command_line_interface, constants, database_migrate_refinery,
    error::{ErrorContext, Result},
    internal_error,
    plugin_auth_crypto::{DatabaseKey, SHA256Output},
};
use futures::future::join_all;
use http::StatusCode;
use r2d2::{Pool, PooledConnection as R2D2PolledConnection};
use r2d2_sqlite::SqliteConnectionManager;
use rusqlite::OpenFlags;
use std::{
    collections::HashMap,
    path::{Path, PathBuf},
    time::Duration,
};
use tokio::fs::remove_file;
use tracing::{debug, error, info, warn};

pub type PooledConnection = R2D2PolledConnection<SqliteConnectionManager>;

#[derive(Clone)]
pub struct ConnectionPool {
    inner: Pool<SqliteConnectionManager>,
    password_hash: SHA256Output,
}

pub type InitDb = tokio::sync::RwLock<HashMap<String, ConnectionPool>>;

pub async fn initialize_db(
    owner: &str,
    init_db: &InitDb,
    database_key: &DatabaseKey,
) -> Result<()> {
    let mut db = init_db.write().await;
    if owner_database_path(owner).exists() {
        return Err(bad_request! { "Account for {owner} already exist" });
    } else if db.get(owner).is_some() {
        return Err(
            internal_error! { "DB file does not exist but connection pool does for {owner}" },
        );
    }

    let pool = initialize_new_pool(owner, database_key)?;

    db.insert(owner.to_string(), pool);

    Ok(())
}

pub async fn get_db_connection(
    owner: &str,
    init_db: &InitDb,
    database_key: &DatabaseKey,
) -> Result<AsyncConnection> {
    let pool = get_pool(owner, init_db, database_key).await?;
    let conn = get_connection_from_pool(&pool, database_key)?;

    // Returned connection is Send, but not Sync, hence it's enforced during compilation
    // that only one thread at a time can use it. That means sqlite can be set to
    // multithread mode, not default serialized
    Ok(AsyncConnection::new(conn))
}

pub async fn remove_db(owner: &str, init_db: &InitDb, database_key: &DatabaseKey) -> Result<()> {
    let pool = get_pool(owner, init_db, database_key).await?;
    if database_key.get_sha256() != pool.password_hash {
        return Err(any_error! {
            StatusCode::UNAUTHORIZED,
            "Invalid password key"
        });
    }
    {
        // caller is the owner, so we can remove the db
        let mut db = init_db.write().await;

        db.remove(owner);

        info!("Removing database at {:?}", owner_database_path(owner));
        // TODO handle errors
        let remove_futures = vec![
            remove_file(owner_database_path(owner)),
            remove_file(owner_database_shm_path(owner)),
            remove_file(owner_database_wal_path(owner)),
        ];
        let result = join_all(remove_futures).await;

        // Return error if main db file was not removed
        if result[0].is_err() {
            error!(
                "Failed to remove database at {:?}",
                owner_database_path(owner)
            );
            return Err(internal_error! { "Failed to remove database" });
        }

        // warn if other files were not removed
        for r in result {
            if r.is_err() {
                warn!("Failed to remove database file: {:?}", r);
            }
        }
    }

    Ok(())
}

fn initialize_new_pool(owner: &str, database_key: &DatabaseKey) -> Result<ConnectionPool> {
    let database_path = owner_database_path(owner);

    let pool = ConnectionPool {
        inner: create_pool(database_path)?,
        password_hash: database_key.get_sha256(),
    };

    // Database might already exist, with some key already set. At this moment we don't know if
    // key provided by owner is valid,
    let mut conn = get_connection_from_pool(&pool, database_key)?;

    // Try to query the db to see if it works
    check_connection(&mut conn)?;

    // Do it under the lock, so it's guaranteed all connections will see a database
    // with migrations in place.
    database_migrate_refinery::migrate(&mut conn)?;

    Ok(pool)
}

fn create_pool<P>(database_path: P) -> Result<Pool<SqliteConnectionManager>>
where
    P: AsRef<Path>,
{
    let manager = {
        SqliteConnectionManager::file(database_path)
            .with_init(|conn| {
                conn.set_prepared_statement_cache_capacity(256);

                Ok(())
            })
            .with_flags(
                // Create the file if not exist
                OpenFlags::SQLITE_OPEN_CREATE
                // Open in "multi thread" mode, sqlite disables synchronization in connection.
                // That is correct as long as one connection is used by exactly one thread at a time,
                // which is the case for the POD
                | OpenFlags::SQLITE_OPEN_NO_MUTEX
                // Open for reading and writing
                | OpenFlags::SQLITE_OPEN_READ_WRITE
                // Use private cache per connection, shared cache introduces a lot of synchronization
                // between them impacting performance seriously.
                | OpenFlags::SQLITE_OPEN_PRIVATE_CACHE,
            )
    };

    let num_connections =
        if let Some(fixed_opened_conns) = command_line_interface::PARSED.opened_connections {
            debug!("Using pre-configured number of connections");
            fixed_opened_conns
        } else {
            u32::max(num_cpus::get() as u32 * 4, 8)
        };

    info!(num_connections, "creating a database pool");

    let pool =
            // Builder blocks until reaches requested number of connections
            r2d2::Pool::builder()
                .max_size(num_connections)
                // Calls to Pool::get will wait this long for a connection to become available before returning an error.
                .connection_timeout(Duration::from_secs(5))
                .build(manager)?;

    Ok(pool)
}

fn owner_database_path_with_extension(owner: &str, extension: &str) -> PathBuf {
    PathBuf::from(constants::DATABASE_DIR)
        .join(owner)
        .with_extension(extension)
}

pub fn owner_database_path(owner: &str) -> PathBuf {
    owner_database_path_with_extension(owner, constants::DATABASE_EXTENSION)
}

fn owner_database_wal_path(owner: &str) -> PathBuf {
    owner_database_path_with_extension(
        owner,
        format!("{}-wal", constants::DATABASE_EXTENSION).as_str(),
    )
}

fn owner_database_shm_path(owner: &str) -> PathBuf {
    owner_database_path_with_extension(
        owner,
        format!("{}-shm", constants::DATABASE_EXTENSION).as_str(),
    )
}

async fn get_pool(
    owner: &str,
    init_db: &InitDb,
    database_key: &DatabaseKey,
) -> Result<ConnectionPool> {
    // Double locking pattern, firstly try to get the pool using cheap read lock.
    // Will succeed always, without any kind of blocking if pool exists.
    // If there is no pool yet, acquire expensive write lock, that happens exactly once
    // per user.
    let pool = {
        let first_try = {
            let db = init_db.read().await;
            db.get(owner).cloned()
        };

        if let Some(pool) = first_try {
            if !owner_database_path(owner).exists() {
                return Err(internal_error! { "No DB, but connection pool exist for {owner}" });
            }

            pool
        } else {
            if !owner_database_path(owner).exists() {
                return Err(any_error! {
                    StatusCode::UNAUTHORIZED,
                    "No such account: {owner}"
                });
            }

            // DB file exists - user registered, but there is no pool yet.
            // Gather write lock
            let mut db = init_db.write().await;

            // There might be situation someone else put it moment ago, so check for existence again
            if let Some(pool) = db.get(owner) {
                pool.clone()
            } else {
                // Under exclusive write lock there is nothing in the hash map, create the pool then
                let pool = initialize_new_pool(owner, database_key)?;

                db.insert(owner.to_string(), pool.clone());

                pool
            }
        }
    };

    Ok(pool)
}

fn check_connection(conn: &mut PooledConnection) -> Result<()> {
    conn.execute_batch("SELECT COUNT(*) FROM sqlite_schema;")?;

    Ok(())
}

/// Get connection from pool and apply required pragmas on them
/// Don't use `with_init` function from the Pool builder, since it
/// will store secret key value over the lifetime of the pool.
fn get_connection_from_pool(
    pool: &ConnectionPool,
    database_key: &DatabaseKey,
) -> Result<PooledConnection> {
    if database_key.get_sha256() != pool.password_hash {
        return Err(any_error! {
            StatusCode::UNAUTHORIZED,
            "Invalid password key"
        });
    }

    let conn = pool.inner.get()?;
    DatabaseKey::execute_sqlite_pragma(database_key, &conn)
        .context_str("Failed to open the database file (did databaseKey change between runs?)")?;

    conn.execute_batch(
        "
        PRAGMA foreign_keys = ON;

        -- Strategy for syncing data to the storage, NORMAL is recommended for WAL
        PRAGMA synchronous = NORMAL;

        -- Temporal storage set to MEMORY, don't use FILE setting, since SQLCipher does not do encryption on them
        PRAGMA temp_store = 2;

        -- Milliseconds that connection will wait to be unblocked with its operations, after that period SQL_BUSY is returned
        PRAGMA busy_timeout = 3000;

        -- Set the maximum number of pages in the database file.
        -- Currently 1GB given a page size of 4096 (default >= 3.12.0). 1000000000 / 4096 = 262144
        PRAGMA max_page_count = 262144;

        -- Use Write Ahead Logging journal that allows multiple readers and at most one writer to work seamlessly
        PRAGMA journal_mode = WAL;

        -- Do auto-checkpoints if journal file exceeds this number of pages
        -- The checkpoint is PASSIVE, meaning will succeed iff there is nothing preventing from doing it
        PRAGMA wal_autocheckpoint = 1000;

        -- The DB will truncate the WAL journal to that size after doing a checkpoint.
        -- This is a trade off made here, keep the journal on the storage for better performance
        -- (do not re-create it each time is needed), but keep it reasonably small so -wal files
        -- will not clutter our cloud space too much.
        -- This setting is a matter of change after releasing POD to the wild and check how it behaves
        -- on users accounts.
        -- Set to 10MB
        -- Note under heavy write load -wal may exceed this limit extensively, it will be shortened
        -- after the write 'wave' goes away.
        PRAGMA journal_size_limit = 10485760;

        -- Number of bytes of the database file that will be accessed using memory-mapped I/O.
        -- mmaping reduces amount of read/write syscalls.
        -- Write happens to -wal file, so there is no that much impact while writing
        -- 30 MB
        PRAGMA mmap_size = 31457280;

        --- Neg value to set Kibibytes (positive for pages)
        PRAGMA cache_size = -1280000;
        ",
    )?;

    Ok(conn)
}

#[cfg(test)]
pub mod tests {
    use rusqlite::{
        ffi::{Error as FFIError, ErrorCode},
        Error::SqliteFailure,
    };
    use tokio::sync::RwLock;

    use crate::error::{Error, ErrorType};

    use super::*;

    struct TempDb {
        pub database_path: PathBuf,
    }

    impl TempDb {
        fn new(owner: &str) -> Self {
            let database_path = owner_database_path(owner);

            if !Path::new(constants::DATABASE_DIR).is_dir() {
                std::fs::create_dir_all(constants::DATABASE_DIR).unwrap();
            }

            Self { database_path }
        }

        fn rm_files(&self) {
            let _ = std::fs::remove_file(&self.database_path);
            let _ = std::fs::remove_file(format!("{}-shm", self.database_path.display()));
            let _ = std::fs::remove_file(format!("{}-wal", self.database_path.display()));
        }
    }

    impl Drop for TempDb {
        fn drop(&mut self) {
            self.rm_files();
        }
    }

    const NORMAL: u32 = 1;
    const ON: u32 = 1;
    const MEMORY: u32 = 2;
    const BUSY_TIMEOUT: u32 = 3000;
    const WAL: &str = "wal";
    const MAX_PAGE_COUNT: u32 = 262144;

    #[tokio::test]
    async fn pragmas_are_set() {
        let owner = "testing_pragmas_are_set";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let conn = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();

        assert_eq!(
            ON,
            conn.query_row("pragma foreign_keys;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );

        assert_eq!(
            NORMAL,
            conn.query_row("pragma synchronous;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );

        assert_eq!(
            MEMORY,
            conn.query_row("pragma temp_store;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );
        assert_eq!(
            BUSY_TIMEOUT,
            conn.query_row("pragma busy_timeout;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );
        assert_eq!(
            WAL,
            conn.query_row("pragma journal_mode;", [], |row| row
                .get::<usize, String>(0))
                .unwrap()
        );
        assert_eq!(
            MAX_PAGE_COUNT,
            conn.query_row("pragma max_page_count;", [], |row| row.get::<usize, u32>(0))
                .unwrap()
        );
    }

    #[tokio::test]
    async fn fails_on_rollback_ok_on_wal() {
        let owner = "testing_fails_on_rollback_ok_on_wal";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let connection_a = get_db_connection(owner, &init_db, &database_key).await;

        let connection_a = connection_a.unwrap();

        let connection_b = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();

        // Journal: this one gets called
        // connection_a.busy_handler(Some(|a| {
        //     println!("busy handler for connection_a called!");
        //     true
        // }))
        // .unwrap();

        connection_a
            .execute_batch(
                "
            create table if not exists t(x integer);
            insert into t values(1);
            insert into t values(2);
            ",
            )
            .unwrap();

        connection_a.execute_batch("begin transaction;").unwrap();

        connection_b.execute_batch("begin transaction;").unwrap();

        // In rollback: connection_b gets SHARED lock
        connection_b.execute_batch("select * from t;").unwrap();

        // In rollback: connection_a gets RESERVED (intent to write lock)
        connection_a
            .execute_batch("insert into t values(1);")
            .unwrap();

        let journal_mode = connection_a
            .query_row("pragma journal_mode;", [], |row| {
                row.get::<usize, String>(0)
            })
            .unwrap();

        if journal_mode == WAL {
            // In WAL: it's designed in such a way readers does not interfere with writers,
            // so this passes
            assert!(matches!(connection_a.execute_batch("commit;"), Ok(())));
        } else {
            // In rollback: tries to go to EXCLUSIVE, can get it, iff there are no other locks
            // but there is one SHARED held by connection_b, connection_a waits BUSY_TIMEOUT and returns with error
            // If there is busy_handler set, then sqlite will call it as long as it returns true,
            // or is able to proceed.
            assert!(matches!(
                connection_a.execute_batch("commit;"),
                Err(SqliteFailure(
                    FFIError {
                        code: ErrorCode::DatabaseBusy,
                        extended_code: 5
                    },
                    Some(_)
                ))
            ))
        }

        connection_b.execute_batch("commit;").unwrap();
    }

    #[test_log::test(tokio::test)]
    async fn deadlock_detection() {
        let owner = "testing_deadlock_detection";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let connection_a = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();

        connection_a
            .busy_handler(Some(|_a| {
                println!("busy handler for connection_a called!");
                true
            }))
            .unwrap();

        let connection_b = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();

        connection_a
            .execute_batch(
                "
            create table if not exists t(x integer);
            insert into t values(1);
            insert into t values(2);
            ",
            )
            .unwrap();

        connection_a.execute_batch("begin transaction;").unwrap();

        connection_b.execute_batch("begin transaction;").unwrap();

        // connection_a gets SHARED lock
        connection_a.execute_batch("select * from t;").unwrap();

        // connection_b gets SHARED lock
        connection_b.execute_batch("select * from t;").unwrap();

        // connection_b gets RESERVED, there can be only 1 at a time of such kind
        connection_b
            .execute_batch("insert into t values(1);")
            .unwrap();

        // connection_a tries to get RESERVED. It fails IMMEDIATELY, because in order to jump to
        // EXCLUSIVE later, all locks has to be cleared out, but that will never happen,
        // because connection_b also holds RESERVED and also will wait to all locks to drop.
        // If both connection_a and connection_b would gather RESERVED, then:
        // connection_a would wait for connection_b to drop, connection_b would wait for connection_a to drop - making a deadlock
        // Such situation is detected by SQLITE, and aborts this call immediately ->
        // allowing connection connection_b to gather exclusive lock later.
        //
        // Note, no busy_handlers are called
        assert!(matches!(
            connection_a.execute_batch("insert into t values(1);"),
            Err(SqliteFailure(
                FFIError {
                    code: ErrorCode::DatabaseBusy,
                    extended_code: 5
                },
                // Result message says "database locked" that suggest SQLITE_LOCKED
                // error, which is completely different thing
                // and makes a lot of confusion.
                Some(_msg)
            ))
        ));

        connection_a.execute_batch("commit;").unwrap();

        connection_b.execute_batch("commit;").unwrap();
    }

    #[tokio::test]
    async fn reader_and_2_writers() {
        let owner = "testing_reader_and_2_writers";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let reader_a = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();

        let writer_b = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();
        let writer_c = get_db_connection(owner, &init_db, &database_key)
            .await
            .unwrap();

        writer_b
            .execute_batch(
                "
            create table if not exists t(x integer);
            insert into t values(1);
            insert into t values(2);
            ",
            )
            .unwrap();

        reader_a.execute_batch("begin transaction;").unwrap();
        reader_a.execute_batch("select * from t;").unwrap();

        writer_b.execute_batch("begin transaction;").unwrap();
        writer_b.execute_batch("insert into t values(1);").unwrap();

        writer_c.execute_batch("begin transaction;").unwrap();

        // In journal and WAL
        // waits busy_timeout milliseconds for writer_A to finish
        assert!(matches!(
            writer_c.execute_batch("insert into t values(1);"),
            Err(SqliteFailure(
                FFIError {
                    code: ErrorCode::DatabaseBusy,
                    extended_code: 5
                },
                // Result message says "database locked" that suggest SQLITE_LOCKED
                // error, which is completely different thing
                // and makes a lot of confusion.
                Some(_msg)
            ))
        ));
    }

    #[tokio::test]
    async fn pool_honors_only_valid_database_key() {
        let owner = "pool_honors_only_valid_database_key";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let database_key_invalid = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbCD".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        {
            let pool = get_pool(owner, &init_db, &database_key).await.unwrap();
            // init all connections with valid key
            let mut conns = vec![];

            for i in 0..pool.inner.max_size() {
                let conn = get_connection_from_pool(&pool, &database_key).unwrap();

                // Insert record to a table with given connection id
                conn.execute_batch(&format!(
                    "
                    create table if not exists t(x integer);
                    insert into t values({});
                    ",
                    i + 1
                ))
                .unwrap();

                conns.push(conn);
            }

            // return all connections to the pool, by dropping conns
        }

        {
            let pool = get_pool(owner, &init_db, &database_key_invalid)
                .await
                .unwrap();

            // try to get connection with INVALID key

            assert!(matches!(
                get_connection_from_pool(&pool, &database_key_invalid),
                Err(Error {
                    context: _,
                    error: ErrorType::Any {
                        code: StatusCode::UNAUTHORIZED,
                        msg: _
                    }
                })
            ));
        }

        {
            let pool = get_pool(owner, &init_db, &database_key).await.unwrap();

            // init all connections with VALID key
            let mut conns = vec![];

            for _ in 0..pool.inner.max_size() {
                let conn = get_connection_from_pool(&pool, &database_key).unwrap();

                // query table with given connection
                // that should PASS!

                {
                    let mut q = conn.prepare("select * from t").unwrap();
                    let mut rows = q.query([]).unwrap();

                    // Check that content is valid
                    let mut i = 0;
                    while let Ok(Some(r)) = rows.next() {
                        assert_eq!(i + 1, r.get::<_, u32>(0).unwrap());
                        i += 1;
                    }
                }
                conns.push(conn);
            }

            // return all connections to the pool, by dropping conns
        }
    }

    #[tokio::test]
    async fn test_remove_db() {
        let owner = "test_remove_db";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let pool = get_pool(owner, &init_db, &database_key).await.unwrap();

        let conn = get_connection_from_pool(&pool, &database_key).unwrap();

        conn.execute_batch(
            "
            create table if not exists t(x integer);
            insert into t values(1);
            ",
        )
        .unwrap();

        info!("database_path_buf: {:?}", owner_database_path(owner));

        assert!(Path::new(&owner_database_path(owner)).exists());
        remove_db(owner, &init_db, &database_key).await.unwrap();
        assert!(!Path::new(&owner_database_path(owner)).exists());
        assert!(!Path::new(&owner_database_shm_path(owner)).exists());
        assert!(!Path::new(&owner_database_wal_path(owner)).exists());
    }

    #[tokio::test]
    async fn test_error_remove_db() {
        let owner = "test_error_remove_db";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let pool = get_pool(owner, &init_db, &database_key).await.unwrap();

        let conn = get_connection_from_pool(&pool, &database_key).unwrap();

        conn.execute_batch(
            "
            create table if not exists t(x integer);
            insert into t values(1);
            ",
        )
        .unwrap();

        info!("database_path_buf: {:?}", owner_database_path(owner));

        // delete file manually
        std::fs::remove_file(owner_database_path(owner)).unwrap();

        assert!(!Path::new(&owner_database_path(owner)).exists());
        assert!(matches!(
            remove_db(owner, &init_db, &database_key).await,
            Err(Error {
                context: _,
                error: ErrorType::Internal(_)
            })
        ));
        assert!(!Path::new(&owner_database_path(owner)).exists());
    }

    #[tokio::test]
    async fn test_remove_db_wrong_credentials() {
        let owner = "test_remove_db_wrong_credentials";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let pool = get_pool(owner, &init_db, &database_key).await.unwrap();

        let conn = get_connection_from_pool(&pool, &database_key).unwrap();

        conn.execute_batch(
            "
            create table if not exists t(x integer);
            insert into t values(1);
            ",
        )
        .unwrap();

        assert!(Path::new(&owner_database_path(owner)).exists());

        // Try to remove database with wrong credentials
        let database_key = DatabaseKey::from(
            "abdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbda".to_string(),
        )
        .unwrap();
        assert!(matches!(
            remove_db(owner, &init_db, &database_key).await,
            Err(Error {
                context: _,
                error: ErrorType::Any {
                    code: StatusCode::UNAUTHORIZED,
                    msg: _
                }
            })
        ));
        assert!(Path::new(&owner_database_path(owner)).exists());
    }

    #[tokio::test]
    async fn test_number_of_connections() {
        let owner = "test_number_of_connections";
        let _temp_db = TempDb::new(owner);

        let database_key = DatabaseKey::from(
            "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb".to_string(),
        )
        .unwrap();

        let init_db = RwLock::new(HashMap::new());

        initialize_db(owner, &init_db, &database_key).await.unwrap();

        let pool = get_pool(owner, &init_db, &database_key).await.unwrap();

        let num_connections = u32::max(num_cpus::get() as u32 * 4, 8);

        let _conns: Vec<PooledConnection> = (0..num_connections)
            .map(|_| get_connection_from_pool(&pool, &database_key).unwrap())
            .collect();

        // Pool is empty, expect error
        assert!(get_connection_from_pool(&pool, &database_key).is_err());
    }
}
