use crate::{
    async_db_connection::{AsyncConnection, AsyncTx as Tx},
    bad_request,
    command_line_interface::{self, CliOptions},
    database_api,
    database_api::{IntegersNameValue, RealsNameValue, Rowid, StringsNameValue},
    database_pool::{get_db_connection, remove_db, InitDb},
    db_model::ItemBase,
    error::Result,
    file_api, forbidden, internal_error,
    plugin_auth_crypto::DatabaseKey,
    schema::{Schema, SchemaPropertyType},
};
use lazy_static::lazy_static;
use rusqlite::types::ValueRef;
use serde_json::{Map, Value};
use sha2::{
    digest::{generic_array::GenericArray, typenum::U32},
    Digest,
};
use std::{
    collections::{HashMap, HashSet},
    ops::Deref,
    str,
};
use tracing::{error, info, warn};

lazy_static! {
    static ref ALLOWED_OWNER_HASHES: HashSet<Vec<u8>> =
        allowed_owner_hashes_fn(&command_line_interface::PARSED);
}

/// Get all properties that the item has, ignoring those
/// that exist in the DB but are not defined in the Schema
pub fn get_item_properties(tx: &Tx, rowid: Rowid, schema: &Schema) -> Result<Map<String, Value>> {
    let mut json = serde_json::Map::new();

    for IntegersNameValue { name, value } in database_api::get_integers_records_for_item(tx, rowid)?
    {
        match schema.property(&name) {
            Some(SchemaPropertyType::Bool) => {
                json.insert(name, (value == 1).into());
            }
            Some(SchemaPropertyType::DateTime) | Some(SchemaPropertyType::Integer) => {
                json.insert(name, value.into());
            }
            other => {
                warn!(
                    "Ignoring item property {}: {} which according to Schema should be a {:?}",
                    name, value, other
                );
            }
        };
    }

    for StringsNameValue { name, value } in database_api::get_strings_records_for_item(tx, rowid)? {
        match schema.property(&name) {
            Some(SchemaPropertyType::Text) => {
                json.insert(name, value.into());
            }
            other => {
                warn!(
                    "Ignoring item property {}: {} which according to Schema should be a {:?}",
                    name, value, other
                );
            }
        }
    }

    for RealsNameValue { name, value } in database_api::get_reals_records_for_item(tx, rowid)? {
        match schema.property(&name) {
            Some(SchemaPropertyType::Real) => {
                json.insert(name, value.into());
            }
            other => {
                warn!(
                    "Ignoring item property {}: {} which according to Schema should be a {:?}",
                    name, value, other
                );
            }
        };
    }

    Ok(json)
}

pub fn check_item_has_all_properties(
    tx: &Tx,
    schema: &Schema,
    rowid: Rowid,
    props: &HashMap<String, Value>,
) -> Result<bool> {
    for (name, value) in props {
        if !check_item_has_property(tx, schema, rowid, name, value)? {
            return Ok(false);
        }
    }
    Ok(true)
}

pub fn check_item_has_property(
    tx: &Tx,
    schema: &Schema,
    rowid: Rowid,
    name: &str,
    value: &Value,
) -> Result<bool> {
    let dbtype = if let Some(t) = schema.property(name) {
        t
    } else {
        return Err(bad_request! {
            "Property {name} not defined in Schema (attempted to use it for json value {value})",
        });
    };

    match value {
        Value::Null => {
            let exists = match dbtype {
                SchemaPropertyType::Text => {
                    database_api::check_string_exists_by_name(tx, rowid, name)?
                }
                SchemaPropertyType::Integer => {
                    database_api::check_integer_exists_by_name(tx, rowid, name)?
                }
                SchemaPropertyType::Real => {
                    database_api::check_real_exists_by_name(tx, rowid, name)?
                }
                SchemaPropertyType::Bool => {
                    database_api::check_integer_exists_by_name(tx, rowid, name)?
                }
                SchemaPropertyType::DateTime => {
                    database_api::check_integer_exists_by_name(tx, rowid, name)?
                }
            };
            Ok(!exists)
        }
        Value::String(value) if dbtype == &SchemaPropertyType::Text => {
            database_api::check_string_exists(tx, rowid, name, value)
        }
        Value::Number(n) if dbtype == &SchemaPropertyType::Integer => {
            if let Some(value) = n.as_i64() {
                database_api::check_integer_exists(tx, rowid, name, value)
            } else {
                Err(bad_request! { "Failed to parse JSON number {n} to i64 ({name})" })
            }
        }
        Value::Number(n) if dbtype == &SchemaPropertyType::Real => {
            if let Some(value) = n.as_f64() {
                database_api::check_real_exists(tx, rowid, name, value)
            } else {
                Err(bad_request! { "Failed to parse JSON number {n} to f64 ({name})" })
            }
        }
        Value::Bool(b) if dbtype == &SchemaPropertyType::Bool => {
            database_api::check_integer_exists(tx, rowid, name, i64::from(*b))
        }
        Value::Number(n) if dbtype == &SchemaPropertyType::DateTime => {
            if let Some(value) = n.as_i64() {
                database_api::check_integer_exists(tx, rowid, name, value)
            } else if let Some(float) = n.as_f64() {
                warn!("Using float-to-integer conversion property {}, value {}. This might not be supported in the future, please use a compatible DateTime format https://gitlab.memri.io/memri/pod#understanding-the-schema", float, name);
                database_api::check_integer_exists(tx, rowid, name, float.round() as i64)
            } else {
                Err(bad_request! {
                    "Failed to parse JSON number {n} to DateTime ({name}), use i64 number instead"
                })
            }
        }
        _ => Err(bad_request! { "Failed to parse json value {value} to {dbtype:?} ({name})" }),
    }
}

pub fn insert_edge(
    tx: &Tx,
    schema: &Schema,
    self_rowid: i64,
    source: i64,
    name: &str,
    target: i64,
) -> Result<()> {
    let src_base = database_api::get_item_base(tx, source)?.ok_or_else(|| {
        internal_error! { "Edge connects to an nonexisting item.rowid {source}" }
    })?;
    let tgt_base = database_api::get_item_base(tx, target)?.ok_or_else(|| {
        internal_error! { "Edge connects to an nonexisting item.rowid {target}" }
    })?;

    let schema_tgt_type = if let Some(t) = schema.item_edge(&src_base._type, name) {
        t
    } else {
        return Err(bad_request! {"Edge {name} not defined in Schema" });
    };

    if !schema_tgt_type.contains(&tgt_base._type) {
        return Err(bad_request! {
            "Attempted to insert Edge with incorrect target type: got {} expected {:?}",
            tgt_base._type, schema_tgt_type,
        });
    };

    database_api::insert_edge(tx, self_rowid, source, name, target)?;

    Ok(())
}

pub fn insert_property(
    tx: &Tx,
    schema: &Schema,
    rowid: Rowid,
    item_type: &str,
    name: &str,
    json: &Value,
) -> Result<()> {
    let dbtype = if let Some(t) = schema.item_property(item_type, name) {
        t
    } else {
        return Err(bad_request! {
            "Property {name} not defined in Schema for {item_type} (attempted to use it for json value {json})",
        });
    };

    database_api::delete_property(tx, rowid, name)?;

    match json {
        Value::Null => (),
        Value::String(value) if dbtype == &SchemaPropertyType::Text => {
            database_api::insert_string(tx, rowid, name, value)?
        }
        Value::Number(n) if dbtype == &SchemaPropertyType::Integer => {
            if let Some(value) = n.as_i64() {
                database_api::insert_integer(tx, rowid, name, value)?
            } else {
                return Err(bad_request! { "Failed to parse JSON number {n} to i64 ({name})" });
            }
        }
        Value::Number(n) if dbtype == &SchemaPropertyType::Real => {
            if let Some(value) = n.as_f64() {
                database_api::insert_real(tx, rowid, name, value)?
            } else {
                return Err(bad_request! { "Failed to parse JSON number {n} to f64 ({name})" });
            }
        }
        Value::Bool(b) if dbtype == &SchemaPropertyType::Bool => {
            database_api::insert_integer(tx, rowid, name, i64::from(*b))?
        }
        Value::Number(n) if dbtype == &SchemaPropertyType::DateTime => {
            if let Some(value) = n.as_i64() {
                database_api::insert_integer(tx, rowid, name, value)?
            } else if let Some(float) = n.as_f64() {
                warn!("Using float-to-integer conversion property {}, value {}. This might not be supported in the future, please use a compatible DateTime format https://gitlab.memri.io/memri/pod#understanding-the-schema", float, name);
                database_api::insert_integer(tx, rowid, name, float.round() as i64)?
            } else {
                return Err(bad_request! {
                    "Failed to parse JSON number {n} to DateTime ({name}), use i64 number instead",
                });
            }
        }
        _ => {
            return Err(bad_request! { "Failed to parse json value {json} to {dbtype:?} ({name})" })
        }
    };
    Ok(())
}

pub fn item_base_to_json(tx: &Tx, item: ItemBase, schema: &Schema) -> Result<Map<String, Value>> {
    let mut props = get_item_properties(tx, item.rowid, schema)?;
    add_item_base_properties(&mut props, item);
    Ok(props)
}

pub fn get_table_from_type(property_type: &SchemaPropertyType) -> String {
    match property_type {
        SchemaPropertyType::DateTime | SchemaPropertyType::Integer | SchemaPropertyType::Bool => {
            "integers"
        }
        SchemaPropertyType::Text => "strings",
        SchemaPropertyType::Real => "reals",
    }
    .to_string()
}

pub fn valueref_to_json_value(
    value_ref: ValueRef,
    schema_type: &SchemaPropertyType,
) -> Result<Value> {
    // Convert a rusqlite valueref to json value
    let value = if value_ref == ValueRef::Null {
        Value::Null
    } else {
        match schema_type {
            SchemaPropertyType::Bool => (value_ref.as_i64().unwrap() == 1).into(),
            SchemaPropertyType::DateTime | SchemaPropertyType::Integer => {
                value_ref.as_i64().unwrap().into()
            }
            SchemaPropertyType::Text => value_ref.as_str().unwrap().into(),
            SchemaPropertyType::Real => value_ref.as_f64().unwrap().into(),
        }
    };
    Ok(value)
}

pub fn add_item_edge_properties(
    tx: &Tx,
    props: &mut Map<String, Value>,
    rowid: Rowid,
) -> Result<()> {
    if let Some(edge_self) = database_api::get_self_edge(tx, rowid)? {
        let source_base = database_api::get_item_base(tx, edge_self.source)?.ok_or_else(|| {
            internal_error! { "Edge {} connects to non-existing source {}", rowid, edge_self.source }
        })?;
        let target_base = database_api::get_item_base(tx, edge_self.target)?.ok_or_else(|| {
            internal_error! { "Edge {} connects to non-existing target {}", rowid, edge_self.target }
        })?;
        props.insert("_source".to_string(), Value::String(source_base.id));
        props.insert("_target".to_string(), Value::String(target_base.id));
        props.insert("_edge".to_string(), Value::String(edge_self.name));
    }
    Ok(())
}

fn add_item_base_properties(props: &mut Map<String, Value>, item: ItemBase) {
    props.insert("id".to_string(), item.id.into());
    props.insert("rowid".to_string(), item.rowid.into());
    props.insert("type".to_string(), item._type.into());
    props.insert("dateCreated".to_string(), item.date_created.into());
    props.insert("dateModified".to_string(), item.date_modified.into());
    props.insert(
        "dateServerModified".to_string(),
        item.date_server_modified.into(),
    );
    props.insert("deleted".to_string(), item.deleted.into());
}

fn allowed_owner_hashes_fn(cli_options: &CliOptions) -> HashSet<Vec<u8>> {
    let owners: &str = &cli_options.owners;
    let mut result = HashSet::new();
    for owner in owners.split(',') {
        let hexed = hex::decode(owner).unwrap_or_else(|err| {
            error!(
                "Pod owners are invalid, failed to decode hex {}, {}",
                owner, err
            );
            std::process::exit(1);
        });
        result.insert(hexed);
    }
    result
}

pub fn check_owner(possible_owner: &str) -> Result<()> {
    if command_line_interface::PARSED.owners == "ANY" {
        return Ok(());
    };
    let possible_hash = hash_of_hex(possible_owner)?;
    if ALLOWED_OWNER_HASHES.contains(possible_hash.deref()) {
        Ok(())
    } else {
        info!(
            "Denying unexpected owner {} with hash {}",
            possible_owner,
            hex::encode(possible_hash)
        );
        Err(forbidden! {
            "Unexpected owner",
        })
    }
}

/// Two methods combined into one to prevent creating a database connection without owner checks.
/// As additional failsafe to the fact that non-owners don't have the database key.
pub async fn check_owner_and_initialize_db(
    owner: &str,
    init_db: &InitDb,
    database_key: &DatabaseKey,
) -> Result<AsyncConnection> {
    check_owner(owner)?;
    get_db_connection(owner, init_db, database_key).await
}

fn hash_of_hex(hex_string: &str) -> Result<GenericArray<u8, U32>> {
    let hex_string: Vec<u8> = hex::decode(hex_string)?;
    let mut hash = sha2::Sha256::new();
    hash.update(&hex_string);
    Ok(hash.finalize())
}

// /// Two methods combined into one to prevent deleting a database without owner checks.
// /// As additional failsafe to the fact that non-owners don't have the database key.
pub async fn check_owner_and_delete_db(
    owner: &str,
    init_db: &InitDb,
    database_key: &DatabaseKey,
) -> Result<()> {
    check_owner(owner)?;
    // TODO: check if init_db.write() lock should be held here through whole function span.
    remove_db(owner, init_db, database_key).await?;
    file_api::remove_owner_files(owner).await
}

#[cfg(test)]
mod tests {
    use super::{super::schema, *};
    use crate::{
        async_db_connection::AsyncTx, database_api::tests::random_id,
        test_helpers::get_memory_connection,
    };
    use serde_json::json;
    use std::ops::Not;

    #[tokio::test]
    async fn test_one_property() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx).unwrap();
            schema.set_type_unchecked(
                "Person".to_string(),
                "age".to_string(),
                SchemaPropertyType::Integer,
            );
            schema.set_type_unchecked(
                "Person".to_string(),
                "strength".to_string(),
                SchemaPropertyType::Real,
            );
            schema.set_type_unchecked(
                "Person".to_string(),
                "myDescription".to_string(),
                SchemaPropertyType::Text,
            );

            let date = schema::utc_millis();
            let item: Rowid = database_api::insert_item_base(
                &tx,
                &random_id(),
                "Person",
                date,
                date,
                date,
                false,
            )?;

            assert!(check_item_has_property(&tx, &schema, item, "age", &json!(20))?.not());

            insert_property(&tx, &schema, item, "Person", "age", &json!(20))?;
            assert!(check_item_has_property(
                &tx,
                &schema,
                item,
                "age",
                &json!(20)
            )?);
            assert!(check_item_has_property(&tx, &schema, item, "age", &json!(99))?.not());

            // Checking non-existing property should yield an error, not a successful "no" response
            assert!(check_item_has_property(&tx, &schema, item, "antiAge", &json!(99)).is_err());

            insert_property(&tx, &schema, item, "Person", "strength", &json!(13.5))?;
            assert!(check_item_has_property(
                &tx,
                &schema,
                item,
                "strength",
                &json!(13.5)
            )?);

            insert_property(
                &tx,
                &schema,
                item,
                "Person",
                "myDescription",
                &json!("Wow such person"),
            )?;
            assert!(check_item_has_property(
                &tx,
                &schema,
                item,
                "myDescription",
                &json!("Wow such person")
            )?);

            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_all_properties() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx).unwrap();
            schema.set_type_unchecked(
                "Person".to_string(),
                "age".to_string(),
                SchemaPropertyType::Integer,
            );
            schema.set_type_unchecked(
                "Person".to_string(),
                "strength".to_string(),
                SchemaPropertyType::Real,
            );
            schema.set_type_unchecked(
                "Person".to_string(),
                "myDescription".to_string(),
                SchemaPropertyType::Text,
            );

            let date = schema::utc_millis();
            let item: Rowid = database_api::insert_item_base(
                &tx,
                &random_id(),
                "Person",
                date,
                date,
                date,
                false,
            )?;

            insert_property(&tx, &schema, item, "Person", "age", &json!(20))?;
            insert_property(&tx, &schema, item, "Person", "strength", &json!(13.5))?;
            insert_property(
                &tx,
                &schema,
                item,
                "Person",
                "myDescription",
                &json!("Wow such person"),
            )?;

            {
                let mut props = HashMap::new();
                props.insert("age".to_string(), json!(20));
                props.insert("strength".to_string(), json!(13.5));
                props.insert("myDescription".to_string(), json!("Wow such person"));
                assert!(check_item_has_all_properties(&tx, &schema, item, &props)?);
            }

            {
                let mut props = HashMap::new();
                props.insert("age".to_string(), json!(20));
                assert!(check_item_has_all_properties(&tx, &schema, item, &props)?);
            }

            {
                let mut props = HashMap::new();
                props.insert("age".to_string(), json!(99999999));
                props.insert("strength".to_string(), json!(13.5));
                assert!(check_item_has_all_properties(&tx, &schema, item, &props)?.not());
            }

            {
                let mut props = HashMap::new();
                props.insert("antiAge".to_string(), json!(-200000000));
                assert!(check_item_has_all_properties(&tx, &schema, item, &props).is_err());
            }

            {
                let props = HashMap::new();
                assert!(check_item_has_all_properties(&tx, &schema, item, &props)?);
            }

            Ok(())
        })
        .await
    }
}
