//! General rule of thumb while working with errors:
//!
//! Add another type to the enum if you want to distinct the error type.
//! Useful while error handling happens in the caller.
//! New type should reflect some functionality in the POD, look for examples.
//!
//! Use internal_error macro for quickly converting 3rd party libs
//! errors to POD error. In such case only message is important, not
//! exact type.
//!
//! Use handcrafted From<T: std::error::Error> for ErrorType implementations,
//! if logic for error type selection becomes non-trivial.
//!
//! Do not write From<T: std::error::Error> for Error.
//!
//! Use bad_request! and other macros to quickly bail out from the function
//! with meaningful error message, when type is not important.
use std::{error::Error as StdError, fmt::Display};

use http::status::StatusCode;
use thiserror::Error;

use crate::api_model::PluginIdentifier;

/// Shorthand for creating errors with handcrafted status code
#[macro_export]
macro_rules! any_error {
    ($err_code: expr, $($arg:tt)*) => {
        $crate::error::Error {
            context: None,
            error: $crate::error::ErrorType::Any { code : $err_code, msg: format!($($arg)*)}
        }
    }
}

/// Shorthand for creating errors caused by user interaction
#[macro_export]
macro_rules! bad_request {
    ($($arg:tt)*) => {
        $crate::error::Error {
            context: None,
            error: $crate::error::ErrorType::BadRequest(format!($($arg)*))
        }
    }
}

/// Shorthand for creating errors caused unauthorized access
#[macro_export]
macro_rules! forbidden {
    ($($arg:tt)*) => {
        $crate::error::Error {
            context: None,
            error: $crate::error::ErrorType::Unauthorized(format!($($arg)*))
        }
    }
}

/// Shorthand for creating errors that are caused by POD invalid internal state
#[macro_export]
macro_rules! internal_error {
    ($($arg:tt)*) => {
        $crate::error::Error {
            context: None,
            error: $crate::error::ErrorType::Internal(format!($($arg)*))
        }
    }
}

/// This is a opaque type for ErrorType that allows include context.
/// If this idea breaks in the future, or becomes too complex,
/// use anyhow::Result<T, ErrorType>, for now It's not done, simply
/// to avoid yet another crate in the dependency.
/// Anyhow might give also other nice features.
#[derive(Error, Debug)]
pub struct Error {
    pub context: Option<String>,
    pub error: ErrorType,
    // NOTE: might want use #[backtrace] in the future, now it's unstable.
    // Note, it's possible to use backtrace-rs on stable, it returns pretty
    // decent trace even in async env (tested on debug)
    // to make Backtrace not recognized by thiserror, wrap it in the box:
    // bt: Box<backtrace::Backtrace>
}

impl Error {
    pub fn inner(&self) -> &ErrorType {
        &self.error
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.error)?;

        if let Some(context) = &self.context {
            write!(f, ", context: {context}")?;
        }

        if let Some(source) = self.error.source() {
            write!(f, ", caused by: {source}")?;
        }

        Ok(())
    }
}

/// This is beautiful rust trait magic. ErrorType implements From conversion for
/// many types, like serde_json::Error, rusqlite::Error, etc, name it T.
/// This trait adds conversion From<T> for opaque type Error,
/// for any type TypeError introduces, now or in the future.
/// No need to write the same thing twice,
/// no need to maintain dozens of From impls for two types.
///
/// The trait says, implement conversion from T to Error for all
/// T's that implements conversion to ErrorType
impl<T> From<T> for Error
where
    T: Into<ErrorType>,
{
    fn from(value: T) -> Self {
        Self {
            context: None,
            error: value.into(),
        }
    }
}

/// Used in tests
impl PartialEq for Error {
    fn eq(&self, other: &Self) -> bool {
        format!("{self}") == format!("{other}")
    }
}

/// Result type used across POD
pub type Result<T> = std::result::Result<T, Error>;

#[derive(Error, Debug)]
pub enum ErrorType {
    #[error("Error in symmetric encryption cipher")]
    AeadEncryption(#[from] chacha20poly1305::aead::Error),

    #[error("Database rusqlite error")]
    Database(#[source] rusqlite::Error),

    #[error("FFI null conversion error")]
    FfiNul(#[from] std::ffi::NulError),

    #[error("{0}")]
    Unauthorized(String),

    #[error("Invalid database key")]
    UnauthorizedDatabaseAccess(#[source] rusqlite::Error),

    #[error("failed to find plugin with id {0:?}")]
    PluginNotFound(PluginIdentifier),

    /// Something is wrong with request provided by the user
    #[error("{0}")]
    BadRequest(String),

    /// POD is in bad internal state
    #[error("{0}")]
    Internal(String),

    /// Error caused by 3rd party lib, hard to decide who's fault was it
    #[error("{msg}")]
    Other {
        msg: String,
        source: Box<dyn std::error::Error + Send + 'static>,
    },

    /// Escape hatch if returned error has to have some uncommon status code
    #[error("{msg}")]
    Any { code: StatusCode, msg: String },
}

macro_rules! convert_to_other_error {
    ($err_type: ty, $msg: expr) => {
        impl From<$err_type> for ErrorType {
            fn from(err: $err_type) -> Self {
                ErrorType::Other {
                    msg: $msg.to_string(),
                    source: Box::new(err),
                }
            }
        }
    };
}

// Errors that are so rare in codebase are not worth to add to the enum
convert_to_other_error!(hex::FromHexError, "Error converting from hex");
convert_to_other_error!(refinery::Error, "Database 'refinery' migration error");
convert_to_other_error!(tokio::task::JoinError, "Async task join error");
convert_to_other_error!(std::num::TryFromIntError, "Parsing integer error");
convert_to_other_error!(std::env::VarError, "Env var error");
convert_to_other_error!(std::io::Error, "IO error");
convert_to_other_error!(std::str::Utf8Error, "UTF8 parse error");
convert_to_other_error!(reqwest::Error, "Reqwest sending error");
convert_to_other_error!(r2d2::Error, "Connection pool error");
convert_to_other_error!(oauth_client::Error, "Oauth error");
convert_to_other_error!(oauth2::url::ParseError, "Invalid URL address");
convert_to_other_error!(serde_json::Error, "JSON formatting error");
convert_to_other_error!(lettre::address::AddressError, "Email address error");
convert_to_other_error!(lettre::error::Error, "Email error");
convert_to_other_error!(lettre::transport::smtp::Error, "Email sending error");

/// serde_path_to_error is used for request body parsing, if something is wrong
/// that's clearly bad request to be blamed
impl From<serde_path_to_error::Error<serde_json::Error>> for ErrorType {
    fn from(value: serde_path_to_error::Error<serde_json::Error>) -> Self {
        ErrorType::BadRequest(format!("JSON deserialization error, {value}"))
    }
}

/// Rusqlite returns 'file is not a database' if decryption failed.
/// We return UnauthorizedDatabaseAccess in such case, otherwise
/// return generic 'Database' error.
impl From<rusqlite::Error> for ErrorType {
    fn from(err: rusqlite::Error) -> Self {
        if err.to_string().contains("file is not a database") {
            ErrorType::UnauthorizedDatabaseAccess(err)
        } else {
            ErrorType::Database(err)
        }
    }
}

impl<RE, T> From<oauth2::RequestTokenError<RE, T>> for ErrorType
where
    RE: std::error::Error + Send + 'static,
    T: oauth2::ErrorResponse + Send + Display + 'static,
{
    fn from(err: oauth2::RequestTokenError<RE, T>) -> Self {
        // Display implementation of internal type gives additional information
        let details = match &err {
            oauth2::RequestTokenError::ServerResponse(e) => format!("{e}"),
            oauth2::RequestTokenError::Request(e) => format!("{e}"),
            oauth2::RequestTokenError::Parse(e, _vec) => format!("{e}"),
            oauth2::RequestTokenError::Other(e) => e.to_string(),
        };

        ErrorType::Other {
            msg: format!("Request token error, details {details}"),
            source: Box::new(err),
        }
    }
}

pub trait ErrorContext<T> {
    fn context<F>(self, context_add: F) -> Result<T>
    where
        F: FnOnce() -> String;
    fn context_str(self, context_add: &str) -> Result<T>;
}

impl<T, E> ErrorContext<T> for std::result::Result<T, E>
where
    E: Into<Error>,
{
    fn context<F>(self, context_add: F) -> Result<T>
    where
        F: FnOnce() -> String,
    {
        match self {
            Ok(t) => Ok(t),
            Err(err) => {
                let mut err: Error = err.into();

                if let Some(ref mut context) = err.context {
                    context.extend([", ", &context_add()]);
                } else {
                    err.context = Some(context_add());
                }

                Err(err)
            }
        }
    }
    fn context_str(self, context_add: &str) -> Result<T> {
        match self {
            Ok(t) => Ok(t),
            Err(err) => {
                let mut err: Error = err.into();

                if let Some(ref mut context) = err.context {
                    context.extend([", ", context_add]);
                } else {
                    err.context = Some(context_add.to_string());
                }

                Err(err)
            }
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn plain_error_shows_basic_info() {
        let f1 = || -> Result<()> { Err(bad_request!("The unfortunate happened"))? };

        let res = f1();

        assert_eq!(&format!("{}", res.unwrap_err()), "The unfortunate happened");
    }

    #[test]
    fn error_shows_context() {
        let f1 = || -> Result<()> { Err(hex::FromHexError::OddLength)? };

        let res = f1()
            .context_str("context 1")
            .context(|| format!("context {}", 2));

        assert_eq!(
            &format!("{}", res.unwrap_err()),
            "Error converting from hex, context: context 1, context 2, caused by: Odd number of digits");
    }

    #[test]
    fn error_with_source_but_no_context() {
        let f1 = || -> Result<()> { Err(hex::FromHexError::OddLength)? };

        let res = f1();

        assert_eq!(
            &format!("{}", res.unwrap_err()),
            "Error converting from hex, caused by: Odd number of digits"
        );
    }

    #[test]
    fn error_with_context_but_no_source() {
        let f1 = || -> Result<()> { Err(bad_request!("Ups!"))? };

        let res = f1().context_str("context 1");

        assert_eq!(&format!("{}", res.unwrap_err()), "Ups!, context: context 1");
    }
}
