use crate::{
    api_model::{
        AuthKey, Bulk, ClientAuth, CreateEdge, CreateItem, CreateUserReq, DeleteEdgeBySourceTarget,
        GetEdges, GetFile, Oauth2AccessTokenRequest, Oauth2AuthUrlRequest,
        Oauth2AuthorizeTokenRequest, OauthAccessTokenPayload, OauthRequestTokenPayload, PodOwner,
        Search, SendEmail, UpdateItem,
    },
    async_db_connection::AsyncTx,
    command_line_interface, database_api,
    database_pool::{self, ConnectionPool, InitDb},
    database_utils, email,
    error::Result,
    file_api, internal_api, oauth1_api, oauth2_api, plugin_auth_crypto, triggers,
};
use async_ffi_macros::async_ffi;
use std::{
    collections::HashMap,
    ffi::{c_char, CStr, CString},
};

/// # Safety
/// The bulk_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn bulk(
    bulk_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let bulk = CStr::from_ptr(bulk_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = bulk_wrapper(&bulk, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn bulk_wrapper(bulk: &str, database_key: &str, owner_key: &str) -> Result<CString> {
    let bulk: Bulk = serde_json::from_str(bulk)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());
    let cli = &command_line_interface::PARSED.clone();

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = {
        let owner = owner.clone();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx)?;
            internal_api::bulk_tx(&tx, &mut schema, bulk, cli, &owner).await
        })
        .await?
    };

    triggers::actions_after_item_create(
        &mut conn,
        &result.create_items,
        &owner,
        cli,
        &database_key,
    )
    .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The create_user_req_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn create_account(create_user_req_ptr: *const c_char) -> *const c_char {
    let create_user_req = CStr::from_ptr(create_user_req_ptr)
        .to_string_lossy()
        .into_owned();

    let result = create_account_wrapper(&create_user_req).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn create_account_wrapper(create_user_req: &str) -> Result<CString> {
    let create_user_req: CreateUserReq = serde_json::from_str(create_user_req)?;

    database_utils::check_owner(&create_user_req.owner_key)?;

    let database_key = plugin_auth_crypto::auth_to_database_key(AuthKey::ClientAuth(ClientAuth {
        database_key: create_user_req.database_key,
    }))?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    database_pool::initialize_db(&create_user_req.owner_key, &init_db, &database_key).await?;

    let result = serde_json::to_string(&create_user_req.owner_key)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The create_edge_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn create_edge(
    create_edge_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let create_edge = CStr::from_ptr(create_edge_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = create_edge_wrapper(&create_edge, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn create_edge_wrapper(
    create_edge: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let create_edge: CreateEdge = serde_json::from_str(create_edge)?;
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx)?;

            internal_api::create_edge(&tx, &mut schema, create_edge)
        })
        .await?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The create_item_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn create_item(
    create_item_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let create_item = CStr::from_ptr(create_item_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = create_item_wrapper(&create_item, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn create_item_wrapper(
    create_item: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let create_item: CreateItem = serde_json::from_str(create_item)?;
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());
    let cli = &command_line_interface::PARSED.clone();

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result =
        internal_api::create_item(create_item, &mut conn, &owner, &database_key, cli).await?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The delete_edge_by_source_target_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn delete_edge(
    delete_edge_by_source_target_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let delete_edge_by_source_target = CStr::from_ptr(delete_edge_by_source_target_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result =
        delete_edge_wrapper(&delete_edge_by_source_target, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn delete_edge_wrapper(
    delete_edge_by_source_target: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let delete_edge_by_source_target: DeleteEdgeBySourceTarget =
        serde_json::from_str(delete_edge_by_source_target)?;
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_write_transaction(|tx: AsyncTx| async move {
            internal_api::delete_edge_by_source_target(&tx, delete_edge_by_source_target)
        })
        .await
        .map(|()| "{}")?
        .to_string();

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The item_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn delete_item(
    item_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let item = CStr::from_ptr(item_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = delete_item_wrapper(&item, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn delete_item_wrapper(item: &str, database_key: &str, owner_key: &str) -> Result<CString> {
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_read_transaction(
            |tx: AsyncTx| async move { internal_api::delete_item_by_id(&tx, item) },
        )
        .await
        .map(|()| "{}")?
        .to_string();

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn delete_user(
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = delete_user_wrapper(&database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn delete_user_wrapper(database_key: &str, owner_key: &str) -> Result<CString> {
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    database_utils::check_owner_and_delete_db(&owner, &init_db, &database_key).await?;

    let result = "{}".to_string();

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The get_edges_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn get_edges(
    get_edges_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let get_edges = CStr::from_ptr(get_edges_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = get_edges_wrapper(&get_edges, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn get_edges_wrapper(
    get_edges: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let get_edges: GetEdges = serde_json::from_str(get_edges)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_read_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            internal_api::get_edges(&tx, get_edges, &schema)
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The get_file_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn get_file(
    get_file_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let get_file = CStr::from_ptr(get_file_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = get_file_wrapper(&get_file, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn get_file_wrapper(get_file: &str, database_key: &str, owner_key: &str) -> Result<CString> {
    let get_file: GetFile = serde_json::from_str(get_file)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_read_transaction(|tx: AsyncTx| async move {
            file_api::get_file(&tx, &owner, &get_file.sha256)
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The item_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn get_item(
    item_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let item = CStr::from_ptr(item_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = get_item_wrapper(&item, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn get_item_wrapper(item: &str, database_key: &str, owner_key: &str) -> Result<CString> {
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_read_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            internal_api::get_item_tx(&tx, &schema, item)
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The item_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn graphql(
    item_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let item = CStr::from_ptr(item_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = graphql_wrapper(&item, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn graphql_wrapper(item: &str, database_key: &str, owner_key: &str) -> Result<CString> {
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_read_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            internal_api::graphql(&tx, &schema, item.to_string())
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The oauth_access_token_payload_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn oauth1_access_token(
    oauth_access_token_payload_ptr: *const c_char,
) -> *const c_char {
    let oauth_access_token_payload = CStr::from_ptr(oauth_access_token_payload_ptr)
        .to_string_lossy()
        .into_owned();

    let result = oauth1_access_token_wrapper(&oauth_access_token_payload).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn oauth1_access_token_wrapper(oauth_access_token_payload: &str) -> Result<CString> {
    let oauth_access_token_payload: OauthAccessTokenPayload =
        serde_json::from_str(oauth_access_token_payload)?;

    let result = oauth1_api::oauth1_access_token(
        oauth_access_token_payload.oauth_token,
        oauth_access_token_payload.oauth_verifier,
        oauth_access_token_payload.oauth_token_secret,
    )
    .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The oauth_request_token_payload_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn oauth1_request_token(
    oauth_request_token_payload_ptr: *const c_char,
) -> *const c_char {
    let oauth_request_token_payload = CStr::from_ptr(oauth_request_token_payload_ptr)
        .to_string_lossy()
        .into_owned();

    let result = oauth1_request_token_wrapper(&oauth_request_token_payload).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn oauth1_request_token_wrapper(oauth_request_token_payload: &str) -> Result<CString> {
    let oauth_request_token_payload: OauthRequestTokenPayload =
        serde_json::from_str(oauth_request_token_payload)?;

    let result = oauth1_api::oauth1_request_token(
        oauth_request_token_payload.callback_url,
        oauth_request_token_payload.service,
    )
    .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The oauth2_access_token_request_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn oauth2_access_token(
    oauth2_access_token_request_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let oauth2_access_token_request = CStr::from_ptr(oauth2_access_token_request_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result =
        oauth2_access_token_wrapper(&oauth2_access_token_request, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn oauth2_access_token_wrapper(
    oauth2_access_token_request: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let oauth2_access_token_request: Oauth2AccessTokenRequest =
        serde_json::from_str(oauth2_access_token_request)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_write_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            oauth2_api::access_token(&tx, &schema, oauth2_access_token_request).await
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The oauth2_auth_url_request_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn oauth2_auth_url(
    oauth2_auth_url_request_ptr: *const c_char,
) -> *const c_char {
    let oauth2_auth_url_request = CStr::from_ptr(oauth2_auth_url_request_ptr)
        .to_string_lossy()
        .into_owned();

    let result = oauth2_auth_url_wrapper(&oauth2_auth_url_request).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn oauth2_auth_url_wrapper(oauth2_auth_url_request: &str) -> Result<CString> {
    let oauth2_auth_url_request: Oauth2AuthUrlRequest =
        serde_json::from_str(oauth2_auth_url_request)?;

    let result = oauth2_api::auth_url(oauth2_auth_url_request)?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The oauth2_authorize_token_request_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn oauth2_authorize(
    oauth2_authorize_token_request_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let oauth2_authorize_token_request = CStr::from_ptr(oauth2_authorize_token_request_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result =
        oauth2_authorize_wrapper(&oauth2_authorize_token_request, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn oauth2_authorize_wrapper(
    oauth2_authorize_token_request: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let oauth2_authorize_token_request: Oauth2AuthorizeTokenRequest =
        serde_json::from_str(oauth2_authorize_token_request)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());
    let cli = &command_line_interface::PARSED.clone();

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx)?;

            oauth2_api::authorize(
                &tx,
                &mut schema,
                &owner,
                cli,
                oauth2_authorize_token_request,
            )
            .await
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The search_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn search(
    search_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let search = CStr::from_ptr(search_ptr).to_string_lossy().into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = search_wrapper(&search, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn search_wrapper(search: &str, database_key: &str, owner_key: &str) -> Result<CString> {
    let search: Search = serde_json::from_str(search)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_read_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            internal_api::search(&tx, &schema, search)
        })
        .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The send_email_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn send_email(
    send_email_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let send_email = CStr::from_ptr(send_email_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = send_email_wrapper(&send_email, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn send_email_wrapper(
    send_email: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let send_email: SendEmail = serde_json::from_str(send_email)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());
    let cli = &command_line_interface::PARSED.clone();

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let _conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let cli = cli.clone();
    let result = tokio::task::spawn_blocking(move || email::send_email(send_email, &cli))
        .await?
        .map(|()| "{}")?
        .to_string();

    let result = CString::new(&*result)?;

    Ok(result)
}

/// # Safety
/// The update_item_ptr must point to a valid String.
/// The database_key_ptr must point to a valid String.
/// The owner_key_ptr must point to a valid String.
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn update_item(
    update_item_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let update_item = CStr::from_ptr(update_item_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = update_item_wrapper(&update_item, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn update_item_wrapper(
    update_item: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let update_item: UpdateItem = serde_json::from_str(update_item)?;

    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn
        .in_write_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            internal_api::update_item_tx(&tx, &schema, &update_item.id, update_item.fields)
        })
        .await
        .map(|()| "{}")?
        .to_string();

    let result = CString::new(&*result)?;

    Ok(result)
}

/*
#[async_ffi(?Send)]
#[no_mangle]
pub async unsafe extern "C" fn upload_file(
    file_body: &'static [u8],
    expected_sha256_ptr: *const c_char,
    database_key_ptr: *const c_char,
    owner_key_ptr: *const c_char,
) -> *const c_char {
    let expected_sha256 = CStr::from_ptr(expected_sha256_ptr)
        .to_string_lossy()
        .into_owned();
    let database_key = CStr::from_ptr(database_key_ptr)
        .to_string_lossy()
        .into_owned();
    let owner_key = CStr::from_ptr(owner_key_ptr).to_string_lossy().into_owned();

    let result = upload_file_wrapper(file_body, &expected_sha256, &database_key, &owner_key).await;

    match result {
        Err(_) => CString::default().as_ptr(),
        Ok(result) => result.as_ptr(),
    }
}

async fn upload_file_wrapper(
    file_body: &[u8],
    expected_sha256: &str,
    database_key: &str,
    owner_key: &str,
) -> Result<CString> {
    let client_auth = ClientAuth {
        database_key: database_key.to_string(),
    };
    let auth_key = AuthKey::ClientAuth(client_auth);
    let owner = PodOwner::from(owner_key.to_string());

    let database_key = plugin_auth_crypto::auth_to_database_key(auth_key)?;

    let hash_map: HashMap<String, ConnectionPool> = HashMap::new();
    let init_db = InitDb::new(hash_map);

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, &init_db, &database_key).await?;

    let result = conn.in_write_transaction(|tx: AsyncTx| async move {
        file_api::upload_file(&tx, &owner, expected_sha256, file_body)
    })
    .await?;

    let result = serde_json::to_string(&result)?;

    let result = CString::new(&*result)?;

    Ok(result)
}
*/
