use crate::async_db_connection::AsyncTx;
use crate::{any_error, bad_request, constants, database_api, error::Result, internal_error};
use chacha20poly1305::{
    aead::{Aead, NewAead},
    Key, XChaCha20Poly1305, XNonce,
};
use http::status::StatusCode;
use rand::random;
use sha2::{Digest, Sha256};
use std::{
    fs::{create_dir_all, OpenOptions},
    io::Write,
    ops::Deref,
    path::PathBuf,
    str::FromStr,
};
use tracing::warn;

pub fn upload_file(tx: &AsyncTx, owner: &str, expected_sha256: &str, body: &[u8]) -> Result<()> {
    if file_exists_on_disk(owner, expected_sha256)? {
        // Note that checking once for file existence here is not enough.
        // To prevent TOCTOU attack, we also need to check file existence below.
        // We could avoid doing a check here at all, but we do it to avoid spending CPU power
        // on hash calculation for conflicting files,
        // which are detected by a relatively cheap test.
        return Err(any_error! {
            StatusCode::CONFLICT,
            "File already exists",
        });
    };
    validate_hash(expected_sha256, body)?;

    let key: [u8; 32] = random();
    let key = Key::from_slice(&key);
    let cipher = XChaCha20Poly1305::new(key);

    let nonce: [u8; 24] = rand::random();
    let nonce = XNonce::from_slice(&nonce); // unique per file
    let body = cipher.encrypt(nonce, body)?;
    update_key_and_nonce(tx, key.deref(), nonce.deref(), expected_sha256)?;

    let file = final_path(owner, expected_sha256)?;
    let file = OpenOptions::new().write(true).create_new(true).open(file);
    let mut file = file.map_err(|err| {
        if err.raw_os_error() == Some(libc::EEXIST) {
            any_error! {
                StatusCode::CONFLICT,
                "File already exists",
            }
        } else {
            internal_error! { "Failed to create target file, {err}" }
        }
    })?;
    file.write_all(&body).map_err(|err| {
        internal_error! { "Failed to write data to target file, {err}" }
    })?;
    Ok(())
}

pub fn get_file(tx: &AsyncTx, owner: &str, sha256: &str) -> Result<Vec<u8>> {
    let file = final_path(owner, sha256)?;
    let file = std::fs::read(file).map_err(|err| {
        internal_error! { "Failed to read data from target file, {err}" }
    })?;
    let (key, nonce) = find_key_and_nonce_by_sha256(tx, sha256)?;
    let (key, nonce) = (Key::from_slice(&key), XNonce::from_slice(&nonce));
    let cipher = XChaCha20Poly1305::new(key);
    let plaintext = cipher.decrypt(nonce, file.as_ref())?;
    Ok(plaintext)
}

fn file_exists_on_disk(owner: &str, sha256: &str) -> Result<bool> {
    let file = final_path(owner, sha256)?;
    Ok(file.exists())
}

fn final_owner_dir(owner: &str) -> Result<PathBuf> {
    let result = files_dir()?;
    let final_dir = result.join(owner).join(constants::FILES_FINAL_SUBDIR);
    create_dir_all(&final_dir).map_err(|err| {
        internal_error! { "Failed to create files directory for owner {owner}, {err}" }
    })?;
    Ok(final_dir)
}

fn final_path(owner: &str, sha256: &str) -> Result<PathBuf> {
    let owner_dir = final_owner_dir(owner)?;
    Ok(owner_dir.join(sha256))
}

pub async fn remove_owner_files(owner: &str) -> Result<()> {
    let owner_dir = final_owner_dir(owner)?;
    tokio::fs::remove_dir_all(owner_dir).await?;
    Ok(())
}

fn validate_hash(expected_sha256: &str, data: &[u8]) -> Result<()> {
    let expected_sha256_vec = hex::decode(expected_sha256)?;
    let mut real_sha256 = Sha256::new();
    real_sha256.update(data);
    let real_sha256 = real_sha256.finalize();
    if real_sha256.as_slice() != expected_sha256_vec.as_slice() {
        Err(bad_request! {
            "Expected file sha256 hash differs from calculated {}",
            hex::encode(real_sha256)
        })
    } else {
        Ok(())
    }
}

/// Update `key` and `nonce` in DB for items that have the given `sha256`
fn update_key_and_nonce(tx: &AsyncTx, key: &[u8], nonce: &[u8], for_sha256: &str) -> Result<()> {
    let item_rowids = database_api::search_strings(tx, "sha256", for_sha256)?;
    if item_rowids.is_empty() {
        Err(any_error! {
            StatusCode::NOT_FOUND,
            "Item with sha256 {for_sha256} not found in database",
        })
    } else {
        let key = hex::encode(key);
        let nonce = hex::encode(nonce);
        for item in item_rowids {
            database_api::insert_string(tx, item, "key", &key)?;
            database_api::insert_string(tx, item, "nonce", &nonce)?;
        }
        Ok(())
    }
}

/// Find first `key` and `nonce` pair in the database for an item with the desired `sha256`
fn find_key_and_nonce_by_sha256(tx: &AsyncTx, sha256: &str) -> Result<(Vec<u8>, Vec<u8>)> {
    let item_rowids = database_api::search_strings(tx, "sha256", sha256)?;
    if let Some(rowid) = item_rowids.first() {
        let mut other_props = database_api::get_strings_for_item(tx, *rowid)?;
        let key = other_props.remove("key").ok_or_else(|| {
            internal_error! { "Item with required hash {sha256} found, but it does not have a 'key' property" }
        })?;

        let nonce = other_props.remove("nonce").ok_or_else(|| {
            internal_error! { "Item with required hash {sha256} found, but it does not have a 'nonce' property" }
        })?;

        let key = hex::decode(key)?;
        let nonce = hex::decode(nonce)?;
        Ok((key, nonce))
    } else {
        Err(any_error! {
            StatusCode::NOT_FOUND,
            "Item with sha256={sha256} not found",
        })
    }
}

/// Directory where files (e.g. media) are stored
fn files_dir() -> Result<PathBuf> {
    PathBuf::from_str(constants::FILES_DIR).map_err(|err| {
        warn!("Failed to create file upload path {}", constants::FILES_DIR);
        internal_error! { "Failed to create file upload path, {err}" }
    })
}

#[cfg(test)]
mod tests {
    use super::{files_dir, final_owner_dir, get_file, remove_owner_files, upload_file};
    use crate::{
        async_db_connection::AsyncTx, command_line_interface, database_api, error::Result,
        internal_api, test_helpers::get_memory_connection,
    };
    use serde_json::json;

    #[tokio::test]
    async fn test_file_upload_get() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx)?;
            let cli = command_line_interface::tests::test_cli();
            let owner = "testOwner".to_string();
            let owner_dir = files_dir()?.join(&owner);
            std::fs::remove_dir_all(&owner_dir).ok();
            let sha =
                "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855".to_string();

            let json = json!({
                "type": "File",
                "sha256": &sha,
            });
            let sha_item = serde_json::from_value(json)?;
            internal_api::create_item_tx(&tx, &mut schema, sha_item, &owner, &cli).await?;

            upload_file(&tx, &owner, &sha, &[])?;

            let result = get_file(&tx, &owner, &sha)?;
            assert_eq!(result.len(), 0, "{}:{}", file!(), line!());
            std::fs::remove_dir_all(owner_dir).ok();
            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_remove_files() -> Result<()> {
        let owner = "testOwner";
        let owner_dir = final_owner_dir(owner)?;
        assert!(owner_dir.exists());
        remove_owner_files(owner).await?;
        assert!(!owner_dir.exists());
        Ok(())
    }
}
