use crate::{
    any_error,
    api_model::{
        Bulk, BulkResponse, CreateEdge, CreateItem, DeleteEdgeBySourceTarget, EdgeDirection,
        GetEdges, Search, SortOrder,
    },
    async_db_connection::{AsyncConnection, AsyncTx},
    bad_request,
    command_line_interface::CliOptions,
    database_api::{self},
    database_api::{
        get_incoming_edges, get_outgoing_edges, search_items, DatabaseSearch, EdgePointer,
        GQLSearchArgs, Rowid,
    },
    database_utils::{add_item_edge_properties, insert_edge, insert_property, item_base_to_json},
    error::{ErrorContext, Result},
    graphql_utils,
    graphql_utils::QueryASTNode,
    internal_error,
    plugin_auth_crypto::DatabaseKey,
    schema,
    schema::{validate_property_name, Schema},
    triggers,
    triggers::SchemaAdditionChange,
    triggers::{add_extra_fields, add_extra_fields_for_twitter},
};

use crate::async_db_connection::AsyncTx as Tx;
use http::StatusCode;
use rand::Rng;
use serde_json::{Map, Value};
use std::{collections::HashMap, str};
use tracing::{debug, info};

pub fn get_project_version() -> String {
    serde_json::json!({
        "cargo": std::env!("CARGO_PKG_VERSION"),
        "git_describe": std::env!("GIT_DESCRIBE"),
        // Env variable reexported from build script
        "build_profile": std::env!("PROFILE"),
    })
    .to_string()
}

pub fn get_item_tx(tx: &Tx, schema: &Schema, id: &str) -> Result<Vec<Value>> {
    debug!("Getting item {}", id);
    let search_query = Search {
        id: Some(id.to_string()),
        _type: None,
        date_server_modified_gte: None,
        date_server_modified_lt: None,
        deleted: None,
        sort_order: SortOrder::Asc,
        limit: 1,
        return_properties: vec![],
        forward_edges: None,
        backward_edges: None,
        other_properties: Default::default(),
    };
    let result = search(tx, schema, search_query)?;
    Ok(result)
}

/// Generate a new random item id.
/// This implementation chooses to generate 32 random hex characters.
pub fn new_random_item_id() -> String {
    new_random_string(32)
}
pub fn new_random_string(length: usize) -> String {
    let mut rng = rand::thread_rng();
    (0..length)
        .map(|_| {
            let idx = rng.gen_range(0..HEX_CHARACTERS.len());
            HEX_CHARACTERS[idx] as char
        })
        .collect()
}
const HEX_CHARACTERS: &[u8] = b"0123456789abcdef";

pub async fn create_item_tx(
    tx: &Tx,
    schema: &mut Schema,
    mut item: CreateItem,
    pod_owner: &str,
    cli: &CliOptions,
) -> Result<String> {
    let _ = schema
        .item(&item._type)
        .ok_or(bad_request!("No schema definition for '{}'", item._type))?;

    let id: String = if let Some(id) = &item.id {
        schema::validate_create_item_id(id)?;
        id.to_string()
    } else {
        let id = new_random_item_id();
        item.id = Some(id.clone());

        id
    };

    let schema_addition = triggers::add_item_as_schema_opt(tx, schema, &item)?;
    if let SchemaAdditionChange::OldSchema { old_id } = schema_addition {
        Ok(old_id)
    } else {
        let time_now = schema::utc_millis();

        if item.date_created.is_none() {
            item.date_created = Some(time_now);
        }

        if item.date_modified.is_none() {
            item.date_modified = Some(time_now);
        }

        if item.date_server_modified.is_none() {
            item.date_server_modified = Some(time_now)
        }

        let rowid = database_api::insert_item_base(
            tx,
            &id,
            &item._type,
            item.date_created.unwrap(),
            item.date_modified.unwrap(),
            time_now,
            item.deleted,
        )?;

        item.rowid = Some(rowid);
        add_extra_fields(&mut item, cli, pod_owner)?;
        add_extra_fields_for_twitter(tx, schema, &mut item).await?;

        for (prop_name, prop_value) in &item.fields {
            insert_property(tx, schema, rowid, &item._type, prop_name, prop_value)?;
        }

        Ok(id)
    }
}

pub fn update_item_tx(
    tx: &Tx,
    schema: &Schema,
    id: &str,
    fields: HashMap<String, Value>,
) -> Result<()> {
    _update_item_tx(tx, schema, id, fields, None)
}

pub fn _update_item_tx(
    tx: &Tx,
    schema: &Schema,
    id: &str,
    mut fields: HashMap<String, Value>,
    skip_from_readonly: Option<Vec<String>>,
) -> Result<()> {
    debug!("Updating item {}", id);
    fields.remove("type");
    fields.remove("dateCreated");

    let readonly_properties: Vec<String> = vec!["containerId".to_string()];

    // if property is readonly but also in skip_from_readonly: still update. This can only be done by the server
    // e.g. when setting the pluginrun.containerId
    for readonly in readonly_properties.into_iter() {
        if let Some(_remove_from_readonly) = &skip_from_readonly {
            if !_remove_from_readonly.contains(&readonly) {
                fields.remove(&readonly);
            }
        } else {
            fields.remove(&readonly);
        }
    }

    let time_now = schema::utc_millis();
    let date_modified = if let Some(dm) = fields.remove("dateModified") {
        if let Some(dm) = dm.as_i64() {
            dm
        } else {
            return Err(bad_request! { "Cannot parse dateModified {dm} to i64" });
        }
    } else {
        time_now
    };
    fields.remove("dateServerModified");
    let deleted = if let Some(d) = fields.remove("deleted") {
        match d {
            Value::Null => None,
            Value::Bool(d) => Some(d),
            _ => return Err(bad_request! { "Failed to parse deleted {d} to bool" }),
        }
    } else {
        None
    };
    for k in fields.keys() {
        validate_property_name(k)?;
    }

    let item = search_items(
        tx,
        &DatabaseSearch {
            id: Some(id),
            ..Default::default()
        },
    )?
    .pop()
    .ok_or_else(|| {
        any_error! {
            StatusCode::NOT_FOUND,
            "Item with id {id} not found",
        }
    })?;

    database_api::update_item_base(tx, item.rowid, date_modified, time_now, deleted)?;
    for (k, v) in &fields {
        insert_property(tx, schema, item.rowid, &item._type, k, v)?;
    }
    Ok(())
}

pub fn delete_item_by_id(tx: &Tx, id: &str) -> Result<()> {
    debug!("Deleting item {}", id);
    let rowid = database_api::get_item_rowid(tx, id)?.ok_or_else(|| {
        any_error! {
            StatusCode::NOT_FOUND,
            "Item with id {id} not found",
        }
    })?;
    delete_item_by_rowid(tx, rowid)
}

pub fn delete_item_by_rowid(tx: &Tx, rowid: Rowid) -> Result<()> {
    let time_now = schema::utc_millis();
    database_api::update_item_base(tx, rowid, time_now, time_now, Some(true))
}

pub async fn bulk_tx(
    tx: &Tx,
    schema: &mut Schema,
    bulk: Bulk,
    cli: &CliOptions,
    pod_owner: &str,
) -> Result<BulkResponse> {
    info!(
        "Performing bulk action with {} new items, {} updated items, {} deleted items, {} created edges",
        bulk.create_items.len(),
        bulk.update_items.len(),
        bulk.delete_items.len(),
        bulk.create_edges.len(),
    );
    let mut created_items = Vec::new();
    for (idx, item) in bulk.create_items.into_iter().enumerate() {
        let id = create_item_tx(tx, schema, item, pod_owner, cli)
            .await
            .context(|| format!("While creating the item #{idx}"))?;
        created_items.push(id);
    }
    for (idx, item) in bulk.update_items.into_iter().enumerate() {
        update_item_tx(tx, schema, &item.id, item.fields)
            .context(|| format!("While updating the item #{idx}"))?;
    }
    for (idx, item_id) in bulk.delete_items.into_iter().enumerate() {
        delete_item_by_id(tx, &item_id).context(|| format!("While deleting the item #{idx}"))?;
    }
    let mut created_edges = Vec::new();
    for (idx, edge) in bulk.create_edges.into_iter().enumerate() {
        let id =
            create_edge(tx, schema, edge).context(|| format!("While creating the edge #{idx}"))?;
        created_edges.push(id);
    }
    let mut search_results = Vec::new();
    for (idx, query) in bulk.search.into_iter().enumerate() {
        let result =
            search(tx, schema, query).context(|| format!("While making the search #{idx}"))?;
        search_results.push(result);
    }

    Ok(BulkResponse {
        create_items: created_items,
        create_edges: created_edges,
        search: search_results,
    })
}

pub fn create_edge(tx: &Tx, schema: &mut Schema, edge: CreateEdge) -> Result<String> {
    let CreateEdge {
        source,
        target,
        name,
        self_id,
    } = edge;
    let date = schema::utc_millis();

    let (self_rowid, self_id) = if let Some(id) = self_id {
        let self_rowid = database_api::get_item_rowid(tx, &id)?.ok_or_else(|| {
            bad_request! { "Failed to create edge with _self: {id}" }
        })?;

        (self_rowid, id)
    } else {
        let self_id = new_random_item_id();
        let self_rowid =
            database_api::insert_item_base(tx, &self_id, "Edge", date, date, date, false)?;
        (self_rowid, self_id)
    };

    let source: Rowid = database_api::get_item_rowid(tx, &source)?.ok_or_else(|| {
        any_error! {
            StatusCode::NOT_FOUND,
            "Edge source not found: {source}",
        }
    })?;
    let target: Rowid = database_api::get_item_rowid(tx, &target)?.ok_or_else(|| {
        any_error! {
            StatusCode::NOT_FOUND,
            "Edge target not found: {target}",
        }
    })?;
    insert_edge(tx, schema, self_rowid, source, &name, target)?;
    // database_api::update_item_date_server_modified(tx, source, date)?;
    Ok(self_id)
}

pub fn delete_edge_by_source_target(tx: &Tx, query: DeleteEdgeBySourceTarget) -> Result<()> {
    debug!(
        "Deleting edge with source {}, target {}",
        query.source, query.target
    );
    let source = database_api::get_item_rowid(tx, &query.source)?.ok_or_else(|| {
        bad_request! { "Edge deletion's _source item ID not found {}", &query.source }
    })?;

    let target = database_api::get_item_rowid(tx, &query.target)?.ok_or_else(|| {
        bad_request! { "Edge deletion's _target item ID not found {}", &query.target }
    })?;

    let self_rowid = database_api::get_edge_self_by_source_target(tx, source, target)?;
    let self_rowid = if let Some(self_rowid) = self_rowid {
        self_rowid
    } else {
        return Err(any_error! {
            StatusCode::NOT_FOUND,
            "Cannot delete edge, edge with source {} ({}) and target {} ({}) not found",
            query.source, source, query.target, target
        });
    };
    database_api::delete_edge(tx, self_rowid)?;
    delete_item_by_rowid(tx, self_rowid)?;
    Ok(())
}

pub fn edge_pointer_to_json(tx: &Tx, schema: &Schema, edge: &EdgePointer) -> Result<Value> {
    let target = database_api::get_item_base(tx, edge.item)?;
    let target = target.ok_or_else(|| {
        internal_error! { "Edge connects to an nonexisting item.rowid {}", edge.item }
    })?;
    let target_json = Value::Object(item_base_to_json(tx, target, schema)?);
    let edge_item = database_api::get_item_base(tx, edge.rowid)?.ok_or_else(|| {
        internal_error! { "Edge does not have an item base, rowid: {}", edge.item }
    })?;
    let mut edge_item = item_base_to_json(tx, edge_item, schema)?;
    edge_item.insert("_edge".to_string(), serde_json::json!(edge.name));
    edge_item.insert("_item".to_string(), serde_json::json!(target_json));
    Ok(serde_json::json!(edge_item))
}
pub fn edge_pointers_to_json(
    tx: &Tx,
    schema: &Schema,
    half_edges: &[EdgePointer],
) -> Result<Vec<Value>> {
    let mut result = Vec::new();
    for edge in half_edges {
        result.push(edge_pointer_to_json(tx, schema, edge)?);
    }
    Ok(result)
}

pub fn get_edges(tx: &Tx, query: GetEdges, schema: &Schema) -> Result<Vec<Value>> {
    let root_item = database_api::get_item_rowid(tx, &query.item)?.ok_or_else(|| {
        any_error! {
            StatusCode::NOT_FOUND,
            "Cannot find item id {}", query.item
        }
    })?;
    let edges = if query.direction == EdgeDirection::Outgoing {
        database_api::get_outgoing_edges(tx, root_item)?
    } else {
        database_api::get_incoming_edges(tx, root_item)?
    };
    let mut result = Vec::new();
    for edge in edges {
        let base = database_api::get_item_base(tx, edge.item)?;
        let base = base.ok_or_else(|| {
            internal_error! { "Edge connects to an nonexisting item.rowid {}", edge.item }
        })?;
        if query.expand_items {
            let item_json = Value::Object(item_base_to_json(tx, base, schema)?);
            result.push(serde_json::json!({
                "name": edge.name,
                "item": item_json,
            }));
        } else {
            result.push(serde_json::json!({
                "name": edge.name,
                "item": { "id": base.id },
            }));
        }
    }
    Ok(result)
}

pub fn search(tx: &Tx, schema: &Schema, query: Search) -> Result<Vec<Value>> {
    debug!("Searching by fields {:?}", query);
    if !query.other_properties.is_empty() {
        info!(
            "Doing a slow search request using non-base properties {:?}",
            query.other_properties.keys()
        );
    }
    let database_search = DatabaseSearch {
        rowid: None,
        id: query.id.as_deref(),
        _type: query._type.as_deref(),
        date_server_modified_gte: query.date_server_modified_gte,
        date_server_modified_lt: query.date_server_modified_lt,
        deleted: query.deleted,
        sort_order: query.sort_order.clone(),
        _limit: query.limit,
    };
    let items = database_api::search_items_with_properties(
        tx,
        &database_search,
        Some(schema),
        Some(&query.other_properties),
    )?;

    let mut result = Vec::new();
    for item in items {
        let rowid = item.rowid;
        let object_map = item_base_to_json(tx, item, schema)?;
        let mut object_map = if query.return_properties.is_empty() {
            object_map
        } else {
            object_map
                .into_iter()
                .filter(|(prop, _value)| query.return_properties.contains(prop))
                .collect()
        };
        add_item_edge_properties(tx, &mut object_map, rowid)?;
        if query.forward_edges.is_some() {
            let edges = get_outgoing_edges(tx, rowid)?;
            let edges = edge_pointers_to_json(tx, schema, &edges)?;
            object_map.insert("[[edges]]".to_string(), Value::Array(edges));
        }
        if query.backward_edges.is_some() {
            let edges = get_incoming_edges(tx, rowid)?;
            let edges = edge_pointers_to_json(tx, schema, &edges)?;
            object_map.insert("~[[edges]]".to_string(), Value::Array(edges));
        }
        result.push(Value::Object(object_map));
        // }
    }
    Ok(result)
}

fn gql_search_recursive(
    tx: &Tx,
    query_ast: &QueryASTNode,
    schema: &Schema,
    item_rowids: Option<Vec<Rowid>>,
) -> Result<Vec<Value>> {
    let sort_property = query_ast
        .arguments
        .sort_property
        .to_owned()
        .unwrap_or_else(|| "dateServerModified".to_owned());
    let sort_order = query_ast
        .arguments
        .sort_order
        .to_owned()
        .unwrap_or_default();

    let query = GQLSearchArgs {
        _type: query_ast.item_type.to_owned(),
        properties: query_ast.properties.to_owned(),
        rowids: item_rowids,
        sort_property,
        sort_order,
        limit: query_ast.arguments.limit,
        offset: query_ast.arguments.offset,
        filters: query_ast.arguments.filters.to_owned(),
    };

    // let now = Instant::now();
    let items = database_api::search_gql_properties(tx, &query, schema)?;

    let mut result = Vec::new();
    for item in items.into_iter() {
        let mut item_json = item.properties;
        if !query_ast.selectors.is_empty() {
            item_json = filter_map_by_selectors(item_json, &query_ast.selectors, schema)
        }
        for (edge_name, edge_item_ast) in &query_ast.edges {
            // let now_edges = Instant::now();
            let edges = if edge_name.contains('~') {
                let normalized = &edge_name[1..edge_name.len()];
                database_api::get_edges(tx, item.rowid, Some(&normalized.to_string()), true)?
            } else {
                database_api::get_edges(tx, item.rowid, Some(edge_name), false)?
            };
            let target_ids: Option<Vec<Rowid>> = Some(edges.iter().map(|e| e.target).collect());
            let edge_jsons = if let Some(_target_ids) = target_ids {
                if !_target_ids.is_empty() {
                    gql_search_recursive(tx, edge_item_ast, schema, Some(_target_ids))?
                } else {
                    Vec::new()
                }
            } else {
                Vec::new()
            };

            item_json.insert(edge_name.to_owned(), edge_jsons.into());
        }
        result.push(item_json.into());
    }
    Ok(result)
}

pub fn graphql(tx: &Tx, schema: &Schema, query: String) -> Result<HashMap<String, Vec<Value>>> {
    let query_ast: QueryASTNode = graphql_utils::parse_graphql_query(&query)?;
    let search_result = gql_search_recursive(tx, &query_ast, schema, None)?;
    let mut result = HashMap::new();
    result.insert("data".to_owned(), search_result);
    Ok(result)
}

pub fn filter_map_by_selectors(
    mut item: Map<String, Value>,
    selectors: &[QueryASTNode],
    schema: &Schema,
) -> Map<String, Value> {
    for selector in selectors.iter() {
        if item["type"].as_str().unwrap() == selector.item_type.as_ref().unwrap() {
            for prop_name in item.clone().keys() {
                if schema.is_base_property(prop_name.as_str()) {
                    continue;
                }
                if !selector.properties.contains(prop_name) {
                    item.remove(prop_name);
                }
            }
        }
    }
    item
}

// TODO: there is no proper way of starting plugins, it happens by creating the item in the DB
// Add dedicated endpoints for that
pub async fn create_item(
    payload: CreateItem,
    conn: &mut AsyncConnection,
    owner: &str,
    database_key: &DatabaseKey,
    cli: &CliOptions,
) -> Result<String> {
    // Commit the item to the DB
    let item_id = conn
        .in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx)?;

            create_item_tx(&tx, &mut schema, payload, owner, cli).await
        })
        .await;

    if let Ok(id) = &item_id {
        // After transaction commit, the item is visible to all readers, now it's safe to call triggers
        triggers::actions_after_item_create(conn, &[id], owner, cli, database_key).await?;
    }

    item_id
}

#[cfg(test)]
mod tests {
    use crate::{
        api_model::{Bulk, CreateItem},
        async_db_connection::AsyncTx,
        command_line_interface::tests::test_cli,

        command_line_interface::{self, CliOptions},
        database_api,
        // database_api::{tests::new_conn, self},
        error::{Error, ErrorType, Result},

        internal_api::{self, create_edge, delete_edge_by_source_target, graphql, search},
        schema::Schema,
        test_helpers::get_memory_connection,
    };
    use serde_json::{json, Value};

    use super::create_item_tx;

    #[tokio::test]
    async fn test_schema_checking() -> Result<()> {
        let mut conn = get_memory_connection()?;
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let cli = command_line_interface::tests::test_cli();

            // first try to insert the Person without Schema
            let item_json = json!({
                "type": "Person",
                "age": 20,
            });
            let item_struct: CreateItem = serde_json::from_value(item_json).unwrap();
            let result = internal_api::create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx)?,
                item_struct.clone(),
                "",
                &cli,
            )
            .await;

            assert!(matches!(
                result.unwrap_err(),
                Error {
                    context: _,
                    error: ErrorType::BadRequest(_)
                }
            ));

            // Then insert the Schema, it should succeed
            let schema_json = json!({
                "type": "ItemPropertySchema",
                "itemType": "Person",
                "propertyName": "age",
                "valueType": "Integer",
            });
            let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
            let result = internal_api::create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx)?,
                schema_struct,
                "",
                &cli,
            )
            .await;
            assert!(result.is_ok());

            // Now insert the Person with already in place
            let result = internal_api::create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx)?,
                item_struct,
                "",
                &cli,
            )
            .await;
            assert!(result.is_ok());

            Ok(())
        })
        .await
    }

    #[tokio::test]
    async fn test_item_insert_schema() {
        let mut conn = get_memory_connection().unwrap();
        let cli = command_line_interface::tests::test_cli();
        conn.in_write_transaction(|tx: AsyncTx| async move {
        let mut minimal_schema = database_api::get_schema(&tx).unwrap();

        let json = json!({
            "type": "ItemPropertySchema",
            "itemType": "Person",
            "propertyName": "age",
            "valueType": "Integer",
        });

        {
            // insert "age" successfully
            let create_item: CreateItem = serde_json::from_value(json.clone()).unwrap();
            internal_api::create_item_tx(&tx, &mut minimal_schema, create_item, "", &cli).await.unwrap();
        }

        {
            let json = json!({
                "type": "ItemPropertySchema",
                "itemType": "Person",
                "propertyName": "dateCreated",
                "valueType": "Bool",
            });
            let create_item: CreateItem = serde_json::from_value(json).unwrap();
            let expected_error = "Schema for property dateCreated is already defined to valueType DateTime, cannot override to Bool";
            let result =
                internal_api::create_item_tx(&tx, &mut minimal_schema, create_item, "", &cli).await;

            assert!(matches!(result.unwrap_err(),
                Error {
                    context: _,
                    error: ErrorType::BadRequest(msg)
                } if msg.contains(expected_error)));
        }

        let mut bad_empty_schema = Schema::empty();
        let create_item: CreateItem = serde_json::from_value(json).unwrap();
        let result =
            internal_api::create_item_tx(&tx, &mut bad_empty_schema, create_item, "", &cli)
                .await.unwrap_err();

        assert!(matches!(result,
        Error {
            context: _,
            error: ErrorType::BadRequest(msg)
        } if msg.contains("No schema definition for 'ItemPropertySchema'")));

        // tx.commit().unwrap();
        Ok(())
    }).await.unwrap()
    }

    #[tokio::test]
    async fn test_search() {
        let mut conn = get_memory_connection().unwrap();
        let cli = command_line_interface::tests::test_cli();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx).unwrap();
            create_item_tx(
                &tx,
                &mut schema,
                serde_json::from_value(json!({
                    "type": "ItemPropertySchema",
                    "itemType": "Person",
                    "propertyName": "age",
                    "valueType": "Integer",
                }))
                .unwrap(),
                "",
                &cli,
            )
            .await
            .unwrap();

            let person = {
                let person = json!({
                    "type": "Person",
                });
                let person: CreateItem = serde_json::from_value(person).unwrap();
                create_item_tx(&tx, &mut schema, person, "", &cli)
                    .await
                    .unwrap()
            };
            {
                let json = json!({
                    "id": person,
                });
                let parsed = serde_json::from_value(json).unwrap();
                let result = search(&tx, &schema, parsed).unwrap();
                assert_eq!(result.len(), 1);
                let result = result.into_iter().next().unwrap();
                let result = match result {
                    Value::Object(r) => r,
                    _ => panic!("Result is not JSON Object"),
                };
                assert!(result.len() > 3);
            };
            {
                let json = json!({
                    "id": person,
                    "_returnProperties": ["id"]
                });
                let parsed = serde_json::from_value(json).unwrap();
                let result = search(&tx, &schema, parsed).unwrap();
                assert_eq!(result.len(), 1);
                let result = result.into_iter().next().unwrap();
                let result = match result {
                    Value::Object(r) => r,
                    _ => panic!("Result is not JSON Object"),
                };
                assert_eq!(result.len(), 1);
                assert!(result.get("id").is_some());
            };

            Ok(())
        })
        .await
        .unwrap()
    }

    async fn prepare_graphql_dataset(tx: &AsyncTx, pod_owner: &str, cli: CliOptions) {
        let mut schema = database_api::get_schema(tx).unwrap();
        internal_api::bulk_tx(
            tx,
            &mut schema,
            Bulk {
                create_items: vec![
                    serde_json::from_value(json!({
                    "type": "ItemPropertySchema",
                    "itemType": "Person",
                    "propertyName": "age",
                    "valueType": "Integer"
                    }))
                    .unwrap(),
                    serde_json::from_value(json!({
                    "type": "ItemPropertySchema",
                    "itemType": "Person1",
                    "propertyName": "age",
                    "valueType": "Integer"
                    }))
                    .unwrap(),
                    serde_json::from_value(json!({
                    "type": "ItemPropertySchema",
                    "itemType": "Person2",
                    "propertyName": "age",
                    "valueType": "Integer"
                    }))
                    .unwrap(),
                    serde_json::from_value(json!({
                    "type": "ItemPropertySchema",
                    "itemType": "Person3",
                    "propertyName": "age",
                    "valueType": "Integer"
                    }))
                    .unwrap(),
                ],
                update_items: vec![],
                delete_items: vec![],
                create_edges: vec![],
                search: vec![],
            },
            &cli,
            pod_owner,
        )
        .await
        .unwrap();

        // create level 1
        for i in 0..1000 {
            let person = json!({
                "type": "Person1",
                "id": "00000".to_owned() + &i.to_string(),
                "age": i
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(tx, &mut schema, person, pod_owner, &cli)
                .await
                .unwrap();
        }

        // create level 2
        for i in 1000..2000 {
            let person = json!({
                "type": "Person2",
                "id": "00000".to_owned() + &i.to_string(),
                "age": i
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(tx, &mut schema, person, pod_owner, &cli)
                .await
                .unwrap();
        }

        // create level 3
        for i in 2000..12000 {
            let person = json!({
                "type": "Person3",
                "id": "00000".to_owned() + &i.to_string()
            });
            let person: CreateItem = serde_json::from_value(person).unwrap();
            create_item_tx(tx, &mut schema, person, pod_owner, &cli)
                .await
                .unwrap();
        }

        // create schema for connections
        let schema_json = json!({
            "type": "ItemEdgeSchema",
            "edgeName": "friend1",
            "sourceType": "Person1",
            "targetType": "Person2",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = create_item_tx(
            tx,
            &mut database_api::get_schema(tx).unwrap(),
            schema_struct,
            pod_owner,
            &cli,
        )
        .await;
        assert!(result.is_ok());

        let schema_json = json!({
            "type": "ItemEdgeSchema",
            "edgeName": "friend2",
            "sourceType": "Person2",
            "targetType": "Person3",
        });
        let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
        let result = create_item_tx(
            tx,
            &mut database_api::get_schema(tx).unwrap(),
            schema_struct,
            pod_owner,
            &cli,
        )
        .await;
        assert!(result.is_ok());

        // Get updated schema
        let mut schema = database_api::get_schema(tx).unwrap();

        // create edges level 1 -> level 2
        for i in 0..100 {
            let _edge = {
                let json = json!({
                    "_source": "00000".to_owned() + &i.to_string(),
                    "_target": "00000".to_owned() + &(i+1000).to_string(),
                    "_name": "friend1",
                });
                let parsed = serde_json::from_value(json).unwrap();
                create_edge(tx, &mut schema, parsed).unwrap()
            };
        }
        // create edges level 2 -> level 3
        for i in 0..1000 {
            let _edge = {
                let json = json!({
                    "_source": "00000".to_owned() + &((i/10)+1000).to_string(),
                    "_target": "00000".to_owned() + &(i+2000).to_string(),
                    "_name": "friend2",
                });
                let parsed = serde_json::from_value(json).unwrap();
                create_edge(tx, &mut schema, parsed).unwrap()
            };
        }
    }

    #[tokio::test]
    async fn test_graphql() {
        let mut conn = get_memory_connection().unwrap();

        // let mut conn = rusqlite::Connection::open("test.sqlite").unwrap();
        // database_migrate_refinery::embedded::migrations::runner()
        //     .run(&mut conn)
        //     .expect("Failed to run refinery migrations");

        conn.in_write_transaction(|tx: AsyncTx| async move {
            let cli = test_cli();
            let pod_owner = "";

            prepare_graphql_dataset(&tx, pod_owner, cli).await;

            let schema = database_api::get_schema(&tx).unwrap();

            let query = "
        query {
            Person1 {
               id
               age
               friend1 {
                   id
                   age
                   friend2 {
                       id
                       age
                   }
               }
            }
        }"
            .to_owned();

            let res = graphql(&tx, &schema, query);
            let data = res.unwrap();
            let data = data.get("data").unwrap();
            // assert_eq!(data.len(), 2000);

            let items_with_friends = data
                .iter()
                .filter_map(|x| x.get("friend1"))
                .filter(|x| !x.as_array().unwrap().is_empty());

            // print!("{:}", serde_json::to_string_pretty(&json!(data)).unwrap());
            assert_eq!(
                items_with_friends.clone().count(),
                100,
                "Expected 100 friends"
            );

            // check if 1st level prop is fetched
            let age: i32 = serde_json::from_value(data[101]["age"].clone()).unwrap_or(0);
            assert!(age > 0, "Cannot get root property: age");

            // // check 2nd level prop
            for friends in items_with_friends {
                let friend_age: i32 =
                    serde_json::from_value(friends[0]["age"].clone()).unwrap_or(0);
                assert!(friend_age > 0, "Cannot get edge item property: age");
            }

            // test limit and offset
            let query = "
        query {
            Person1 (limit: 10, offset: 5) {
               id
               friend1 {
                    id
                }
            }
        }"
            .to_owned();
            // print!("{:}", serde_json::to_string_pretty(&json!(data)).unwrap());

            let res = graphql(&tx, &schema, query);
            let data = res.unwrap();
            let data = data.get("data").unwrap();
            assert_eq!(data.len(), 10);
            Ok(())
        })
        .await
        .unwrap()
    }

    #[tokio::test]
    async fn test_graphql_filters() {
        let mut conn = get_memory_connection().unwrap();

        // let mut conn = rusqlite::Connection::open("test.sqlite").unwrap();
        // database_migrate_refinery::embedded::migrations::runner()
        //     .run(&mut conn)
        //     .expect("Failed to run refinery migrations");

        conn.in_write_transaction(|tx: AsyncTx| async move {
            let cli = test_cli();
            let pod_owner = "";

            prepare_graphql_dataset(&tx, pod_owner, cli).await;

            let schema = database_api::get_schema(&tx).unwrap();

            let query = "
        query {
            Person1 {
               id
               friend1 {
                   id
                   friend2 {
                       id
                   }
               }
            }
        }"
            .to_owned();

            let res = graphql(&tx, &schema, query);
            let data = res.unwrap();
            let data = data.get("data").unwrap();
            // assert_eq!(data.len(), 2000);

            let items_with_friends = data
                .iter()
                .filter_map(|x| x.get("friend1"))
                .filter(|x| !x.as_array().unwrap().is_empty());

            // print!("{:}", serde_json::to_string_pretty(&json!(data)).unwrap());
            assert_eq!(items_with_friends.count(), 100);

            let query = "
        query {
            Person1 (filter: {id: {ne: \"000000\"}}) {
               id
               friend1 {
                   id
               }
            }
        }"
            .to_owned();
            let res = graphql(&tx, &schema, query);
            let data = res.unwrap();
            let data = data.get("data").unwrap();
            assert_eq!(data.len(), 999);

            Ok(())
        })
        .await
        .unwrap()
    }

    #[tokio::test]
    async fn test_graphql_unions() {
        let mut conn = get_memory_connection().unwrap();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let cli = test_cli();
            let pod_owner = "";

            prepare_graphql_dataset(&tx, pod_owner, cli.clone()).await;

            // enable edgy unions: friend1 = Union[Person1, Person2]
            let schema_json = json!({
                "type": "ItemEdgeSchema",
                "edgeName": "friend1",
                "sourceType": "Person1",
                "targetType": "Person1",
            });
            let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
            let result = create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx).unwrap(),
                schema_struct,
                pod_owner,
                &cli,
            )
            .await;
            assert!(result.is_ok());

            let schema = &mut database_api::get_schema(&tx).unwrap();

            // create edges level 1 w edgy union
            let json = json!({
                "_source": "00000101",
                "_target": "00000999",
                "_name": "friend1",
            });
            let parsed = serde_json::from_value(json).unwrap();
            create_edge(&tx, schema, parsed).unwrap();

            // create edges level 1 w edgy union
            let json = json!({
                "_source": "00000101".to_owned(),
                "_target": "000001999".to_owned(),
                "_name": "friend1",
            });
            let parsed = serde_json::from_value(json).unwrap();
            create_edge(&tx, schema, parsed).unwrap();

            let query = "
        query {
            Person1 (filter: {id: {eq: \"00000101\"}}) {
               id
               age
               friend1 {
                   ... on Person1 {
                       age
                   }
                   ... on Person2 {
                       id
                   }
               }
            }
        }"
            .to_owned();

            let res = graphql(&tx, schema, query);
            let data = res.unwrap();
            let data = data.get("data").unwrap();

            let union_friends = data[0].get("friend1").unwrap();
            assert_eq!(
                union_friends.clone().as_array().unwrap().clone().len(),
                2,
                "Expected 2 union friends"
            );
            let age_1 = match union_friends[0].get("age") {
                Some(age) => age.as_i64().unwrap(),
                None => panic!("Expected age"),
            };
            let _ = match union_friends[1].get("age") {
                Some(_) => panic!("Not expecting age property on Person2"),
                None => 0,
            };
            assert_eq!(age_1, 999, "Expected age=999 for Person1-friend1-Person1");

            Ok(())
        })
        .await
        .unwrap()
    }

    #[tokio::test]
    /// Test creating an item that has type="Edge"
    async fn test_create_edge_item() {
        let mut conn = get_memory_connection().unwrap();
        let cli = command_line_interface::tests::test_cli();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx).unwrap();
            let edge_item = json!({
                "type": "Edge",
            });
            let edge_item: CreateItem = serde_json::from_value(edge_item).unwrap();
            // Creating raw "Edge" items should be successful.
            // (It is also the way how Memri clients push edge information to the Pod.)
            create_item_tx(&tx, &mut schema, edge_item, "", &cli).await
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_edge_schema() {
        let mut conn = get_memory_connection().unwrap();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let cli = command_line_interface::tests::test_cli();

            // Test creating an edge item. Should succeed.
            let schema_json = json!({
              "type": "ItemEdgeSchema",
              "edgeName": "friend",
              "sourceType": "Person",
              "targetType": "Person",
            });

            let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
            let result = internal_api::create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx).unwrap(),
                schema_struct.clone(),
                "",
                &cli,
            )
            .await;
            assert!(result.is_ok());

            // Test duplicate edge target type entry. Should fail.
            let duplicate_result = internal_api::create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx).unwrap(),
                schema_struct,
                "",
                &cli,
            )
            .await;
            assert!(duplicate_result.is_ok());
            // make sure it gets the original id properly
            assert!(duplicate_result.as_ref().unwrap() == result.as_ref().unwrap());

            // Test edgy union, new target type addition. Should succeed
            let schema_json = json!({
                "type": "ItemEdgeSchema",
                "edgeName": "friend",
                "sourceType": "Person",
                "targetType": "Dog",
            });

            let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
            internal_api::create_item_tx(
                &tx,
                &mut database_api::get_schema(&tx).unwrap(),
                schema_struct,
                "",
                &cli,
            )
            .await
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_edge_search() {
        let mut conn = get_memory_connection().unwrap();
        let cli = command_line_interface::tests::test_cli();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx).unwrap();

            let schema_json = json!({
              "type": "ItemEdgeSchema",
              "edgeName": "friend",
              "sourceType": "Person",
              "targetType": "Person",
            });
            let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
            let result =
                internal_api::create_item_tx(&tx, &mut schema, schema_struct, "", &cli).await;
            assert!(result.is_ok());

            let schema_struct: CreateItem = serde_json::from_value(json!({
              "type": "ItemPropertySchema",
              "itemType": "Person",
              "propertyName": "age",
              "valueType": "Integer"
            }))
            .unwrap();
            let result =
                internal_api::create_item_tx(&tx, &mut schema, schema_struct, "", &cli).await;
            assert!(result.is_ok());

            let person1 = {
                let person = json!({
                    "type": "Person",
                });
                let person: CreateItem = serde_json::from_value(person).unwrap();
                create_item_tx(&tx, &mut schema, person, "", &cli)
                    .await
                    .unwrap()
            };
            let person2 = {
                let person = json!({
                    "type": "Person",
                });
                let person: CreateItem = serde_json::from_value(person).unwrap();
                create_item_tx(&tx, &mut schema, person, "", &cli)
                    .await
                    .unwrap()
            };
            let _edge = {
                let json = json!({
                    "_source": person1,
                    "_target": person2,
                    "_name": "friend",
                });
                let parsed = serde_json::from_value(json).unwrap();
                create_edge(&tx, &mut schema, parsed).unwrap()
            };

            let result = {
                let json = json!({
                    "id": person1,
                    "[[edges]]": {},
                });
                let parsed = serde_json::from_value(json).unwrap();
                search(&tx, &schema, parsed).unwrap()
            };
            assert_eq!(result.len(), 1);
            let result = result.into_iter().next().unwrap();
            let result = match result {
                Value::Object(r) => r,
                _ => panic!("Result is not JSON Object"),
            };

            // Example:
            // {
            //   "[[edges]]": [
            //     {
            //       "_edge": "friend",
            //       "_item": {
            //         "dateCreated": 1624363823546,
            //         "dateModified": 1624363823546,
            //         "dateServerModified": 1624363823546,
            //         "deleted": false,
            //         "id": "1539da9dfc14a0467694e79730135df6",
            //         "type": "Person"
            //       },
            //       "dateCreated": 1624363823546,
            //       "dateModified": 1624363823546,
            //       "dateServerModified": 1624363823546,
            //       "deleted": false,
            //       "id": "4ea088d17a670cbb26f4154bce809950",
            //       "type": "Edge"
            //     }
            //   ],
            //   "dateCreated": 1624363823545,
            //   "dateModified": 1624363823545,
            //   "dateServerModified": 1624363823545,
            //   "deleted": false,
            //   "id": "59dd635884bff1de1c8447ac3763543b",
            //   "type": "Person"
            // }

            let result_id = result.get("id").unwrap().as_str().unwrap();
            assert_eq!(result_id, person1);

            let result_edge = result.get("[[edges]]").unwrap().as_array().unwrap();
            let result_edge = result_edge.iter().next().unwrap();

            assert_eq!(
                result_edge.get("_edge").unwrap().as_str().unwrap(),
                "friend"
            );

            let result_edge_item = result_edge.get("_item").unwrap();
            let result_edge_item = match result_edge_item {
                Value::Object(i) => i,
                _ => panic!("edge item is not an JSON Object"),
            };
            assert_eq!(
                result_edge_item.get("id").unwrap().as_str().unwrap(),
                person2
            );

            Ok(())
        })
        .await
        .unwrap();
    }

    #[tokio::test]
    async fn test_edge_delete() {
        let mut conn = get_memory_connection().unwrap();
        let cli = command_line_interface::tests::test_cli();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx).unwrap();

            let schema_json = json!({
              "type": "ItemEdgeSchema",
              "edgeName": "friend",
              "sourceType": "Person",
              "targetType": "Person",
            });
            let schema_struct: CreateItem = serde_json::from_value(schema_json).unwrap();
            let result =
                internal_api::create_item_tx(&tx, &mut schema, schema_struct, "", &cli).await;
            assert!(result.is_ok());

            let schema_struct: CreateItem = serde_json::from_value(json!({
              "type": "ItemPropertySchema",
              "itemType": "Person",
              "propertyName": "age",
              "valueType": "Integer"
            }))
            .unwrap();
            let result =
                internal_api::create_item_tx(&tx, &mut schema, schema_struct, "", &cli).await;
            assert!(result.is_ok());

            let person1 = {
                let person = json!({
                    "type": "Person",
                });
                let person: CreateItem = serde_json::from_value(person).unwrap();
                create_item_tx(&tx, &mut schema, person, "", &cli)
                    .await
                    .unwrap()
            };
            let person2 = {
                let person = json!({
                    "type": "Person",
                });
                let person: CreateItem = serde_json::from_value(person).unwrap();
                create_item_tx(&tx, &mut schema, person, "", &cli)
                    .await
                    .unwrap()
            };
            let _edge = {
                let json = json!({
                    "_source": person1,
                    "_target": person2,
                    "_name": "friend",
                });
                let parsed = serde_json::from_value(json).unwrap();
                create_edge(&tx, &mut schema, parsed).unwrap()
            };

            let result = {
                let json = json!({
                    "id": person1,
                    "[[edges]]": {},
                });
                let parsed = serde_json::from_value(json).unwrap();
                search(&tx, &schema, parsed).unwrap()
            };
            assert_eq!(result.len(), 1);

            {
                let json = json!({
                    "_source": person1,
                    "_target": person2,
                });
                let parsed = serde_json::from_value(json).unwrap();
                delete_edge_by_source_target(&tx, parsed).unwrap()
            };
            let result = {
                let json = json!({
                    "id": person1,
                    "[[edges]]": {},
                });
                let parsed = serde_json::from_value(json).unwrap();
                search(&tx, &schema, parsed).unwrap()
            };
            assert_eq!(result.len(), 1);
            let result = result.into_iter().next().unwrap();
            let result = match result {
                Value::Object(r) => r,
                _ => panic!("Result is not JSON Object"),
            };
            let result_edges = result.get("[[edges]]").unwrap().as_array().unwrap();
            assert!(
                result_edges.is_empty(),
                "There should be 0 edges after the deletion of the edge"
            );
            Ok(())
        })
        .await
        .unwrap();
    }
}
