use crate::async_db_connection::AsyncTx;
use crate::db_model::EXPIRES_AFTER;
use crate::{
    api_model::{
        CreateItem, Oauth2AccessTokenRequest, Oauth2AccessTokenResponse, Oauth2AuthUrlRequest,
        Oauth2AuthUrlResponse, Oauth2AuthorizeTokenRequest, Platforms, Search, SortOrder,
    },
    bad_request,
    command_line_interface::CliOptions,
    db_model::{
        Oauth2Flow, ACCESS_TOKEN, OAUTH2_FLOW, PLATFORM, REDIRECT_URI, REFRESH_TOKEN, TOKEN_TYPE,
    },
    error::{ErrorContext, Result},
    internal_api::{self, search},
    internal_error,
    schema::Schema,
};
use lazy_static::lazy_static;
use oauth2::{
    basic::{BasicClient, BasicTokenResponse},
    reqwest::async_http_client,
    AuthUrl, AuthorizationCode, ClientId, ClientSecret, CsrfToken, PkceCodeChallenge,
    PkceCodeVerifier, RedirectUrl, RefreshToken, Scope, TokenResponse, TokenUrl,
};
use oauth2::{ExtraTokenFields, StandardTokenResponse, TokenType};

use serde_json::json;
use std::time::SystemTime;
use std::{collections::HashMap, env};
use tracing::{debug, warn};

pub fn auth_url(payload: Oauth2AuthUrlRequest) -> Result<Oauth2AuthUrlResponse> {
    debug!("Creating auth url for {:?}", payload.platform);

    let client = get_client(&payload.platform, payload.redirect_uri)?;

    let (pkce_challenge, pkce_verifier) = PkceCodeChallenge::new_random_sha256();

    let (authorize_url, csfr_token) = client
        .authorize_url(CsrfToken::new_random)
        .add_scopes(payload.scopes.into_iter().map(Scope::new))
        .set_pkce_challenge(pkce_challenge)
        .url();

    // TODO: PKCE verifier should be encrypted, well it depends on the security model we assume,
    // but worth considering - re-use idea from PluginAuth
    // same for csfr_token, that one should be used on FE to validate response from the
    // oauth server in order to check for MITM attack.
    Ok(Oauth2AuthUrlResponse {
        authorize_url: authorize_url.to_string(),
        pkce_verifier: pkce_verifier.secret().to_string(),
        csfr_token: csfr_token.secret().to_string(),
    })
}

pub async fn access_token(
    tx: &AsyncTx,
    schema: &Schema,
    payload: Oauth2AccessTokenRequest,
) -> Result<Oauth2AccessTokenResponse> {
    debug!("Retrieving access token for {:?}", payload.platform);

    let oauth2_item = get_oauth2_flow(tx, schema, &payload.platform)?;
    if token_should_be_refreshed(&oauth2_item) {
        debug!("Refreshing token...");

        let token = exchange_refresh_token(&oauth2_item).await?;

        let expires_after = get_expiration_time_from_response(&token);

        internal_api::update_item_tx(
            tx,
            schema,
            &oauth2_item.base.id,
            json!({
                ACCESS_TOKEN: token.access_token().secret(),
                REFRESH_TOKEN: token.refresh_token().expect("missing refresh token").secret(),
                TOKEN_TYPE: token.token_type().as_ref().to_string(),
                EXPIRES_AFTER: expires_after
            })
            .as_object()
            .expect("Invalid json object for oauth2")
            .clone()
            .into_iter()
            .collect(),
        )?;

        Ok(Oauth2AccessTokenResponse {
            access_token: token.access_token().secret().clone(),
            token_type: token.token_type().as_ref().to_string(),
        })
    } else {
        Ok(Oauth2AccessTokenResponse {
            access_token: oauth2_item.access_token,
            token_type: oauth2_item.token_type,
        })
    }
}

pub async fn authorize(
    tx: &AsyncTx,
    schema: &mut Schema,
    owner: &str,
    cli: &CliOptions,
    payload: Oauth2AuthorizeTokenRequest,
) -> Result<Oauth2AccessTokenResponse> {
    debug!("Authorizing token for {:?}", payload.platform);

    let token = exchange_code(&payload).await?;

    let expires_after = get_expiration_time_from_response(&token);

    // It's allowed to call authorize many times for the same platform,
    // re-use the Oauth2Flow item in the database if exist.
    match get_oauth2_flow(tx, schema, &payload.platform) {
        Ok(item) => {
            // Update already existing item in the database
            internal_api::update_item_tx(
                tx,
                schema,
                &item.base.id,
                json!({
                    ACCESS_TOKEN: token.access_token().secret(),
                    REFRESH_TOKEN: token.refresh_token().expect("missing refresh token").secret(),
                    TOKEN_TYPE: token.token_type().as_ref().to_string(),
                    REDIRECT_URI: payload.redirect_uri,
                    EXPIRES_AFTER: expires_after
                })
                .as_object()
                .expect("Invalid json object for oauth2")
                .clone()
                .into_iter()
                .collect(),
            )?;

            debug!(
                "Access token for {:?}, updated {}",
                payload.platform, item.base.id
            );
        }
        Err(_) => {
            // Create new Oauth2Flow item
            let item = CreateItem {
                _type: OAUTH2_FLOW.to_string(),
                fields: json!({
                    ACCESS_TOKEN: token.access_token().secret(),
                    REFRESH_TOKEN: token.refresh_token().expect("missing refresh token").secret(),
                    TOKEN_TYPE: token.token_type().as_ref().to_string(),
                    REDIRECT_URI: payload.redirect_uri,
                    PLATFORM: payload.platform,
                    EXPIRES_AFTER: expires_after
                })
                .as_object()
                .expect("Invalid json object for oauth2")
                .clone()
                .into_iter()
                .collect(),
                ..Default::default()
            };

            let id = internal_api::create_item_tx(tx, schema, item, owner, cli).await?;
            debug!("Access token for {:?}, created {id}", payload.platform);
        }
    };

    Ok(Oauth2AccessTokenResponse {
        access_token: token.access_token().secret().clone(),
        token_type: token.token_type().as_ref().to_string(),
    })
}

fn get_expiration_time_from_response<EF, TT>(token: &StandardTokenResponse<EF, TT>) -> i64
where
    EF: ExtraTokenFields,
    TT: TokenType,
{
    if let Some(expires_in) = token.expires_in() {
        debug!("Token expires in {expires_in:?}");
        SystemTime::now() + expires_in
    } else {
        warn!("No 'expires in' information in token response, setting it to 0");
        SystemTime::now()
    }
    .duration_since(SystemTime::UNIX_EPOCH)
    .expect("Failed to get duration from system time")
    .as_secs() as i64
}

struct PlatformConfiguration {
    client_id_env: &'static str,
    client_secret_env: &'static str,
    auth_url: AuthUrl,
    token_url: TokenUrl,
}

lazy_static! {
    static ref PLATFORM_CONFIG: HashMap<Platforms, PlatformConfiguration> = HashMap::from([
        (
            Platforms::Gitlab,
            PlatformConfiguration {
                client_id_env: "GITLAB_CLIENT_ID",
                client_secret_env: "GITLAB_CLIENT_SECRET",
                auth_url: AuthUrl::new("https://gitlab.memri.io/oauth/authorize".to_string())
                    .unwrap(),
                token_url: TokenUrl::new("https://gitlab.memri.io/oauth/token".to_string())
                    .unwrap(),
            },
        ),
        (
            Platforms::Twitter,
            PlatformConfiguration {
                client_id_env: "TWITTER_V2_CLIENT_ID",
                client_secret_env: "TWITTER_V2_CLIENT_SECRET",
                auth_url: AuthUrl::new("https://twitter.com/i/oauth2/authorize".to_string())
                    .unwrap(),
                token_url: TokenUrl::new("https://api.twitter.com/2/oauth2/token".to_string())
                    .unwrap(),
            },
        )
    ]);
}

async fn exchange_refresh_token(oauth2_item: &Oauth2Flow) -> Result<BasicTokenResponse> {
    let client = get_client(&oauth2_item.platform, oauth2_item.redirect_uri.clone())?;

    let res = client
        .exchange_refresh_token(&RefreshToken::new(oauth2_item.refresh_token.clone()))
        .request_async(async_http_client)
        .await?;

    Ok(res)
}

async fn exchange_code(payload: &Oauth2AuthorizeTokenRequest) -> Result<BasicTokenResponse> {
    let client = get_client(&payload.platform, payload.redirect_uri.clone())?;

    let res = client
        .exchange_code(AuthorizationCode::new(payload.auth_code.clone()))
        .set_pkce_verifier(PkceCodeVerifier::new(payload.pkce_verifier.clone()))
        .request_async(async_http_client)
        .await?;

    Ok(res)
}

/// Creates oauth2 client for specific platform
fn get_client(platform: &Platforms, redirect_uri: String) -> Result<BasicClient> {
    let config = PLATFORM_CONFIG
        .get(platform)
        .ok_or(internal_error! { "Missing configuration for {platform:?}"})?;

    let client_id = ClientId::new(
        env::var(config.client_id_env).context(|| format!("missing: {}", config.client_id_env))?,
    );
    let client_secret = ClientSecret::new(
        env::var(config.client_secret_env)
            .context(|| format!("missing: {}", config.client_secret_env))?,
    );

    Ok(BasicClient::new(
        client_id,
        Some(client_secret),
        config.auth_url.clone(),
        Some(config.token_url.clone()),
    )
    .set_redirect_uri(RedirectUrl::new(redirect_uri)?))
}

fn get_oauth2_flow(tx: &AsyncTx, schema: &Schema, platform: &Platforms) -> Result<Oauth2Flow> {
    let query = Search {
        _type: Some(OAUTH2_FLOW.to_string()),
        deleted: Some(false),
        limit: 5,
        sort_order: SortOrder::Desc,
        other_properties: HashMap::from([(
            PLATFORM.to_string(),
            serde_json::to_value(platform).expect("Cannot serialize Platforms to Value"),
        )]),
        ..Default::default()
    };

    let mut items = search(tx, schema, query)?;

    debug_assert!(
        items.len() <= 1,
        "There is more than one Oauth2Flow item in the DB for platform {platform:?}, {items:#?}"
    );

    let oauth2_item = items
        .pop()
        .ok_or(bad_request! { "Could not find Oauth2Flow for {platform:?}" })?;

    let oauth2_item: Oauth2Flow = serde_json::from_value(oauth2_item)
        .context(|| format!("Cannot deserialize Oauth2Flow for platform {platform:?}",))?;

    debug!("Found Oauth2 item {}", oauth2_item.base.id);

    Ok(oauth2_item)
}

const EXPIRY_THRESHOLD: i64 = 60;

fn token_should_be_refreshed(oauth2: &Oauth2Flow) -> bool {
    let now_ts = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .expect("Failed to retrieve duration from SystemTime")
        .as_secs() as i64
        + EXPIRY_THRESHOLD;

    debug!("Token time to expire {:}s", oauth2.expires_after - now_ts);
    now_ts > oauth2.expires_after
}
