use crate::{
    api_model::{AuthKey, PluginAuth, PluginAuthData},
    bad_request,
    error::Result,
    global_static,
};
use chacha20poly1305::{
    aead::{generic_array::GenericArray, Aead, NewAead},
    Key, XChaCha20Poly1305, XNonce,
};
use rusqlite::Connection;
use sha2::{digest::typenum, Digest, Sha256};
use std::ops::Not;
use zeroize::Zeroize;

pub type SHA256Output = GenericArray<u8, typenum::consts::U32>;

pub fn generate_xchacha20poly1305_cipher() -> XChaCha20Poly1305 {
    let key: [u8; 32] = rand::random();
    let key = Key::from_slice(&key);
    XChaCha20Poly1305::new(key)
}

/// Given a `database_key` from an already authorized request,
/// create a new PluginAuth to be passed to external Plugins.
pub fn create_plugin_auth(database_key: &str) -> Result<PluginAuth> {
    let nonce: [u8; 24] = rand::random();
    let nonce = XNonce::from_slice(&nonce); // MUST be unique
    let cipher = &global_static::CIPHER;
    let database_key: Vec<u8> = hex::decode(database_key)?;
    let encrypted = cipher.encrypt(nonce, database_key.as_slice())?;
    Ok(PluginAuth {
        data: PluginAuthData {
            nonce: hex::encode(nonce),
            encrypted_permissions: hex::encode(encrypted),
        },
    })
}

/// Given a request that uses PluginAuth, decrypt its `database_key`.
pub fn extract_database_key(plugin_auth: &PluginAuth) -> Result<DatabaseKey> {
    if plugin_auth.data.nonce.len() != 48 {
        return Err(bad_request! { "PluginAuth nonce should be 48 hex characters" });
    }
    let nonce = hex::decode(&plugin_auth.data.nonce)?;
    let encrypted_permissions = hex::decode(&plugin_auth.data.encrypted_permissions)?;
    let nonce = XNonce::from_slice(&nonce);
    let cipher = &global_static::CIPHER;
    let decrypted = cipher.decrypt(nonce, encrypted_permissions.as_ref())?;
    if decrypted.len() != 32 && decrypted.is_empty().not() {
        return Err(bad_request! { "Key has incorrect length: {}", decrypted.len() });
    }
    Ok(DatabaseKey {
        database_key: hex::encode_upper(&decrypted),
    })
}

/// Returns database key structure.
pub fn auth_to_database_key(auth: AuthKey) -> Result<DatabaseKey> {
    match auth {
        AuthKey::ClientAuth(c) => DatabaseKey::from(c.database_key),
        AuthKey::PluginAuth(p) => extract_database_key(&p),
    }
}

/// Database key stored in raw String format (as supplied to sqlcipher).
pub struct DatabaseKey {
    /// PRIVATE key (intentionally not public)!
    /// The wrapper struct only exposes "safe" methods to work with the key.
    database_key: String,
}

impl Drop for DatabaseKey {
    fn drop(&mut self) {
        self.database_key.zeroize();
    }
}

impl std::fmt::Debug for DatabaseKey {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("DatabaseKey(hidden_fields)")
    }
}

impl DatabaseKey {
    /// Set encryption for a database connection, taking care of memory zeroing in the process.
    pub fn execute_sqlite_pragma(&self, conn: &Connection) -> Result<()> {
        let mut sql = format!(
            "
            PRAGMA key = \"x'{}'\";
            PRAGMA cipher_compatibility = 4;
        ",
            self.database_key
        );
        conn.execute_batch(&sql)?;
        sql.zeroize();
        Ok(())
    }

    /// Returns SHA256 hash of underlying priv key
    pub fn get_sha256(&self) -> SHA256Output {
        Sha256::new_with_prefix(&self.database_key).finalize()
    }

    pub fn from(mut key_as_hex_string: String) -> Result<DatabaseKey> {
        key_as_hex_string.make_ascii_uppercase();
        for c in key_as_hex_string.chars() {
            if !(c.is_ascii_uppercase() || c.is_ascii_digit()) {
                key_as_hex_string.zeroize();
                return Err(bad_request! {
                    // important: don't log the actual char
                    "Invalid database key character"
                });
            }
        }
        Ok(DatabaseKey {
            database_key: key_as_hex_string,
        })
    }
    pub fn create_plugin_auth(&self) -> Result<PluginAuth> {
        create_plugin_auth(&self.database_key)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_encrypt_decrypt() {
        let db_key = "0".repeat(64);
        let db_key_struct = DatabaseKey::from(db_key.clone()).unwrap();
        let auth = db_key_struct.create_plugin_auth().unwrap();
        assert!(!auth.data.nonce.contains(&db_key));
        assert!(!auth.data.encrypted_permissions.contains(&db_key));
        assert_eq!(
            auth.data.encrypted_permissions.len(),
            db_key.len() + 32,
            "Encrypted permissions should include the 16-byte (32 hex characters) TAG."
        );
        let decrypted = extract_database_key(&auth).unwrap();
        assert_eq!(decrypted.database_key, db_key);
    }
}
