use crate::{
    any_error,
    api_model::{Search, SortOrder},
    bad_request,
    command_line_interface::CliOptions,
    db_model::{PluginRunItem, Trigger},
    error::{ErrorContext, Result},
    internal_api::search,
    plugin_run::get_plugin_base_endpoint,
    schema::Schema,
    triggers::AfterAction,
};

use crate::async_db_connection::AsyncTx;
use reqwest::header::CONTENT_TYPE;
use serde_json::{json, Value};

use tracing::{debug, error, info, trace};

pub async fn on_new_items(
    tx: &AsyncTx,
    schema: &Schema,
    items: &[Value],
    cli: &CliOptions,
) -> Result<AfterAction> {
    let triggers = get_all_triggers(tx, schema)?;

    let mut after_action = AfterAction::default();

    // For each trigger find subset of items
    let items_for_triggers: Vec<Vec<&Value>> = triggers
        .iter()
        .map(|trigger| {
            let property_pointer = format!("/{}", trigger.filter_created_after_property_name);

            let items_for_trigger: Vec<&Value> = items
                .iter()
                .filter(
                    |item| match item_should_do_trigger(item, &property_pointer, trigger) {
                        Ok(should_trigger) => should_trigger,
                        Err(e) => {
                            error!("Unable to parse item, reason: {e}");
                            false
                        }
                    },
                )
                .collect();

            items_for_trigger
        })
        .collect();

    for (idx, trigger) in triggers.into_iter().enumerate() {
        if items_for_triggers[idx].is_empty() {
            debug!("No messages for trigger {}", trigger.id);
            continue;
        }

        let plugin_instance = get_plugin_instance(tx, schema, &trigger.plugin_run_id);

        let mut delete_trigger = false;

        match plugin_instance {
            Ok(plugin_instance) => {
                if let Some(status) = &plugin_instance.status {
                    match status.as_str() {
                        "done" | "error" => {
                            // Plugin was registered for the trigger that has finished, delete the trigger
                            info!(
                                "Deleting trigger {}, because plugin {} has status {status}",
                                trigger.id, trigger.plugin_run_id
                            );
                            delete_trigger = true;
                        }
                        // Plugin still works, notify it
                        _ => {
                            notify_plugin(&trigger, &items_for_triggers[idx], &plugin_instance, cli)
                                .await?
                        }
                    }
                } else {
                    trace!(
                        "There is no status field for plugin {}, yet",
                        trigger.plugin_run_id
                    );
                }
            }
            Err(e) => {
                info!("Deleting trigger {}, because of {e}", trigger.id);
                delete_trigger = true;
            }
        }

        if delete_trigger {
            after_action.items_to_delete.insert(trigger.id);
        }
    }

    Ok(after_action)
}

fn item_should_do_trigger(item: &Value, item_property: &str, trigger: &Trigger) -> Result<bool> {
    if trigger.trigger_on == item["type"].as_str().unwrap() {
        let created_after = item
            .pointer(item_property)
            .and_then(|property| property.as_i64())
            .ok_or(bad_request! { "Cannot get property {item_property}" })?;

        Ok(created_after > trigger.filter_created_after)
    } else {
        Ok(false)
    }
}

fn get_all_triggers(tx: &AsyncTx, schema: &Schema) -> Result<Vec<Trigger>> {
    let query = Search {
        _type: Some("Trigger".to_string()),
        deleted: Some(false),
        limit: u64::MAX,
        // Return triggers in order they were inserted
        sort_order: SortOrder::Asc,
        ..Default::default()
    };

    search(tx, schema, query)?
        .into_iter()
        .map(serde_json::from_value)
        .collect::<std::result::Result<Vec<Trigger>, serde_json::Error>>()
        .context_str("Failed to retrieve Triggers from the DB")
}

fn get_plugin_instance(
    tx: &AsyncTx,
    schema: &Schema,
    plugin_run_id: &str,
) -> Result<PluginRunItem> {
    let query = Search {
        id: Some(plugin_run_id.to_string()),
        deleted: Some(false),
        limit: 1,
        ..Default::default()
    };

    let plugin_instance = search(tx, schema, query)?.pop().ok_or(bad_request! {
        "Could not find PluginRun {plugin_run_id}, maybe it's deleted?",
    })?;

    let plugin_instance: PluginRunItem = serde_json::from_value(plugin_instance)
        .context(|| format!("Cannot parse PluginRun with id {plugin_run_id}"))?;

    Ok(plugin_instance)
}

async fn notify_plugin(
    trigger: &Trigger,
    items: &[&Value],
    plugin_instance: &PluginRunItem,
    cli: &CliOptions,
) -> Result<()> {
    let endpoint = format!(
        "{}/{}",
        get_plugin_base_endpoint(plugin_instance, cli),
        trigger.action
    );

    let item_ids: Vec<String> = items
        .iter()
        .map(|item| item["id"].as_str().unwrap().to_string())
        .collect();
    let body = json!({ "item_ids": item_ids, "trigger_id": trigger.id.clone() }).to_string();

    debug!(
        "Going to trigger PluginRun instance {} by Trigger {} because of arrival of {}",
        trigger.plugin_run_id, trigger.id, items[0]["type"]
    );
    debug!("POST: {endpoint}, with body {body:#}");

    let client = reqwest::Client::new();
    let res = client
        .post(&endpoint)
        .header(CONTENT_TYPE, "application/json")
        .body(body.clone())
        .send()
        .await
        .context(|| {
            format!(
                "Failed to trigger plugin instance {}, endpoint {endpoint}",
                trigger.plugin_run_id
            )
        })?;

    debug!("Got POST {} response {res:?}", trigger.action);

    if !res.status().is_success() {
        return Err(any_error! {
            res.status(),
            "Failure while triggering the plugin. PluginRun = {}, action = {} body = {}, error = {:?}",
            trigger.plugin_run_id, trigger.action, body, res.text().await
        });
    }

    Ok(())
}
