use crate::{
    api_model::{AuthKey, ClientAuth, CreateItem, Search},
    async_db_connection::{AsyncConnection, AsyncTx},
    command_line_interface::PARSED,
    database_api::{self, dangerous_permament_remove_item},
    database_pool::{get_db_connection, initialize_db, InitDb},
    db_model::ItemBase,
    error::Result,
    internal_api::{self, search},
    plugin_auth_crypto::auth_to_database_key,
};
use std::ops::Deref;
use tracing::{debug, info, warn};

// TODO: there is no proper plugin state management
// if pod restarts - plugins are unable to reach it, plugin auth is broken
// plugins listens for that, and shuts down
// garbage in the db stays
// if plugin dies, pod does nothing with it

pub async fn db_connection(init_db: &InitDb) -> Result<Option<AsyncConnection>> {
    if let (Some(owner), Some(database_key)) = (
        PARSED.owner_key_for_pod.clone(),
        PARSED.db_key_for_pod.clone(),
    ) {
        let database_key = auth_to_database_key(AuthKey::ClientAuth(ClientAuth { database_key }))?;
        let conn = get_db_connection(&owner, init_db, &database_key).await?;

        Ok(Some(conn))
    } else {
        Ok(None)
    }
}

pub async fn initialize(init_db: &InitDb) -> Result<()> {
    let (Some(owner_key), Some(database_key)) = (PARSED.owner_key_for_pod.clone(), PARSED.db_key_for_pod.clone()) else {
        info!("No POD credentials provided, skipping startup of shared plugins");
        return Ok(());
    };

    let database_key = auth_to_database_key(AuthKey::ClientAuth(ClientAuth { database_key }))?;

    let _ = initialize_db(&owner_key, init_db, &database_key).await;

    let mut conn = get_db_connection(&owner_key, init_db, &database_key).await?;

    // Remove PluginRun from the DB before starting new plugins
    // Currently plugins have listeners that will detect lack of POD connection
    // and will close themselves.
    // This is very simple approach, that might become invalid in the future.
    // But for now it behaves well.
    clean_db_from_plugins(&mut conn).await?;

    for plugin in &PARSED.shared_plugins {
        let payload: CreateItem = serde_json::from_str(plugin)?;
        let plugin_name = payload.fields.get("pluginName").cloned();
        let id = payload.id.clone();

        match internal_api::create_item(
            payload,
            &mut conn,
            &owner_key,
            &database_key,
            PARSED.deref(),
        )
        .await
        {
            Ok(id) => info!("Started plugin {plugin_name:?} with id {id}"),
            // TODO: check if plugin still exists, if not - delete from the db, using kubectl probably, because
            // what if is in status pending - health endpoint would not work
            Err(e) => warn!("Plugin with id {id:?} not started, reason {e}",),
        }
    }

    Ok(())
}

async fn clean_db_from_plugins(conn: &mut AsyncConnection) -> Result<()> {
    conn.in_write_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;

        let query = Search {
            _type: Some("PluginRun".to_string()),
            limit: u64::MAX,
            ..Default::default()
        };

        let plugins = search(&tx, &schema, query)?
            .into_iter()
            .map(|element| {
                serde_json::from_value::<ItemBase>(element).unwrap_or_else(|err| {
                    panic!("Unable to deserialize value into ItemBase, reason {err:?}")
                })
            })
            .collect::<Vec<_>>();

        debug!("Number of plugins to remove {}", plugins.len());

        for plugin in plugins {
            debug!("Removing {}", plugin.id);

            if let Err(e) = dangerous_permament_remove_item(&tx, plugin.rowid) {
                warn!(
                    "Failed to remove PluginRun with id {}, reason {e:?}",
                    plugin.id
                );
            }
        }

        Ok(())
    })
    .await
}
