//
// Pod triggers that should "run" on particular DB/data changes (e.g. item insertion)
//

use std::{collections::HashSet, fmt::Debug};

use crate::{
    any_error,
    api_model::{CreateItem, Oauth2AccessTokenRequest},
    async_db_connection::{AsyncConnection, AsyncTx},
    bad_request,
    command_line_interface::CliOptions,
    database_api::{self},
    db_model::{
        ItemWithBase, PluginRunItem, CONTAINER_ID, UNIQUE_ID, WEBSERVER_PORT, WEBSERVER_URL,
    },
    error::{ErrorContext, Result},
    internal_api::{self, get_item_tx},
    internal_error, oauth2_api,
    plugin_auth_crypto::DatabaseKey,
    plugin_run, plugin_trigger,
    schema::{Schema, SchemaEdge, SchemaItem, ITEM_EDGE_SCHEMA, ITEM_PROPERTY_SCHEMA},
};

use crate::async_db_connection::AsyncTx as Tx;
use serde::Deserialize;
use serde_json::{json, Value};

use tracing::{debug, error};

#[derive(Debug, Eq, PartialEq)]
pub enum SchemaAdditionChange {
    NotASchema,
    NewSchemaAdded,
    OldSchema { old_id: String },
}

/// If an item is a Schema, add it to the schema.
/// Fail if an incompatible schema is attempted to be inserted.
pub fn add_item_as_schema_opt(
    tx: &Tx,
    schema: &mut Schema,
    item: &CreateItem,
) -> Result<SchemaAdditionChange> {
    if item._type == ITEM_PROPERTY_SCHEMA {
        // We'll do something ugly here.
        // We'll convert the item into JSON, and then back into the desired type
        // for type check and parsing. This is easier code-wise than to do manual conversions.
        // It only triggers for specific, rarely used items. This implementation might change later.
        let json = serde_json::to_value(item)?;
        let parsed: SchemaItem = serde_json::from_value(json)
            .context(|| format!("Parsing of Schema item {:?}, {}:{}", item, file!(), line!()))?;

        // The result of the operation depends on:
        // * the "heterogeneous" property's type
        // * this particular itemType's property type
        // * whether an ID is present in the request
        match (
            schema.property(&parsed.property_name),
            schema.item_property(&parsed.item_type, &parsed.property_name),
            &item.id,
        ) {
            (Some(old_value_type), _, _) if &parsed.value_type != old_value_type => {
                Err(bad_request! {
                    "Schema for property {} is already defined to valueType {}, cannot override to {}",
                    parsed.property_name,
                    old_value_type,
                    parsed.value_type,
                })
            }
            (_, None, _) | (_, _, Some(_)) => {
                schema.set_type_unchecked(
                    parsed.item_type,
                    parsed.property_name,
                    parsed.value_type,
                );
                Ok(SchemaAdditionChange::NewSchemaAdded)
            }
            (_, Some(old_value_type), None) if &parsed.value_type == old_value_type => {
                let old_id = database_api::get_schema_item_id_for_property(
                    tx,
                    &parsed.item_type,
                    &parsed.property_name,
                )?;
                Ok(SchemaAdditionChange::OldSchema { old_id })
            }
            (union_value_type, Some(old_value_type), None) => Err(internal_error! {
                "Unexpected failure, {}.{} has valueType {}. However, this Schema should not have been allowed as {} is already defined as {:?}",
                parsed.item_type,
                parsed.property_name,
                old_value_type,
                parsed.property_name,
                union_value_type,
            }),
        }
    } else if item._type == ITEM_EDGE_SCHEMA {
        let json = serde_json::to_value(item)?;
        let parsed: SchemaEdge = serde_json::from_value(json).context(|| {
            format!(
                "Parsing of Schema item edge {:?}, {}:{}",
                item,
                file!(),
                line!()
            )
        })?;

        let edge_type_option = schema.item_edge(&parsed.source_type, &parsed.edge_name);

        match edge_type_option {
            // do not add if edge already exists
            Some(edge_type_vector) if edge_type_vector.contains(&parsed.target_type) => {
                let old_id = database_api::get_schema_edge_item_id(
                    tx,
                    &parsed.edge_name,
                    &parsed.source_type,
                    &parsed.target_type,
                )?;
                Ok(SchemaAdditionChange::OldSchema { old_id })
            }
            // in case the edge does not exist or it doesn't contain the target type, add it
            Some(_) | None => {
                schema.set_edgetype_unchecked(
                    parsed.source_type,
                    parsed.edge_name,
                    parsed.target_type,
                );
                Ok(SchemaAdditionChange::NewSchemaAdded)
            }
        }
    } else {
        Ok(SchemaAdditionChange::NotASchema)
    }
}

/// Used during item creation, completes the item by fields that are not possible to set by the client.
pub fn add_extra_fields(item: &mut CreateItem, cli: &CliOptions, pod_owner: &str) -> Result<()> {
    if item._type == "PluginRun"
        // When attaching to already existing plugin those fields will already exist, so skip this step
        && !item.fields.contains_key(WEBSERVER_PORT)
        && !item.fields.contains_key(WEBSERVER_URL)
    {
        // New plugin instance is going to be created, get free port for it's webserver
        // and store it in the DB
        let webserver_port = plugin_run::get_free_port(cli);
        item.fields
            .insert(WEBSERVER_PORT.to_string(), json!(webserver_port));

        let container_id = plugin_run::get_container_id(item, cli, pod_owner)?;
        item.fields
            .insert(CONTAINER_ID.to_string(), json!(container_id));

        let webserver_url = plugin_run::resolve_webserver_url(item, cli, pod_owner)?;
        item.fields
            .insert(WEBSERVER_URL.to_string(), json!(webserver_url));
    }

    Ok(())
}

pub async fn add_extra_fields_for_twitter(
    tx: &Tx,
    schema: &Schema,
    item: &mut CreateItem,
) -> Result<()> {
    debug!("add_extra_fields_for_twitter");
    if item._type == "PluginRun" {
        debug!("item is PluginRun");
        if let Some(plugin_name) = item
            .fields
            .get("pluginName")
            .cloned()
            .map(|name| serde_json::from_value::<String>(name).unwrap())
        {
            debug!("Plugin name is {plugin_name}");
            if plugin_name == "TwitterPlugin" {
                let payload = Oauth2AccessTokenRequest {
                    platform: crate::api_model::Platforms::Twitter,
                };
                let resp = oauth2_api::access_token(tx, schema, payload)
                    .await
                    .context_str("while retrieving access token for twitter hack")?;

                let twitter_user_id = {
                    let client = reqwest::Client::new();
                    let res = client
                        .get("https://api.twitter.com/2/users/me")
                        .bearer_auth(resp.access_token)
                        .send()
                        .await
                        .context_str("Failed to verify twitter credentials")?;

                    debug!("Got twitter  response {res:#?}");

                    if !res.status().is_success() {
                        return Err(any_error! {
                            res.status(),
                            "Failure getting the credentials"
                        });
                    }

                    #[derive(Deserialize, Debug)]
                    struct TwitterUserResponse {
                        data: TwitterUserData,
                    }

                    #[derive(Deserialize, Debug)]
                    struct TwitterUserData {
                        id: String,
                    }

                    let res: TwitterUserResponse = res.json().await?;

                    res.data.id
                };

                debug!("Storing {UNIQUE_ID}={twitter_user_id} into PluginRun");
                item.fields
                    .insert(UNIQUE_ID.to_string(), json!(twitter_user_id));
            }
        }
    }

    Ok(())
}

#[derive(Default)]
pub struct AfterAction {
    pub items_to_delete: HashSet<String>,
}

pub async fn actions_after_item_create<T>(
    conn: &mut AsyncConnection,
    created_items_ids: &[T],
    owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<()>
where
    T: AsRef<str>,
{
    let tx_res = conn
        .in_read_transaction(|tx: AsyncTx| async move {
            let schema = database_api::get_schema(&tx)?;

            execute_actions_after_item_create(
                &tx,
                &schema,
                created_items_ids,
                owner,
                cli,
                database_key,
            )
            .await
        })
        .await;

    match tx_res {
        Ok(actions) => {
            // Write TX might be expensive to execute, therefore,
            // firstly check if there is anything to modify
            if !actions.items_to_delete.is_empty() {
                conn.in_write_transaction(|tx: AsyncTx| async move {
                    for del_item in actions.items_to_delete {
                        // Report the error, but do not break the loop
                        if let Err(e) = internal_api::delete_item_by_id(&tx, &del_item) {
                            error!(
                                "Removing item {del_item} after item creation failed, reason {e}"
                            );
                        }
                    }

                    Ok(())
                })
                .await
            } else {
                Ok(())
            }
        }

        Err(e) => {
            error!("Failed upon item triggering, reason {e}");
            Err(e)
        }
    }
}

/// After item is in the DB do appropriate actions.
/// Invariants:
///  1. This can be called only after successful commit of the write TX. Meaning all new data is visible to all readers.
///  2. Code inside this function cannot do write/update to fields, otherwise it introduces a datarace with readers
///     spawned as a result of this call. Again data will be visible for POD, since its in the middle of the TX,
///     but not outside the TX, as long as it's not committed. When that happens is unknown, hence no assumptions can be made.
/// Returns actions that has to be done in separate transaction.
#[allow(clippy::too_many_arguments)]
async fn execute_actions_after_item_create<T>(
    tx: &Tx,
    schema: &Schema,
    created_items: &[T],
    pod_owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<AfterAction>
where
    T: AsRef<str>,
{
    let created_items = created_items
        .iter()
        .map(|item_created_id| {
            get_item_tx(tx, schema, item_created_id.as_ref())?
                .pop()
                .ok_or(
                    bad_request! { "Could not find item in the db {}", item_created_id.as_ref() },
                )
        })
        .collect::<Result<Vec<Value>>>()?;

    handle_plugin_run_items(tx, schema, &created_items, pod_owner, cli, database_key).await?;

    plugin_trigger::on_new_items(tx, schema, &created_items, cli).await
}

async fn handle_plugin_run_items(
    tx: &Tx,
    schema: &Schema,
    created_items: &[Value],
    pod_owner: &str,
    cli: &CliOptions,
    database_key: &DatabaseKey,
) -> Result<()>
where {
    let mut errors_during_plugin_run_actions = vec![];

    for item in created_items.iter() {
        if item["type"] == "PluginRun" {
            if let Err(e) = {
                let plugin: ItemWithBase<PluginRunItem> = serde_json::from_value(item.clone())?;

                plugin_run::run_plugin_container(
                    tx,
                    schema,
                    &plugin.base.id,
                    &plugin.item,
                    pod_owner,
                    database_key,
                    cli,
                )
                .await
            } {
                // Collect errors, but do not break the loop, since following items might execute without problems
                errors_during_plugin_run_actions.push((item["id"].as_str().unwrap(), e));
            }
        }
    }

    if !errors_during_plugin_run_actions.is_empty() {
        let collected_error_msgs = errors_during_plugin_run_actions.into_iter().fold(
            "Bulk trigger stage failed for following reasons: ".to_string(),
            |mut acc, (item_id, error)| {
                acc.push_str(&format!("\n{item_id}: {error}"));

                acc
            },
        );

        Err(bad_request!("{collected_error_msgs}"))
    } else {
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::add_item_as_schema_opt;
    use crate::api_model::CreateItem;

    use crate::async_db_connection::AsyncTx;
    use crate::internal_error;
    use crate::test_helpers::get_memory_connection;
    use crate::{
        error::Result,
        schema::{Schema, SchemaPropertyType},
        triggers::SchemaAdditionChange,
    };
    use serde_json::json;

    #[tokio::test]
    async fn my_test() -> Result<()> {
        let mut conn = get_memory_connection()?;

        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = Schema::empty();
            schema.set_type_unchecked(
                "Person".to_string(),
                "age".to_string(),
                SchemaPropertyType::Integer,
            );

            let json = json!({
                "type": "Something"
            });
            let create_item: CreateItem = serde_json::from_value(json).unwrap();
            let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
            assert_eq!(result, Ok(SchemaAdditionChange::NotASchema));

            let json = json!({
                "type": "ItemPropertySchema",
                "itemType": "Person",
                "propertyName": "age",
                "valueType": "Integer",
            });
            let create_item: CreateItem = serde_json::from_value(json).unwrap();
            let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
            assert_eq!(
                result,
                Err(internal_error! { "Failed to find property Person.age" })
            );

            let json = json!({
                "type": "ItemPropertySchema",
                "itemType": "Person",
                "propertyName": "agility",
                "valueType": "Integer",
            });
            let create_item: CreateItem = serde_json::from_value(json).unwrap();
            assert_eq!(schema.all_properties_len(), 1);
            let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
            assert_eq!(schema.all_properties_len(), 2);
            assert_eq!(
                schema.property("agility"),
                Some(SchemaPropertyType::Integer).as_ref()
            );
            assert_eq!(result, Ok(SchemaAdditionChange::NewSchemaAdded));

            let json = json!({
                "type": "ItemPropertySchema",
                "itemType": "Person",
                "propertyName": "age",
                "valueType": "Text",
            });
            let create_item: CreateItem = serde_json::from_value(json).unwrap();
            let result = add_item_as_schema_opt(&tx, &mut schema, &create_item);
            assert!(result.is_err(), "result should be an error {result:?}");

            Ok(())
        })
        .await
    }
}
