use crate::actix_endpoints::{
    bulk, create_account, create_edge, create_item, delete_edge_by_source_target, delete_item,
    delete_user, get_edges, get_file, get_item, get_logs, graphql, not_found, oauth1_access_token,
    oauth1_request_token, oauth2_access_token, oauth2_auth_url, oauth2_authorize, plugin_api,
    plugin_api_call, plugin_attach, plugins_status, search, trace_filter, trigger_status,
    update_item, upload_file, upload_file_b, version,
};
use actix_cors::Cors;
use actix_web::{
    self,
    web::{self},
    App, FromRequest, Handler, HttpServer, Responder,
};
use libpod::{
    command_line_interface::CliOptions,
    database_pool::ConnectionPool,
    error::{ErrorContext, Result},
    internal_error, shared_plugins,
};
use rustls::{Certificate, PrivateKey, ServerConfig};
use rustls_pemfile::{certs, pkcs8_private_keys};

use std::{
    collections::HashMap,
    fs::File,
    io::BufReader,
    net::{IpAddr, SocketAddr},
    ops::Deref,
};
use tracing::{info, warn};
use tracing_subscriber::{filter::EnvFilter, reload::Handle};

/// Start web framework with specified APIs.
pub async fn run_server<S: 'static>(
    cli_options: CliOptions,
    trace_handle: Handle<EnvFilter, S>,
) -> Result<()> {
    const DEFAULT_JSON_PAYLOAD_LIMIT: usize = 5 * 1024 * 1024;
    const FILE_BYTES_PAYLOAD_LIMIT: usize = 500 * 1024 * 1024;

    let init_db = web::Data::new(tokio::sync::RwLock::new(
        HashMap::<String, ConnectionPool>::new(),
    ));

    let shared_plugin_db = init_db.clone();

    let cli_options = web::Data::new(cli_options);
    let trace_handle = web::Data::new(trace_handle);

    let cli = cli_options.clone();
    let server = HttpServer::new(move || {
        App::new()
            // TODO: it should not be used in production, but it is what it is - from WARP:
            // headers.insert("Access-Control-Allow-Origin", HeaderValue::from_static("*"));
            .wrap(Cors::permissive())
            .app_data(init_db.clone())
            .app_data(cli.clone())
            .app_data(trace_handle.clone())
            // Endpoints avoid JSON extractor, raw Bytes are taken and deserialized manually,
            // go to pod_handlers, to figure out why
            .app_data(web::PayloadConfig::new(DEFAULT_JSON_PAYLOAD_LIMIT))
            .default_service(web::to(not_found))
            .service(version)
            .route("/trace_filter", web::post().to(trace_filter::<S>))
            .service(
                web::scope("/v4")
                    .service(create_item)
                    .service(get_item)
                    .service(oauth1_request_token)
                    .service(oauth1_access_token)
                    .service(oauth2_auth_url)
                    .service(oauth2_authorize)
                    .service(oauth2_access_token)
                    .service(update_item)
                    .service(bulk)
                    .service(delete_item)
                    .service(delete_user)
                    .service(create_account)
                    .service(create_edge)
                    .service(delete_edge_by_source_target)
                    .service(get_edges)
                    .service(search)
                    .service(graphql)
                    // Change to higher payload limits for file endpoints
                    // TODO: whole api should be refactored more into rest manner
                    // add scope("file"), set a limit here - that's API breaking change
                    // so now post_with_limit hack is in use
                    .service(post_with_limit(
                        "{owner}/upload_file/{database_key}/{expected_sha256}",
                        FILE_BYTES_PAYLOAD_LIMIT,
                        upload_file,
                    ))
                    .service(post_with_limit(
                        "{owner}/upload_file_b/{auth_json}/{expected_sha256}",
                        FILE_BYTES_PAYLOAD_LIMIT,
                        upload_file_b,
                    ))
                    .service(post_with_limit(
                        "{owner}/get_file",
                        FILE_BYTES_PAYLOAD_LIMIT,
                        get_file,
                    ))
                    .service(
                        web::scope("/{owner}/plugin")
                            .service(plugins_status)
                            .service(plugin_api)
                            .service(plugin_api_call)
                            .service(plugin_attach),
                    )
                    .service(get_logs)
                    .service(trigger_status),
            )
    });

    let (address, use_tls) = if cli_options.non_tls || cli_options.insecure_non_tls.is_some() {
        let ip = if let Some(ip) = cli_options.insecure_non_tls {
            if ip.is_loopback() {
                info!(
                    "You seem to use --insecure-non-tls option with a loopback address. \
                    It is recommended to instead use --non-tls CLI option \
                    for clarity and simplicity."
                );
            } else if check_public_ip(&ip) {
                warn!(
                    "The server is asked to run on a public IP {} without https encryption. \
                    This is discouraged as an intermediary (even your router on a local network) \
                    could spoof the traffic sent to the server and do a MiTM attack.\
                    Please consider using Pod with https encryption, \
                    or run it on a non-public network.",
                    ip
                );
            };
            ip
        } else {
            IpAddr::from([127, 0, 0, 1])
        };

        (SocketAddr::new(ip, cli_options.port), false)
    } else {
        (
            SocketAddr::new(IpAddr::from([0, 0, 0, 0]), cli_options.port),
            true,
        )
    };

    let server_handle = actix_web::rt::spawn(async move {
        if use_tls {
            let tls_conf = load_rustls_config(&cli_options)?;
            server.bind_rustls(address, tls_conf)?.run().await?;
        } else {
            server.bind(address)?.run().await?;
        }

        Ok(())
    });

    // Note initializing shared plugins must happen after webserver is up and ready,
    // otherwise starting plugin might not be able to reach the pod - RC for the webserver
    if let Err(e) = shared_plugins::initialize(shared_plugin_db.deref()).await {
        warn!("Failed to initialize shared plugins: {e}");
    }

    server_handle.await?
}

fn load_rustls_config(cli_options: &CliOptions) -> Result<rustls::ServerConfig> {
    // init server config builder with safe defaults
    let config = ServerConfig::builder()
        .with_safe_defaults()
        .with_no_client_auth();

    // load TLS key/cert files
    let cert_file = &mut BufReader::new(
        File::open(&cli_options.tls_pub_crt)
            .context(|| format!("while opening {}", cli_options.tls_pub_crt))?,
    );
    let key_file = &mut BufReader::new(
        File::open(&cli_options.tls_priv_key)
            .context(|| format!("while opening {}", cli_options.tls_priv_key))?,
    );

    // convert files to key/cert objects
    let cert_chain = certs(cert_file)
        .context_str("while loading certificates")?
        .into_iter()
        .map(Certificate)
        .collect();
    let mut keys: Vec<PrivateKey> = pkcs8_private_keys(key_file)
        .context_str("while loading private keys")?
        .into_iter()
        .map(PrivateKey)
        .collect();

    // exit if no keys could be parsed
    if keys.is_empty() {
        return Err(internal_error!("Could not locate PKCS 8 private keys."));
    }

    let config = config
        .with_single_cert(cert_chain, keys.remove(0))
        .map_err(|e| internal_error!("error occurred while building the config: {e}"))?;
    Ok(config)
}

/// Override limit for this exact endpoint
fn post_with_limit<F, Args>(path: &str, limit: usize, handler: F) -> actix_web::Resource
where
    F: Handler<Args>,
    Args: FromRequest + 'static,
    F::Output: Responder + 'static,
{
    web::resource(path)
        .app_data(web::PayloadConfig::new(limit))
        .route(web::post().to(handler))
}

fn check_public_ip(addr: &IpAddr) -> bool {
    match addr {
        IpAddr::V4(v4) => v4.is_private() || v4.is_loopback() || v4.is_link_local(),
        IpAddr::V6(v6) if v6.is_loopback() => true,
        // https://en.wikipedia.org/wiki/Unique_local_address
        // Implementation copied from `v6.is_unique_local()`,
        // which is not yet stabilized in Rust
        IpAddr::V6(v6) if (v6.segments()[0] & 0xfe00) == 0xfc00 => true,
        // https://en.wikipedia.org/wiki/Link-local_address
        // Implementation copied from `v6.is_unicast_link_local()`,
        // which is not yet stabilized in Rust
        IpAddr::V6(v6) if (v6.segments()[0] & 0xffc0) == 0xfe80 => true,
        _ => false,
    }
}
