//! Endpoints definitions in this file HAS to be agnostic in terms of underlying webserver.
//! Use only generic types here.

use http::StatusCode;
use libpod::{
    any_error,
    api_model::{
        AuthKey, Bulk, BulkResponse, ClientAuth, CreateEdge, CreateItem, CreateUserReq,
        DeleteEdgeBySourceTarget, GetEdges, GetFile, Oauth2AccessTokenRequest,
        Oauth2AccessTokenResponse, Oauth2AuthUrlRequest, Oauth2AuthUrlResponse,
        Oauth2AuthorizeTokenRequest, OauthAccessTokenPayload, OauthRequestTokenPayload,
        PayloadWrapper, PluginAttachReq, PluginCallApiReq, PluginCallApiRes, PluginGetApiReq,
        PluginGetApiRes, PluginStatusReq, PluginStatusRes, PodOwner, Search, SendEmail,
        TraceRequest, UpdateItem,
    },
    async_db_connection::AsyncTx,
    bad_request,
    command_line_interface::CliOptions,
    database_api::{self},
    database_pool::{initialize_db, InitDb},
    database_utils, email,
    error::Result,
    error::{ErrorContext, ErrorType},
    file_api, internal_api, internal_error, oauth1_api, oauth2_api,
    plugin_auth_crypto::{auth_to_database_key, DatabaseKey},
    plugin_run, shared_plugins, triggers,
};
use serde_json::Value;
use sha2::{Digest, Sha256};
use std::{
    collections::HashMap,
    sync::{atomic::AtomicU16, Arc},
};
use tokio::task;
use tracing::{debug, info, warn};
use tracing_subscriber::{reload::Handle, EnvFilter};

#[inline(always)]
pub fn version() -> String {
    internal_api::get_project_version()
}

#[inline(always)]
pub async fn create_item(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<CreateItem>,
    cli: &CliOptions,
) -> Result<String> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    internal_api::create_item(payload, &mut conn, &owner, &database_key, cli).await
}

#[inline(always)]
pub async fn get_item(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<String>,
) -> Result<Vec<Value>> {
    let auth = body.auth;
    let payload = body.payload;

    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_read_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;
        internal_api::get_item_tx(&tx, &schema, &payload)
    })
    .await
}
pub async fn oauth1_request_token(body: PayloadWrapper<OauthRequestTokenPayload>) -> Result<Value> {
    let payload = body.payload;
    oauth1_api::oauth1_request_token(payload.callback_url, payload.service).await
}

pub async fn oauth1_access_token(body: PayloadWrapper<OauthAccessTokenPayload>) -> Result<Value> {
    let payload = body.payload;
    oauth1_api::oauth1_access_token(
        payload.oauth_token,
        payload.oauth_verifier,
        payload.oauth_token_secret,
    )
    .await
}
#[inline(always)]
pub async fn oauth2_auth_url(
    body: PayloadWrapper<Oauth2AuthUrlRequest>,
) -> Result<Oauth2AuthUrlResponse> {
    let payload = body.payload;

    oauth2_api::auth_url(payload)
}
#[inline(always)]
pub async fn oauth2_authorize(
    owner: PodOwner,
    body: PayloadWrapper<Oauth2AuthorizeTokenRequest>,
    init_db: &InitDb,
    cli: &CliOptions,
) -> Result<Oauth2AccessTokenResponse> {
    let payload = body.payload;

    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        let mut schema = database_api::get_schema(&tx)?;

        oauth2_api::authorize(&tx, &mut schema, &owner, cli, payload).await
    })
    .await
}
#[inline(always)]
pub async fn oauth2_access_token(
    owner: PodOwner,
    body: PayloadWrapper<Oauth2AccessTokenRequest>,
    init_db: &InitDb,
) -> Result<Oauth2AccessTokenResponse> {
    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;
    let payload = body.payload;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;

        oauth2_api::access_token(&tx, &schema, payload).await
    })
    .await
}
#[inline(always)]
pub async fn update_item(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<UpdateItem>,
) -> Result<()> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;
        internal_api::update_item_tx(&tx, &schema, &payload.id, payload.fields)
    })
    .await
}
#[inline(always)]
pub async fn bulk(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<Bulk>,
    cli: &CliOptions,
) -> Result<BulkResponse> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    let bulk_res = {
        let owner = owner.clone();
        conn.in_write_transaction(|tx: AsyncTx| async move {
            let mut schema = database_api::get_schema(&tx)?;
            internal_api::bulk_tx(&tx, &mut schema, payload, cli, &owner).await
        })
        .await?
    };

    triggers::actions_after_item_create(
        &mut conn,
        &bulk_res.create_items,
        &owner,
        cli,
        &database_key,
    )
    .await?;

    Ok(bulk_res)
}

#[inline(always)]
pub async fn delete_item(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<String>,
) -> Result<()> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;
    conn.in_write_transaction(|tx: AsyncTx| async move {
        internal_api::delete_item_by_id(&tx, &payload)
    })
    .await
}
#[inline(always)]
pub async fn delete_user(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<String>,
) -> Result<()> {
    info!("Deleting user {}", owner);
    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;
    database_utils::check_owner_and_delete_db(&owner, init_db, &database_key).await?;
    Result::Ok(())
}
#[inline(always)]
pub async fn create_account(init_db: &InitDb, body: CreateUserReq) -> Result<PodOwner> {
    info!("Creating account for {}", body.owner_key);
    database_utils::check_owner(&body.owner_key)?;

    let database_key = auth_to_database_key(AuthKey::ClientAuth(ClientAuth {
        database_key: body.database_key,
    }))?;

    initialize_db(&body.owner_key, init_db, &database_key).await?;

    debug!("Account created");

    Ok(body.owner_key)
}
#[inline(always)]
pub async fn create_edge(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<CreateEdge>,
) -> Result<String> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        let mut schema = database_api::get_schema(&tx)?;
        internal_api::create_edge(&tx, &mut schema, payload)
    })
    .await
}
#[inline(always)]
pub async fn delete_edge_by_source_target(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<DeleteEdgeBySourceTarget>,
) -> Result<()> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        internal_api::delete_edge_by_source_target(&tx, payload)
    })
    .await
}
#[inline(always)]
pub async fn get_edges(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<GetEdges>,
) -> Result<Vec<Value>> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_read_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;
        internal_api::get_edges(&tx, payload, &schema)
    })
    .await
}
#[inline(always)]
pub async fn search(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<Search>,
) -> Result<Vec<Value>> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_read_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;
        internal_api::search(&tx, &schema, payload)
    })
    .await
}
#[inline(always)]
pub async fn graphql(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<String>,
) -> Result<HashMap<String, Vec<Value>>> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_read_transaction(|tx: AsyncTx| async move {
        let schema = database_api::get_schema(&tx)?;
        internal_api::graphql(&tx, &schema, payload)
    })
    .await
}

#[inline(always)]
pub async fn upload_file(
    owner: PodOwner,
    init_db: &InitDb,
    database_key: String,
    expected_sha256: String,
    body: &[u8],
) -> Result<()> {
    let database_key = DatabaseKey::from(database_key).map(Arc::new)?;

    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        file_api::upload_file(&tx, &owner, &expected_sha256, body)
    })
    .await
}

#[inline(always)]
pub async fn upload_file_b(
    owner: PodOwner,
    init_db: &InitDb,
    auth_json: String,
    expected_sha256: String,
    body: &[u8],
) -> Result<()> {
    let auth_json = percent_encoding::percent_decode_str(&auth_json)
        .decode_utf8()
        .context_str("Failed to decode UTF-8 authorization JSON")?;

    let auth: AuthKey = serde_json::from_str(&auth_json)?;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_write_transaction(|tx: AsyncTx| async move {
        file_api::upload_file(&tx, &owner, &expected_sha256, body)
    })
    .await
}

#[inline(always)]
pub async fn get_file(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<GetFile>,
) -> Result<Vec<u8>> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_read_transaction(|tx: AsyncTx| async move {
        file_api::get_file(&tx, &owner, &payload.sha256)
    })
    .await
}
#[inline(always)]
pub async fn get_logs(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<String>,
    cli: &CliOptions,
) -> Result<Value> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    conn.in_read_transaction(|tx: AsyncTx| async move { plugin_run::get_logs(&tx, &payload, cli) })
        .await
}

#[inline(always)]
pub async fn send_email(
    owner: PodOwner,
    init_db: &InitDb,
    body: PayloadWrapper<SendEmail>,
    cli: &CliOptions,
) -> Result<()> {
    let auth = body.auth;
    let payload = body.payload;
    let database_key = auth_to_database_key(auth)?;
    let _conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    let cli = cli.clone();
    task::spawn_blocking(move || email::send_email(payload, &cli)).await?
}

#[inline(always)]
pub fn trace_filter<S>(body: TraceRequest, trace_handler: &Handle<EnvFilter, S>) -> Result<()> {
    const SECRET_HASH: &str = "E40C1223ED2AB3C68BFC6F738EE817839F8C3E5155C2E8D67968F39B62293E93";

    let mut hasher = Sha256::new();
    hasher.update(body.secret);
    let res = hasher.finalize();

    if SECRET_HASH != hex::encode_upper(res) {
        return Err(any_error! {
            StatusCode::UNAUTHORIZED,
            "unauthorized"
        });
    }

    warn!("Changing trace filter to {}", body.filter);

    let new_filter = body
        .filter
        .parse::<tracing_subscriber::filter::EnvFilter>()
        .map_err(|e| bad_request! { "{e}" })?;

    trace_handler.reload(new_filter).map_err(|e| {
        internal_error! {"{e}"}
    })
}

#[inline(always)]
pub async fn plugins_status(
    owner: PodOwner,
    init_db: &InitDb,
    cli: &CliOptions,
    body: PayloadWrapper<PluginStatusReq>,
) -> Result<PluginStatusRes> {
    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    let plugins = plugin_run::get_plugins_status(&mut conn, cli, &body.payload).await?;

    let shared_plugins = if let Some(mut conn) = shared_plugins::db_connection(init_db).await? {
        plugin_run::get_plugins_status(&mut conn, cli, &body.payload).await?
    } else {
        HashMap::new()
    };

    Ok(PluginStatusRes {
        plugins,
        shared_plugins,
    })
}
#[inline(always)]
pub async fn plugin_api(
    owner: PodOwner,
    init_db: &InitDb,
    cli: &CliOptions,
    body: PayloadWrapper<PluginGetApiReq>,
) -> Result<PluginGetApiRes> {
    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    let mut api = plugin_run::get_plugin_api(&mut conn, cli, &body.payload).await;

    let plugin_not_found = if let Err(e) = &api {
        matches!(e.inner(), ErrorType::PluginNotFound(_))
    } else {
        false
    };

    if plugin_not_found {
        // Plugin not found, maybe it's shared plugin?
        if let Some(mut conn) = shared_plugins::db_connection(init_db).await? {
            api = plugin_run::get_plugin_api(&mut conn, cli, &body.payload).await;
        }
    }

    api
}
#[inline(always)]
pub async fn plugin_api_call(
    owner: PodOwner,
    init_db: &InitDb,
    cli: &CliOptions,
    body: PayloadWrapper<PluginCallApiReq>,
) -> Result<PluginCallApiRes> {
    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    let mut api_response = plugin_run::call_plugin_api(&mut conn, cli, &body.payload).await;

    let plugin_not_found = if let Err(e) = &api_response {
        matches!(e.inner(), ErrorType::PluginNotFound(_))
    } else {
        false
    };

    if plugin_not_found {
        // Plugin not found, maybe it's shared plugin?
        if let Some(mut conn) = shared_plugins::db_connection(init_db).await? {
            api_response = plugin_run::call_plugin_api(&mut conn, cli, &body.payload).await
        }
    }

    api_response
}

#[inline(always)]
pub async fn plugin_attach(
    owner: PodOwner,
    init_db: &InitDb,
    cli: &CliOptions,
    body: PayloadWrapper<PluginAttachReq>,
) -> Result<String> {
    let auth = body.auth;
    let database_key = auth_to_database_key(auth)?;
    let mut conn =
        database_utils::check_owner_and_initialize_db(&owner, init_db, &database_key).await?;

    let attached_id = plugin_run::plugin_attach(&mut conn, cli, body.payload, &owner).await?;

    Ok(attached_id)
}

/// Get unique identifier for given trace span
pub fn trace_uid() -> u16 {
    static UID: AtomicU16 = AtomicU16::new(0);

    // It safely wraps around during overflow
    UID.fetch_add(1, std::sync::atomic::Ordering::Relaxed)
}

pub fn trigger_status(body: &Value) -> Result<()> {
    // TODO: TBD what to do with the notification
    if let Some(status) = body.pointer("/payload/status") {
        info!("Trigger status: {status:#}");
        Ok(())
    } else {
        Err(bad_request!("Missing status property in the response"))
    }
}
