# Integration tests
This directory contains integration tests of the POD.

## Structure
In `common` directory, there are structs and functions that are useful across the tests. The `test_data` contains context for tests, `pod_client` is a set of functions that allows to communicate, or gather data from the POD.

Note how Rust integration tests works. Each test file is a separate binary. Tests inside a binary are executed **concurrently**, however binaries are executed **one after another**. Keep that in mind if you encounter some weird behavior.

## How it works
The current infrastructure allows to run POD as a process and do tests of it. Most likely it will be HTTP API tests, look to `test_trigger.rs` for examples.

Integration test can be decorated with `#[test_context(TestData)]` that gives the running test access to the Context. The context contains means to communicate with the POD, like the HTTP client. Code execution of integration test with the context is as follows:
```rust
let ctx = TestData{};
ctx.setup().await; // Start the POD, create HTTP client, and do other repetitive stuff that you would have to do in every test manually.

test_to_execute(ctx).await;

ctx.teardown().await; // Teardown is called, even if test_to_execute panics
// It kills the POD under test, removes plugins that were started during test execution, removes the database used by the POD
```
Context is created for each test separately.
You as a developer worries only about test implementation, `setup` and `teardown` are called behind the scenes. Each test operates on it's own dedicated POD instance. Each test uses unique `owner` and `database` keys.
Test isolation is the key to have them robust. Effort should be put in future development, to persist that property.

To mark the tests as *slow* use feature, called `include_slow_tests`. They will be excluded from compilation, if the feature is not enabled.
To disable whole file add an entry in Cargo.toml:

```
[[test]]
name = "test_trigger"
path = "tests/test_trigger.rs"
required-features = ["include_slow_tests"]
```

To disable only specific test in the file use:

```
#[cfg(feature = "include_slow_tests")]
```

## Cheat sheet
Run unit + integration tests by:

```cargo test```

Run only integration tests by:
```
--test name...
    Test the specified integration test. This flag may be specified multiple times and supports common Unix glob patterns.
```

```RUST_LOG=info cargo test   --test "*" -- --nocapture```

Some tests can be really slow, and are disabled by default, if you want to execute them, add feature `include_slow_tests`, like that:

```RUST_LOG=info cargo test   --test "*" --features=include_slow_tests -- --nocapture```

There is also possibility to run CI tasks locally, simply run the following script:
```tools/local_ci.sh```
