use bollard::{
    container::{ListContainersOptions, StopContainerOptions},
    service::ContainerSummary,
    Docker,
};
use serde::Serialize;
use sha2::Digest;

use std::{
    collections::HashMap, ffi::OsStr, fs, hash::Hash, path::PathBuf, process::Stdio,
    sync::atomic::AtomicU32, time::Duration,
};
use test_context::AsyncTestContext;
use tokio::{
    process::{Child, Command},
    sync::mpsc::UnboundedReceiver,
};
use tracing::info;
use tracing_subscriber::{EnvFilter, FmtSubscriber};

use super::pod_client::{start_log_collection, PodClient};
use libpod::database_pool::owner_database_path;
/// Context of the test
pub struct TestData {
    pub owner_key: String,
    pub database_key: String,
    pub pod_handle: Child,
    pub pod_url: String,
    pub pod_port: u32,
    pub pod_client: PodClient,
    pub pod_logs: UnboundedReceiver<String>,
    pub docker_api: Docker,
}

/// Get next free port for POD instance
pub fn pod_port() -> u32 {
    // NOTE: this is not isolated among processes
    static NEXT_PORT: AtomicU32 = AtomicU32::new(3040);

    NEXT_PORT.fetch_add(1, std::sync::atomic::Ordering::Relaxed)
}

pub fn owner_key(port: u32) -> String {
    format!("133700000000000000000000000000000000000000000000000000000000{port}")
}

pub fn database_key(port: u32) -> String {
    format!("DBDBD0000000000000000000000000000000000000000000000000000000{port}")
}

/// Get absolute path to root of the POD project
pub fn root_path() -> PathBuf {
    PathBuf::from(env!("CARGO_MANIFEST_DIR"))
}

/// Depending on build mode find valid POD executable
fn get_pod_path_executable() -> PathBuf {
    let build_mode;

    #[cfg(debug_assertions)]
    {
        build_mode = "debug";
    }
    #[cfg(not(debug_assertions))]
    {
        build_mode = "release";
    }

    // From root path go to partent workspace root, where target dir is present
    root_path()
        .parent()
        .unwrap()
        .join("target")
        .join(build_mode)
        .join("pod")
}

/// Creates a POD process for given owner, returns process handle, together with log handle
pub async fn create_pod<T, U>(args: T) -> (Child, UnboundedReceiver<String>)
where
    T: IntoIterator<Item = U>,
    U: AsRef<OsStr>,
{
    let mut pod_handle = Command::new(get_pod_path_executable())
        .env("RUST_LOG", option_env!("RUST_LOG").unwrap_or("info"))
        .args(args)
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()
        .unwrap();

    // Take the process some time to startup the webservice
    tokio::time::sleep(Duration::from_secs(1)).await;

    let pod_logs = start_log_collection(
        pod_handle.stdout.take().unwrap(),
        pod_handle.stderr.take().unwrap(),
    );

    (pod_handle, pod_logs)
}

impl TestData {
    pub async fn stop_containers(&self) {
        let containers = self
            .list_containers(ListContainersOptions {
                filters: HashMap::from([("name", vec![self.owner_key.as_ref()])]),
                ..Default::default()
            })
            .await;

        for container in containers {
            info!("Stop container: {:?}", container.id);

            let _res = self
                .docker_api
                .stop_container(
                    container.id.as_ref().unwrap(),
                    Some(StopContainerOptions { t: 5 }),
                )
                .await;
        }
    }

    pub async fn list_containers<T>(
        &self,
        filter: ListContainersOptions<T>,
    ) -> Vec<ContainerSummary>
    where
        T: Into<String> + Eq + Hash + Serialize,
    {
        self.docker_api
            .list_containers(Some(filter))
            .await
            .expect("Failed to list running containers")
    }

    pub async fn remove_db(&self) {
        let database_path = owner_database_path(&self.owner_key);
        let path = root_path().join(database_path);
        info!("Path to remove {path:?}");
        let _ = fs::remove_file(path);
    }
}

#[async_trait::async_trait]
impl AsyncTestContext for TestData {
    async fn setup() -> Self {
        let subscriber = FmtSubscriber::builder()
            .with_env_filter(
                EnvFilter::builder()
                    .with_default_directive("info".parse().unwrap())
                    .from_env_lossy(),
            )
            .without_time()
            .finish();

        let _ = tracing::subscriber::set_global_default(subscriber);

        let pod_port = pod_port();
        let owner_key = owner_key(pod_port);

        let owner_hash = sha2::Sha256::new_with_prefix(hex::decode(&owner_key).unwrap()).finalize();
        let owner_hash = hex::encode(owner_hash);

        let (pod_handle, pod_logs) = create_pod([
            format!("--owners={owner_hash}"),
            "--non-tls".to_string(),
            format!("--port={pod_port}"),
        ])
        .await;

        let database_key = database_key(pod_port);
        let pod_url = format!("http://localhost:{pod_port}");

        let docker_api = Docker::connect_with_local_defaults().unwrap();

        Self {
            pod_client: PodClient::new(&database_key, &owner_key, &pod_url),
            pod_logs,
            pod_port,
            owner_key,
            database_key,
            pod_url,
            pod_handle,
            docker_api,
        }
    }

    async fn teardown(mut self) {
        self.stop_containers().await;

        let _ = self.pod_handle.kill().await;
        self.remove_db().await;
    }
}
