mod common;

use common::pod_client::PodClient;
use common::test_data::{create_pod, TestData};

use reqwest::{Response, StatusCode};
use serde_json::json;
use test_context::test_context;

async fn use_pod(client: &PodClient) -> Response {
    let payload = json!(
    {
        "type": "ItemPropertySchema",
        "itemType": "Message",
        "propertyName": "Content",
        "valueType": "Text"
    });

    client.post_to(payload, "create_item").await
}

// The truth table of possible states.
// The DB can already exist or not,
// The connection pool can already exist or not.
// User uses /register endpoint, or any other

// # | db exist  | connection pool created  | call /register xor call other endpoint | expected result
// 1 |  0       |         0                |                               1         | state unaffected, error: unknown account
// 2 |  0       |         0                |       1                                 | db, and pool created
// 3 |  0       |         1                |       1                                 | unreachable -> return internal server error
// 4 |  0       |         1                |                               1         | unreachable -> will return rusqlite error upon trying to access the db
// 5 |  1       |         0                |       1                                 | error -> user already exists
// 6 |  1       |         0                |                               1         | create connection pool, validate credentials, proceed
// 7 |  1       |         1                |       1                                 | error -> user already exists
// 8 |  1       |         1                |                               1         | check credentials, proceed

#[test_context(TestData)]
#[tokio::test]
async fn test_account_use_not_registered(ctx: &mut TestData) {
    let valid_credentials = PodClient::new(&ctx.database_key, &ctx.owner_key, &ctx.pod_url);

    // State #1, trying to use pod, while not yet registered
    let res = use_pod(&valid_credentials).await;
    assert_eq!(res.status(), StatusCode::UNAUTHORIZED);
}

#[test_context(TestData)]
#[tokio::test]
async fn test_account_create(ctx: &mut TestData) {
    let valid_credentials = PodClient::new(&ctx.database_key, &ctx.owner_key, &ctx.pod_url);

    // State #2, register from the list - the happy path
    let register_request = json!( {
        "ownerKey":  valid_credentials.owner_key,
        "databaseKey": valid_credentials.database_key

    });

    let res = ctx
        .pod_client
        .client
        .post(format!("{}/v4/account", valid_credentials.pod_url))
        .json(&register_request)
        .send()
        .await
        .unwrap();

    assert_eq!(res.status(), StatusCode::OK);

    // Cannot register twice, state #7
    let res = ctx
        .pod_client
        .client
        .post(format!("{}/v4/account", valid_credentials.pod_url))
        .json(&register_request)
        .send()
        .await
        .unwrap();

    assert_eq!(res.status(), StatusCode::BAD_REQUEST);

    assert_eq!(
        res.json::<String>().await.unwrap(),
        format!(
            "Failure: Error 400 Bad Request, Account for {} already exist",
            valid_credentials.owner_key
        )
    );

    // Can use the POD, state #8
    let res = use_pod(&valid_credentials).await;
    assert_eq!(res.status(), StatusCode::OK);

    // Cannot use the POD with invalid password
    let good_login_bad_password = PodClient::new(&ctx.owner_key, &ctx.owner_key, &ctx.pod_url);
    let res = use_pod(&good_login_bad_password).await;
    assert_eq!(res.status(), StatusCode::UNAUTHORIZED);
    assert_eq!(
        res.json::<String>().await.unwrap(),
        "Failure: Error 401 Unauthorized, Invalid password key"
    );

    // Remove the db to simulate situation defined in state #3:
    // db disappeared, user tries incorrectly to register again
    ctx.remove_db().await;

    let res = ctx
        .pod_client
        .client
        .post(format!("{}/v4/account", valid_credentials.pod_url))
        .json(&register_request)
        .send()
        .await
        .unwrap();

    assert_eq!(res.status(), StatusCode::INTERNAL_SERVER_ERROR);

    assert_eq!(
        res.json::<String>().await.unwrap(),
        format!(
            "Failure: Error 500 Internal Server Error, DB file does not exist but connection pool does for {}",
            valid_credentials.owner_key
        )
    );

    // State #4: db disappeared, user tries to use the pod
    let res = use_pod(&valid_credentials).await;
    assert_eq!(res.status(), StatusCode::INTERNAL_SERVER_ERROR);

    assert_eq!(
        res.json::<String>().await.unwrap(),
        format!(
            "Failure: Error 500 Internal Server Error, No DB, but connection pool exist for {}",
            valid_credentials.owner_key
        )
    );
}

#[test_context(TestData)]
#[tokio::test]
async fn test_account_works_after_pod_restart(ctx: &mut TestData) {
    let valid_credentials = PodClient::new(&ctx.database_key, &ctx.owner_key, &ctx.pod_url);

    // # 2 register from the list - the happy path
    let register_request = json!( {
        "ownerKey":  valid_credentials.owner_key,
        "databaseKey": valid_credentials.database_key

    });

    let res = ctx
        .pod_client
        .client
        .post(format!("{}/v4/account", valid_credentials.pod_url))
        .json(&register_request)
        .send()
        .await
        .unwrap();

    assert_eq!(res.status(), StatusCode::OK);

    let res = use_pod(&valid_credentials).await;
    assert_eq!(res.status(), StatusCode::OK);

    // Restart the POD
    let _ = ctx.pod_handle.kill().await;
    ctx.pod_logs.close();

    let (pod_handle, pod_logs) = create_pod([
        "--owners=ANY",
        "--non-tls",
        &format!("--port={}", ctx.pod_port),
    ])
    .await;

    ctx.pod_handle = pod_handle;
    ctx.pod_logs = pod_logs;

    // Can use the POD, state #6
    let res = use_pod(&valid_credentials).await;
    assert_eq!(res.status(), StatusCode::OK);

    // Cannot register again, state #5
    let res = ctx
        .pod_client
        .client
        .post(format!("{}/v4/account", valid_credentials.pod_url))
        .json(&register_request)
        .send()
        .await
        .unwrap();

    assert_eq!(res.status(), StatusCode::BAD_REQUEST);

    assert_eq!(
        res.json::<String>().await.unwrap(),
        format!(
            "Failure: Error 400 Bad Request, Account for {} already exist",
            valid_credentials.owner_key
        )
    );
}

#[test_context(TestData)]
#[tokio::test]
async fn test_account_outside_allow_list(ctx: &mut TestData) {
    let user_outside_list = PodClient::new(
        &ctx.database_key,
        &String::from_utf8(ctx.owner_key.bytes().rev().collect()).unwrap(),
        &ctx.pod_url,
    );

    // # 1 trying to use pod, while not yet registered
    let res = use_pod(&user_outside_list).await;
    assert_eq!(res.status(), StatusCode::FORBIDDEN);
}
