mod common;

use crate::common::pod_request::{migrate_plugin_definition, register_account, start_plugin};
use common::test_data::TestData;
use libpod::api_model::{PluginCallApiRes, PluginGetApiRes};
use serde_json::json;
use test_context::test_context;

#[test_context(TestData)]
#[tokio::test]
async fn test_plugin_api(ctx: &mut TestData) {
    register_account(&ctx.pod_client).await;

    migrate_plugin_definition(&ctx.pod_client).await;

    let (id, alias) = start_plugin(ctx).await;

    test_plugin_get_api(id.clone(), alias.clone(), ctx).await;

    test_plugin_call_api(id, alias, ctx).await;
}

async fn test_plugin_get_api(id: String, alias: String, ctx: &TestData) {
    let payload = json!({ "id": id });

    let response = ctx.pod_client.post_to(payload, "plugin/api").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginGetApiRes>().await.unwrap();

    assert_eq!(
        json!({
            "/v1/api":  {
                "method": "GET",
                "args": {},
            },
            "/v1/echo":  {
                "method": "POST",
                "args": {
                    "x": "int",
                    "req": "ReqModel",
                    "a": "typing.List[~AnyStr]",
                    "b": "dict",
                    "t": "tuple",
                },
            },
            "/v1/health":  {
                "method": "GET",
                "args": {},
            },

        }),
        serde_json::to_value(response).unwrap()
    );

    // Same, but using alias
    let payload = json!({ "id": alias });

    let response = ctx.pod_client.post_to(payload, "plugin/api").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginGetApiRes>().await.unwrap();

    assert_eq!(
        json!({
            "/v1/api":  {
                "method": "GET",
                "args": {},
            },
            "/v1/echo":  {
                "method": "POST",
                "args": {
                    "x": "int",
                    "req": "ReqModel",
                    "a": "typing.List[~AnyStr]",
                    "b": "dict",
                    "t": "tuple",
                },
            },
            "/v1/health":  {
                "method": "GET",
                "args": {},
            },

        }),
        serde_json::to_value(response).unwrap()
    );

    let payload = json!(
        {
            "id" : "invalid-id"
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/api").await;

    assert!(!response.status().is_success());

    let response: String = response.json().await.unwrap();

    assert_eq!(
        response,
        "Failure: Error 400 Bad Request, failed to find plugin with id Alias(\"invalid-id\")"
    );
}

async fn test_plugin_call_api(id: String, alias: String, ctx: &TestData) {
    // Valid call
    let payload = json!({
      "id": id,
      "method": "POST",
      "endpoint": "/v1/echo",

      "query": {
        "x": 1234
      },

      "jsonBody": {
        "a": ["string"],

        "b": {},

        "req": {
            "x": 112233,
            "y": false,
            "d": "2022-10-24T19:38:09.718Z"
        },

        "t": [3,2,1]
      }
    });

    let response = ctx.pod_client.post_to(payload, "plugin/api/call").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginCallApiRes>().await.unwrap();

    assert_eq!(
        serde_json::to_value(response).unwrap(),
        json!({
            "statusCode": 200,
            "body":{
                "x": 1234,
                "a": ["string"],
                "b": {},
                "req": {
                    "x": 112233,
                    "y": false,
                    "d": "2022-10-24T19:38:09.718000+00:00"
                },
                "t": [3,2,1]
            }
        })
    );

    // Same, but using alias
    let payload = json!({
      "id": alias,
      "method": "POST",
      "endpoint": "/v1/echo",

      "query": {
        "x": 1234
      },

      "jsonBody": {
        "a": ["string"],

        "b": {},

        "req": {
            "x": 112233,
            "y": false,
            "d": "2022-10-24T19:38:09.718Z"
        },

        "t": [3,2,1]
      }
    });

    let response = ctx.pod_client.post_to(payload, "plugin/api/call").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginCallApiRes>().await.unwrap();

    assert_eq!(
        serde_json::to_value(response).unwrap(),
        json!({
            "statusCode": 200,
            "body":{
                "x": 1234,
                "a": ["string"],
                "b": {},
                "req": {
                    "x": 112233,
                    "y": false,
                    "d": "2022-10-24T19:38:09.718000+00:00"
                },
                "t": [3,2,1]
            }
        })
    );

    // Invalid call, wrong method
    let payload = json!({
      "id": id,
      "method": "PUT",
      "endpoint": "/v1/echo",

      "query": {
        "x": 1234
      },

      "jsonBody": {
        "a": ["string"],

        "b": {},

        "req": {
            "x": 112233,
            "y": false,
            "d": "2022-10-24T19:38:09.718Z"
        },

        "t": [3,2,1]
      }
    });

    let response = ctx.pod_client.post_to(payload, "plugin/api/call").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginCallApiRes>().await.unwrap();

    assert_eq!(
        serde_json::to_value(response).unwrap(),
        json!({
            "statusCode": 405,
            "body": {
                "detail": "Method Not Allowed"
            }
        })
    );

    // Request status of selected invalid plugin
    let payload = json!(
        {
            "id" : "invalid-id",
            "method": "POST",
            "endpoint": "/v1/echo",
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/api/call").await;

    assert!(!response.status().is_success());

    let response: String = response.json().await.unwrap();

    assert_eq!(
        response,
        "Failure: Error 400 Bad Request, failed to find plugin with id Alias(\"invalid-id\")"
    );
}
