mod common;
use crate::common::pod_request::{migrate_plugin_definition, register_account, start_plugin};
use common::test_data::TestData;
use libpod::api_model::{PluginStatus, PluginStatusMetadata, PluginStatusRes};
use serde_json::json;
use test_context::test_context;
use tracing::error;

#[test_context(TestData)]
#[tokio::test]
async fn test_plugin_state(ctx: &mut TestData) {
    register_account(&ctx.pod_client).await;

    migrate_plugin_definition(&ctx.pod_client).await;

    // Request status of all plugins on fresh instance, should be none
    let payload = json!(
        {
            "plugins" : []
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/status").await;

    assert!(response.status().is_success());
    assert_eq!(
        response
            .json::<PluginStatusRes>()
            .await
            .unwrap()
            .plugins
            .len(),
        0
    );

    let (plugin_id, plugin_alias) = start_plugin(ctx).await;

    // Request status of all plugins
    let payload = json!(
        {
            "plugins" : []
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/status").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginStatusRes>().await.unwrap();
    assert_eq!(response.plugins.len(), 1);

    error!("status: {:#?}", response.plugins.get(&plugin_id).unwrap());

    assert!(matches!(
        response.plugins.get(&plugin_id).unwrap(),
        PluginStatusMetadata {
            status: PluginStatus::Running(_),
            alias: Some(alias)
        } if alias == &plugin_alias
    ));

    // Request status of selected plugins
    let payload = json!(
        {
            "plugins" : [plugin_id, "invalid-id"]
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/status").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginStatusRes>().await.unwrap();
    assert_eq!(response.plugins.len(), 1);
    assert!(matches!(
        response.plugins.get(&plugin_id).unwrap(),
        PluginStatusMetadata {
            status: PluginStatus::Running(_),
            alias: Some(alias)
        }  if alias == &plugin_alias
    ));

    // Request status of selected plugin, using alias
    let payload = json!(
        {
            "plugins" : [plugin_alias, "invalid-id"]
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/status").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginStatusRes>().await.unwrap();
    assert_eq!(response.plugins.len(), 1);
    assert!(matches!(
        response.plugins.get(&plugin_id).unwrap(),
        PluginStatusMetadata {
            status: PluginStatus::Running(_),
            alias: Some(alias)
        } if alias == &plugin_alias
    ));

    // Request status of selected invalid plugin
    let payload = json!(
        {
            "plugins" : ["invalid-id"]
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/status").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginStatusRes>().await.unwrap();
    assert_eq!(response.plugins.len(), 0);

    ctx.stop_containers().await;

    // Request status of selected plugin, that does not run
    let payload = json!(
        {
            "plugins" : [plugin_id, "invalid-id"]
        }
    );

    let response = ctx.pod_client.post_to(payload, "plugin/status").await;

    assert!(response.status().is_success());

    let response = response.json::<PluginStatusRes>().await.unwrap();
    assert_eq!(response.plugins.len(), 1);
    assert!(matches!(
        response.plugins.get(&plugin_id).unwrap(),
        PluginStatusMetadata {
            status: PluginStatus::Unreachable,
            alias: Some(alias)
        }  if alias == &plugin_alias
    ));
}
