mod common;
use crate::common::{
    pod_request::{migrate_plugin_definition, register_account},
    test_data::root_path,
};
use bollard::{
    container::{ListContainersOptions, StopContainerOptions},
    Docker,
};
use common::{
    pod_client::PodClient,
    test_data::{create_pod, database_key, owner_key, pod_port, TestData},
};
use libpod::{
    api_model::{PluginStatus, PluginStatusRes},
    database_pool::owner_database_path,
};
use serde_json::json;
use std::{collections::HashMap, fs, time::Duration};
use test_context::{test_context, AsyncTestContext};
use tokio::time::timeout;
use tracing::info;
use tracing_subscriber::{EnvFilter, FmtSubscriber};

const SHARED_PLUGIN1: &str = "566d75e8-6163-4f04-a5eb-7a6239dc7437";
const SHARED_PLUGIN2: &str = "8562b3fd-f227-43ba-9d21-657a075651a9";

#[test_context(TestDataSharedPlugin)]
#[tokio::test]
async fn test_shared_plugin(ctx: &mut TestDataSharedPlugin) {
    // POD started with 2 shared plugins, user creates own plugin:
    register_account(&ctx.base.pod_client).await;

    migrate_plugin_definition(&ctx.base.pod_client).await;

    let payload = json!(
        {
            "createItems":[
                {
                    // Valid plugin
                    "type": "PluginRun",
                    "id": "d669792f-cc3f-4598-984e-63f63931925d",

                    "containerImage": "gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest",
                    "pluginModule": "plugin_to_trigger.plugin",
                    "pluginName": "ClassifierPlugin",
                    "status": "idle",
                    "targetItemId": "d669792f-cc3f-4598-984e-63f63931925d"

                }
            ]
        }
    );

    let resp = ctx.base.pod_client.post_to(payload, "bulk").await;

    assert_eq!(resp.status(), reqwest::StatusCode::OK);

    timeout(Duration::from_secs(30), async {
        loop {
            // Request status of all plugins, should contain shared, and user specific
            let payload = json!(
                {
                    "plugins" : []
                }
            );
            let response = ctx.base.pod_client.post_to(payload, "plugin/status").await;
            assert!(response.status().is_success());

            let response_plugins = response.json::<PluginStatusRes>().await.unwrap();

            info!("plugins status: {response_plugins:#?}");

            if response_plugins.plugins.len() == 1
                && response_plugins.shared_plugins.len() == 2
                && response_plugins
                    .plugins
                    .iter()
                    .all(|(_id, metadata)| metadata.status != PluginStatus::Unreachable)
                && response_plugins
                    .shared_plugins
                    .iter()
                    .all(|(_id, metadata)| metadata.status != PluginStatus::Unreachable)
            {
                break;
            }

            info!("Waiting for plugins to settle...");
            tokio::time::sleep(Duration::from_secs(1)).await;
        }
    })
    .await
    .unwrap();

    // It is possible to get API of shared plugin
    let payload = json!({ "id": SHARED_PLUGIN1 });

    let response = ctx.base.pod_client.post_to(payload, "plugin/api").await;

    assert!(response.status().is_success());

    // It is possible to call that API
    let payload = json!({
      "id": SHARED_PLUGIN1,
      "method": "POST",
      "endpoint": "/v1/echo",

      "query": {
        "x": 1234
      },

      "jsonBody": {
        "a": ["string"],

        "b": {},

        "req": {
            "x": 112233,
            "y": false,
            "d": "2022-10-24T19:38:09.718Z"
        },

        "t": [3,2,1]
      }
    });

    let response = ctx
        .base
        .pod_client
        .post_to(payload, "plugin/api/call")
        .await;

    assert!(response.status().is_success());

    // TODO: check for shared plugins
    // get health for all
    // get api for shared
}

struct TestDataSharedPlugin {
    pub base: TestData,
    pub pod_owner_key: String,
}

impl TestDataSharedPlugin {
    async fn stop_containers(&self) {
        self.base.stop_containers().await;

        let containers = self
            .base
            .list_containers(ListContainersOptions {
                filters: HashMap::from([("name", vec![self.pod_owner_key.as_ref()])]),
                ..Default::default()
            })
            .await;

        for container in containers {
            info!("Stop shared container: {:?}", container.id);

            let _res = self
                .base
                .docker_api
                .stop_container(
                    container.id.as_ref().unwrap(),
                    Some(StopContainerOptions { t: 5 }),
                )
                .await;
        }
    }
    async fn remove_db(&self) {
        self.base.remove_db().await;

        let database_path = owner_database_path(&self.pod_owner_key);
        let path = root_path().join(database_path);
        info!("Path to remove {path:?}");
        let _ = fs::remove_file(path);
    }
}

#[async_trait::async_trait]
impl AsyncTestContext for TestDataSharedPlugin {
    async fn setup() -> Self {
        let subscriber = FmtSubscriber::builder()
            .with_env_filter(
                EnvFilter::builder()
                    .with_default_directive("info".parse().unwrap())
                    .from_env_lossy(),
            )
            .without_time()
            .finish();

        let _ = tracing::subscriber::set_global_default(subscriber);

        let pod_owner_key = owner_key(pod_port());
        let pod_db_key = "9684544858277421540522235608665732630370684094198630717255861961";

        let pod_port = pod_port();
        let owner_key = owner_key(pod_port);

        let args = vec![
            "--non-tls".to_string(),
            format!("--port={pod_port}"),
            format!("--db-key-for-pod={pod_db_key}"),
            format!("--owner-key-for-pod={pod_owner_key}"),
            format!(
                "--shared-plugins={}",
                json!({
                    "type": "PluginRun",
                    "id": SHARED_PLUGIN1,

                    "containerImage": "gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest",
                    "pluginModule": "plugin_to_trigger.plugin",
                    "pluginName": "ClassifierPlugin",
                    "status": "idle",
                    "targetItemId": SHARED_PLUGIN1
                })
            ),
            format!(
                "--shared-plugins={}",
                json!({
                    "type": "PluginRun",
                    "id": SHARED_PLUGIN2,

                    "containerImage": "gitlab.memri.io:5050/szimnowoda/plugin_to_trigger:main-latest",
                    "pluginModule": "plugin_to_trigger.plugin",
                    "pluginName": "ClassifierPlugin",
                    "status": "idle",
                    "targetItemId": SHARED_PLUGIN2
                })
            ),
            format!("--owners=ANY"),
        ];

        let (pod_handle, pod_logs) = create_pod(args).await;

        let database_key = database_key(pod_port);
        let pod_url = format!("http://localhost:{pod_port}");

        // Take some time to start the plugins, and the pod webserver
        timeout(Duration::from_secs(30), async {
            while reqwest::get(format!("{pod_url}/version")).await.is_err() {
                tokio::time::sleep(Duration::from_millis(200)).await;
            }
        })
        .await
        .unwrap();

        let docker_api = Docker::connect_with_local_defaults().unwrap();

        Self {
            base: TestData {
                pod_client: PodClient::new(&database_key, &owner_key, &pod_url),
                pod_logs,
                pod_port,
                owner_key,
                database_key,
                pod_url,
                pod_handle,
                docker_api,
            },
            pod_owner_key,
        }
    }

    async fn teardown(mut self) {
        self.stop_containers().await;

        let _ = self.base.pod_handle.kill().await;
        self.remove_db().await;
    }
}
