"""
This file simulates FrontEnd that does some random Read/Write
And and Importer plugin that sends messages. Those messages
trigger 'classifier' plugin, that sends the response back to the POD.

The importer plugin creates docker container for the 'classifier' plugin
Be careful while simulating hundreds of users, because each of them will have
the classifier running.
"""
import asyncio
import json
import os
from unicodedata import name

import aiohttp
from locust import FastHttpUser, between, events, tag, task

key = "dbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdbdb"

POD_OWNERS_COUNT = None
POD_OWNERS = None

POD_DB_DIR_ENV = "POD_DB_DIR"
CLIENT_AUTH = {"type": "ClientAuth", "databaseKey": key}
# TODO:
# use returned id, don't hardcode id
# split deploy plugin to two requests in order to create an edge


class FrontEnd(FastHttpUser):
    CURRENT_OWNER = 0

    wait_time = between(0.5, 1.5)

    def get_owner(self):
        owner = POD_OWNERS[FrontEnd.CURRENT_OWNER]
        FrontEnd.CURRENT_OWNER += 1
        return owner

    def on_start(self):
        self.pod_owner = self.get_owner()
        print(f"FrontEnd started for {self.pod_owner}")

    def on_stop(self):
        FrontEnd.CURRENT_OWNER -= 1

    @task(5)
    def queries_data(self):
        self.client.post("/v4/"+self.pod_owner+"/search",
                         json={
                             "auth": CLIENT_AUTH,
                             "payload": {
                                 "dateServerModified>=": 0,
                                 "[[edges]]": {},
                                 "_limit": 100,
                                 "_sortOrder": "Desc"
                             }}, name=f"{self.pod_owner} {self.__class__.__name__}.queries_data")

    @task(1)
    def some_write(self):
        self.client.post("/v4/"+self.pod_owner+"/create_item",
                         json={"auth": CLIENT_AUTH,
                               "payload": {
                                   "type": "File",
                                   "sha256": "fake sha256",
                                   "key": "fake file key",
                                   "nonce": "fake nonce"
                               }},
                         name=f"{self.pod_owner} {self.__class__.__name__}.some_write"
                         )


class ClassifierPlugin(FastHttpUser):
    """
    Starts the 'classifier' plugin
    Registers trigger on new message -> will be passed through to the classifier
    Starts a task that sends the message to the pod.
    Task makes following traffic:
    Message -> POD - (trigger) -> Plugin
    Plugin -> (response to trigger) -> POD

    """
    CURRENT_OWNER = 0
    PLUGIN_DIR = "./plugin_request"

    DEPLOY_DUMMY_PLUGIN = "deploy_dummy_plugin.json"
    START_PLUGIN = "start_plugin.json"
    START_DUMMY_PLUGIN = "start_dummy_plugin.json"
    REGISTER_TRIGGER = "register_trigger.json"

    wait_time = between(1, 1.5)

    def get_owner(self):
        owner = POD_OWNERS[ClassifierPlugin.CURRENT_OWNER]
        ClassifierPlugin.CURRENT_OWNER += 1
        return owner

    def create_initial_messages(self, num_messages=500):
        """
        Simulate the initial import of an importer,
        with dateReceived < trigger.filterCreatedAfter
        """
        messages = []
        for _ in range(num_messages):
            messages.append({
                "type": "Message",
                "content": "a"*500,
                "dateReceived": 1655170839991,
                "isMock": True
            })

        self.client.post("/v4/"+self.pod_owner+"/bulk", json={
            "auth": CLIENT_AUTH,
            "payload": {
                "createItems": messages,
            }
        })

    def on_start(self):
        """Start the plugin"""
        self.pod_owner = self.get_owner()
        self.create_initial_messages()

        with open(os.path.join(ClassifierPlugin.PLUGIN_DIR, ClassifierPlugin.START_DUMMY_PLUGIN), 'r') as file:
            payload = json.loads(file.read())

            self.client.post("/v4/"+self.pod_owner+"/bulk", json={
                "auth": CLIENT_AUTH,
                "payload": payload}, name=f"{self.pod_owner} {self.__class__.__name__}.on_start")

        with open(os.path.join(ClassifierPlugin.PLUGIN_DIR, ClassifierPlugin.REGISTER_TRIGGER), 'r') as file:
            payload = json.loads(file.read())

            self.client.post("/v4/"+self.pod_owner+"/bulk", json={
                "auth": CLIENT_AUTH,
                "payload": payload}, name=f"{self.pod_owner} {self.__class__.__name__}.on_start")

        print(f"ClassifierPlugin started for {self.pod_owner}")

    def on_stop(self):
        print(f"ClassifierPlugin stop for {self.pod_owner}")
        ClassifierPlugin.CURRENT_OWNER -= 1
        # TODO: plugin stop??

    @task(1)
    def trigger_message_wait_for_classification(self):
        # Send a message, trigger will fire up, ping the classifier plugin to do the stuff
        response = self.client.post("/v4/"+self.pod_owner+"/create_item",
                                    json={"auth": CLIENT_AUTH,
                                          "payload": {
                                              "type": "Message",
                                              "content": "a"*500,
                                              "dateReceived": 1655170839993,
                                              "isMock": True,
                                          }},
                                    name=f"{self.pod_owner} {self.__class__.__name__}.classification"
                                    )

        message_id = response.text.strip('"')

        search_result = self.client.post("/v4/"+self.pod_owner+"/search",
                                         json={
                                             "auth": CLIENT_AUTH,
                                             "payload": {
                                                 "id": message_id,
                                                 "_limit": 1,
                                                 "_sortOrder": "Desc"
                                             }}, name=f"{self.pod_owner} {self.__class__.__name__}.queries_data")

        # TODO: check for classification to complete


async def post_bulk(session, base_url, owner, payload):
    url = base_url + "/v4/" + owner + "/bulk"

    async with session.request("POST", url,
                               json={"auth": CLIENT_AUTH, "payload": payload}) as resp:
        return (resp.status, url)


async def create_accounts(base_url, session):
    tasks = []
    for owner in POD_OWNERS:
        tasks.append(asyncio.create_task(create_new_account(owner, base_url, session)))

    _responses = await asyncio.gather(*tasks)


async def create_new_account(owner, base_url, session):
    url = base_url + "/v4/account"

    async with session.request("POST", url,
                               json={"ownerKey": owner, "databaseKey": key}) as resp:
        return (resp.status, url)


async def flutter_migration(base_url, session):
    """Fills the database with the schema definitions from the frontend"""
    FLUTTER_MIGRATION = "./flutter_migration"
    for bulk in sorted(os.listdir(FLUTTER_MIGRATION)):
        print(f"bulk {bulk}")
        with open(os.path.join(FLUTTER_MIGRATION, bulk), 'r') as file:
            payload = json.loads(file.read())

            tasks = []
            for owner in POD_OWNERS:
                tasks.append(asyncio.create_task(post_bulk(session, base_url, owner, payload)))

            _responses = await asyncio.gather(*tasks)


async def fill_db(base_url):
    """Create DB for each POD owner with some data
    note each DB makes POD to allocate at least that many FDs:
    opened_fds = nr_of_cores * 3 (.sqlite, -shm. -wal files) + at least one TCP connection from locust

    Please keep it in mind when launching with default linux FD limit set to 1024
    """

    async with aiohttp.ClientSession() as session:
        await create_accounts(base_url, session)
        await flutter_migration(base_url, session)


@events.test_start.add_listener
def on_test_start(environment, **kwargs):
    global POD_OWNERS_COUNT
    global POD_OWNERS

    num_users = environment.parsed_options.num_users

    if not num_users:
        print("Please pass num users as an parameter to locust CLI")
        os.abort()

    user_classes = environment.user_classes

    print(f"user_classes: {user_classes}")

    # The num_users passed as a parameter to locust does not correspond to the number of POD owners
    # Divide num_users by number of locust client class that are taken to the simulation
    POD_OWNERS_COUNT = num_users // len(user_classes)

    # Note beef has to be lowercase, otherwise kubernetes container will not run
    POD_OWNERS = ["beef%05d" %
                  i for i in range(POD_OWNERS_COUNT)]

    # Set exactly one FrontEnd and one ClassifierPlugin per POD owner
    # fixed_count cannot be greater than num_users
    FrontEnd.fixed_count = POD_OWNERS_COUNT
    ClassifierPlugin.fixed_count = POD_OWNERS_COUNT

    print(f"A new test for {environment.host} is starting. Fill up the DB...")
    asyncio.run(fill_db(environment.host))


@events.test_stop.add_listener
def on_test_stop(environment, **kwargs):
    pass

    # if not os.getenv(POD_DB_DIR_ENV):
    #     print(f"{POD_DB_DIR_ENV} env not set, cleanup aborted")
    #     return

    # DB_SUFFIX = ".p4.4.sqlite"

    # print(f"A test for {environment.host} is ending, cleaning up databases...")

    # db_dir = os.getenv(POD_DB_DIR_ENV)

    # for owner in POD_OWNERS:
    #     db_path = os.path.join(db_dir, owner)

    #     try:
    #         os.remove(db_path + DB_SUFFIX)
    #     except Exception as e:
    #         pass

    #     try:
    #         os.remove(db_path + DB_SUFFIX + "-wal")
    #     except Exception as e:
    #         pass

    #     try:
    #         os.remove(db_path + DB_SUFFIX + "-shm")
    #     except Exception as e:
    #         pass
